package ME::VFS;

=head1 NAME

ME::VFS - the new world order!

=head1 METHODS

=over 4

=cut
use strict;
use warnings;

# Display Modes {{{
use constant DisplayModeFilesOnly => 0;
use constant DisplayModeIndexes => 1;
use constant DisplayModePhotoGallery => 2;
use constant DisplayModeRedirect => 4;
use constant DisplayModeCloaked => 5;
use constant DisplayModeRedirectInternal => 6;
use constant DisplayModeIndexesUpload => 7;

# Multiple gallery display styles now promoted to 100 and above
use constant DisplayModePhotoGalleries => 100;

our %DisplayModes = (
  0 => ['Files only', 'F'],
  1 => ['Files & Dir listing', 'F'],
  2 => ['Photo gallery', 'F'],
  #3 =>  was old photo gallery style
  4 => ['Redirect to external site', 'R', 'E'],
  5 => ['Cloaked redirect', 'R', 'E'],
  6 => ['Internal redirect to site', 'R', 'I'],
  7 => ['Files & Dir listing with upload', 'F'],
  100 => ['Photo gallery - light', 'F'],
  101 => ['Photo gallery - dark', 'F'],
);

our @DisplayModes = (0, 1, 7, 100, 101, 4, 5, 6);

# }}}

use Socket;
use ME::User;
use ME::VFS::Cache;
use ME::VFS::Node;
use ME::VFS::IOHandle;
use Archive::Zip qw( :ERROR_CODES );
use ME::Zip;
use ME::DB;
use File::Temp;
use File::Copy;
use Data::UUID;
use MIME::Base64;
use Digest::SHA;
use Encode;
use Scalar::Util qw( weaken );
use ME::RateTrack;

our $UUIDGEN = Data::UUID->new();

use constant VFS_NAMESPACE => 'ME:';

sub new {
  my $class = shift;
  my $UserArg = shift;

  # figure out who the user is
  my $RefPassed = ref $UserArg;
  my $User;
  if ($RefPassed) {
    $User = $UserArg;
  }
  elsif ($UserArg =~ m/^\d+$/) {
    $User = ME::User->new($UserArg);
  }
  else {
    $User = ME::User->new_unqname($UserArg);
  }

  my $Self = bless {
    User => $User,
    Nodes => {},
  }, ref($class) || $class;

  # Weaken the User ref if it was passed in, so we don't prevent GC
  # If we created the user, then don't make it a weak ref.
  weaken $Self->{User} if $RefPassed;

  ME::VFS::Cache::Init();

  my $VFSDbh = ME::VFS::Cache::VFSDbh();

  my ($id) = $VFSDbh->selectrow_array(<<EOF, {}, $User->UserId());
SELECT NodeId 
  FROM Nodes 
 WHERE IsDeleted = 0
   AND ParentNodeId = 1 
   AND UserId = ?
EOF

  if ($id) {
    my $nd = ME::VFS::Cache::Get('ND', $id);
    $Self->{BaseDirId} = $id;
    $Self->{HomeDirId} = $nd->{0}{files};
    $Self->{TempDirId} = $nd->{0}{temp};
  }

  # XXX - just for copying stuff across!

  return $Self;
}

sub Init {
  my $Self = shift;
  $Self->CreateVFS(); # ensure we have a home dir!
}

sub Node {
  my $Self = shift;
  my $NodeId = shift;
  my %Opts = @_;

  my $Node = ME::VFS::Node->new($Self, $NodeId);

  weaken $Node->{VFS}; # make it cleanupable

  # Enforce ACLs more often
  unless ($Opts{IgnoreACLs}) {
    return unless $Node->CanRead();
  }

  return $Node;
}

sub newByToken {
  my $Token = shift;

  return unless $Token =~ m/^v1-([a-z0-9]+)-(\d+)-/;
  my $UnqId = $1;
  my $NodeId = $2;

  my $UserId = ME::User::UnqIdToUserId($UnqId);
  return unless $UserId;
  my $User = ME::User->new($UserId);
  return unless $User;

  my $VFS = $User->VFS();
  return unless $VFS; # don't init
  my $Node = $VFS->Node($NodeId);
  return unless $Node;
  return if $Node->IsDeleted();
  return unless $Node->IsFile();

  # make sure the token is valid
  my $RealToken = $Node->FileToken('v1');
  return unless $RealToken eq $Token;

  return ($User, $VFS, $Node);
}

sub User {
  my $Self = shift;
  return $Self->{User};
}

sub UserId {
  my $Self = shift;
  use Carp;
  Carp::confess unless $Self->User;
  return $Self->User->UserId();
}

sub SysPropertyNamespace {
  my $Self = shift;
  return VFS_NAMESPACE;
}

sub UserPropertyNamespace {
  my $Self = shift;
  return VFS_NAMESPACE . $Self->UserId();
}

sub MakeUUID {
  my $Self = shift;

  return $UUIDGEN->create_str();
}

sub VFSExists {
  my $Self = shift;
  return $Self->{BaseDirId} ? 1 : 0;
}

=item I<DeleteEntireVFS()>

Removes everything owned by the user!

=cut
sub DeleteEntireVFS {
  my $Self = shift;

  # If no file storage for this user, just return silently.
  return unless $Self->VFSExists();

  # tidy up the filesystem as much as possible.
  my $Node = $Self->BaseDir();
  $Node->Delete(IgnoreACLs => 1) if $Node;

  # clean up web access
  my $WebIds = ME::VFS::Cache::Get('UW', $Self->UserId());
  foreach my $Id (map { values %{$_} } values %{$WebIds}) {
    ME::VFS::Cache::Delete('W', $Id);
  }

  ME::VFS::Cache::Commit();
}

sub BaseName {
  my $Self = shift;
  my $User = shift || $Self->User();

  # I'm pretty sure this never creates 'dodgy' names
  my $BaseName = $User->UnqName();
  $BaseName =~ s/@/./;

  return $BaseName;
}
sub BaseDir {
  my $Self = shift;

  unless (exists $Self->{BaseDir}) {
    unless ($Self->{BaseDirId}) {
      $Self->CreateVFS();
    }
    $Self->{BaseDir} = $Self->Node($Self->{BaseDirId}, IgnoreACLs => 1);
  }

  return $Self->{BaseDir};
}

=item I<PostRename()>

Run this after renaming the user.  If the home directory has changed, update it

=cut
sub PostRename {
  my $Self = shift;

  # check that we have a VFS
  return 0 unless $Self->{BaseDirId};

  # Avoid ACLs with direct access
  my $BaseName = $Self->BaseName();
  my $nd = ME::VFS::Cache::Get('N', $Self->{BaseDirId});
  
  # check that it has changed
  return 1 if $nd->{NodeName} eq $BaseName;

  # there'd better not be anyone else with this nodename, or
  # the database constraints on the Users table have already failed
  ME::VFS::Cache::Update('N', $Self->{BaseDirId}, {NodeName => $BaseName});
  ME::VFS::Cache::Commit();

  # done - changed it!
  return 2; 
}

=item I<HomeDir()>

Finds the user's Home Directory node if it exists.

=cut
sub HomeDir {
  my $Self = shift;

  unless (exists $Self->{HomeDir}) {
    unless ($Self->{HomeDirId}) {
      return $Self->CreateVFS();
    }
    $Self->{HomeDir} = $Self->Node($Self->{HomeDirId}, IgnoreACLs => 1);
  }

  return $Self->{HomeDir};
}

=item I<TempDir()>

Finds the user's Temporary directory and creates it if necessary

=cut
sub TempDir {
  my $Self = shift;

  unless (exists $Self->{TempDir}) {
    unless ($Self->{TempDirId}) {
      return $Self->CreateTempDir();
    }
    $Self->{TempDir} = $Self->Node($Self->{TempDirId}, IgnoreACLs => 1);
  }

  return $Self->{TempDir};
}

sub Chroot {
  my $Self = shift;
  my $Node = shift;
  $Self->{Chroot} = $Node;
}

sub RootNode {
  my $Self = shift;
  return $Self->{Chroot} if $Self->{Chroot};
  return $Self->Node(1, IgnoreACLs => 1);
}

# No longer cached because otherwise deep Delete fails.  It's not that
# bad though, all the underlying data is cached.
sub PathToNode {
  my $Self = shift;
  my $Path = shift || '/';

  # be a "Node" interface too for extra funkies
  if ($Path =~ m/^\d+$/) {
    return $Self->Node($Path);
  }

  return $Self->RootNode() if $Path eq '/';

  die "Need full path name\n" unless $Path =~ s:^/::;

  return $Self->RootNode()->PathToNode($Path);
}

=item I<CreateVFS()>

Creates the file storage directory for this user, and
the quota file and the /private and /public directories.

=cut
sub CreateVFS {
  my $Self = shift;
  my %Opts = @_;
  if ($Self->{HomeDirId}) {
    return $Self->HomeDir();
  }

  my $User = $Self->User();
  my $UserId = $User->UserId();

  my $TopLevelName = $Self->BaseName($User);

  # NOTE: THIS CODE PATH USES DIRECT ME::VFS::Cache ACCESS
  # TO AVOID ACL CHECKS BECAUSE IT IS BOOTSTRAPPING

  my $time = time();

  # Create top level dir with appropriate permissions
  my $TopLevelNodeId = ME::VFS::Cache::Create('N', {
    UserId => $UserId,
    ParentNodeId => 1,
    NodeName => $TopLevelName,
    NodeType => 'd',
    Created => $time,
    Modified => $time,
    IsDeleted => 0,
  });
  die "Unable to create /$TopLevelName" unless $TopLevelNodeId;
  $Self->{BaseDirId} = $TopLevelNodeId;

  # Create files dir with appropriate permissions
  my $FilesNodeId = ME::VFS::Cache::Create('N', {
    UserId => $UserId,
    ParentNodeId => $TopLevelNodeId,
    NodeName => 'files',
    NodeType => 'd',
    Created => $time,
    Modified => $time,
    IsDeleted => 0,
  });
  die "Unable to create /$TopLevelName/files" unless $FilesNodeId;
  $Self->{HomeDirId} = $FilesNodeId;

  # Create temp dir with appropriate permissions
  my $TempNodeId = ME::VFS::Cache::Create('N', {
    UserId => $UserId,
    ParentNodeId => $TopLevelNodeId,
    NodeName => 'temp',
    NodeType => 'd',
    Created => $time,
    Modified => $time,
    IsDeleted => 0,
  });
  die "Couldn't create Temp Directory" unless $TempNodeId;
  $Self->{TempDirId} = $TempNodeId;

  ME::VFS::Cache::Create('NQ', {
    NodeId => $FilesNodeId,
    Used => 0,
    Total => $Self->CalculateQuotaTotal(),
  });

  my $HomeDir = $Self->Node($FilesNodeId);
  $Self->{HomeDir} = $HomeDir;

  # Default to expanded for their home directory.
  $HomeDir->SetUserProperty('IsExpanded', 1, SkipCommit => 1);

  # enable if we want to autocreate a public directory for the user
  #$HomeDir->Mkdir('public', SkipCommit => 1);

  # Create a photo album
  my $Album = $HomeDir->Mkdir('album', SkipCommit => 1);
  $Album->SetUserProperty('ViewMode', 'photos');

  if ($User->Business) {
    $User->SetPrivacy($User->{IsPrivate}); # refresh ACLs for admin
  }

  # Remove our cached memories of contents and zap
  ME::VFS::Cache::Commit() unless $Opts{SkipCommit};

  return $HomeDir;
}

=item I<CreateTempDir()>

Creates the file storage temporary directory for this user:
NOTE: called automatically from TempDir, use that instead!

NOTE2: deprecated, will be replaced by the time everyone has a TempDir

=cut
sub CreateTempDir {
  my $Self = shift;
  my %Opts = @_;

  my $Parent = $Self->BaseDir();

  my $nd = ME::VFS::Cache::Get('ND', $Parent->NodeId());
  if ($nd->{0}{temp}) {
    $Self->{TempDirId} = $nd->{temp}{0};
    return $Self->Node($nd->{temp}{0});
  }

  my $time = time();
  my $UserId = $Self->UserId();
  my $ParentNodeId = $Parent->NodeId();

  # Create temp dir with appropriate permissions
  my $TempNodeId = ME::VFS::Cache::Create('N', {
    UserId => $UserId,
    ParentNodeId => $ParentNodeId,
    NodeName => 'temp',
    NodeType => 'd',
    Created => $time,
    Modified => $time,
    IsDeleted => 0,
    TempExpires => undef,
  });
  die "Couldn't create Temp Directory" unless $TempNodeId;
  $Self->{TempDirId} = $TempNodeId;

  ME::VFS::Cache::Commit() unless $Opts{SkipCommit};
 
  return $Self->Node($TempNodeId);
}

# Quota Commands

=item I<RateUpdate( $SizeInBytes, $FileName, %Opts )>

Updates bandwidth usage via RateTrack

=cut
sub RateUpdate {
  my ($Self, $FileSize, $FileName, %Opts) = @_;

  my $ForceFlag = $Opts{ForceFlag};
  my $Source = $Opts{Source};

  my $UnqName = $Opts{UnqName} || $Self->User->UnqName();

  my $RateTrack = $Self->get_ratetrack;

  my $Result = $RateTrack->log_file($UnqName, $FileName, $FileSize, $ForceFlag);

  my $ReturnCode = 0;

  if ($Result =~ /^2/){
    $ReturnCode = 1;
  }
  elsif ($Result =~ /^5/){
    warn "ME:VFS - Unexpected result from RateTrack - $Result";
    $ReturnCode = 1;
  } else {
    $ReturnCode = 0;
  }
  
  return wantarray ? ($ReturnCode, $Result) : $ReturnCode;
}

=item I<RateCheck( $SizeInBytes, $FileName, %Opts )>

Checks to see how a ratetrack transaction would work out without actually
updating any usage records.

=cut
sub RateCheck {
  my ($Self, $FileSize, $FileName, %Opts) = @_;

  my $RateTrack = $Self->get_ratetrack;

  my $UnqName = $Opts{UnqName} || $Self->User->UnqName();

  my @Opts;
  push @Opts, (IP => $Opts{IP}) if $Opts{IP};
  push @Opts, (ThrottleMode => $Opts{ThrottleMode}) if $Opts{ThrottleMode};
  push @Opts, (WebAccessId => $Opts{WebAccessId}) if $Opts{WebAccessId};
  my $Result = $RateTrack->log_file_test($UnqName, $FileName, $FileSize, @Opts);

  my $ReturnCode = 0;

  if ($Result =~ /^2/){
    $ReturnCode = 1;
  }
  elsif ($Result =~ /^5/){
    warn "ME:VFS - Unexpected result from RateTrack - $Result";
    $ReturnCode = 1;
  } else {
    $ReturnCode = 0;
  }

  return wantarray ? ($ReturnCode, $Result) : $ReturnCode;
}

sub GetQuotaUsed {
  my $Self = shift;

  my $HomeDir = $Self->HomeDir();
  return unless $HomeDir;

  return $HomeDir->GetQuotaUsed();
}

sub GetQuota {
  my $Self = shift;

  my $HomeDir = $Self->HomeDir();
  return unless $HomeDir;

  return $HomeDir->GetQuota();
}

sub SetQuotaTotal {
  my $Self = shift;
  my $Total = shift;
  my %Opts = @_;

  my $HomeDir = $Self->HomeDir();

  my $Used = $HomeDir->GetQuotaUsed();
  if ($Total < $Used) {
    warn "setting quota $Total below currently used amount $Used\n";
  }

  # NOTE: this looks like a race condition, but it actually isn't because we
  # die before committing if the same key has been updated in another session!
  # We need to do all keys here because it might actually make a new record!
  return $HomeDir->SetQuotaTotal($Total, %Opts, SkipQuotaCheck => 1);
}

sub FixQuota {
  my $Self = shift;
  my %Opts = @_;

  my $VFSDbh = ME::VFS::Cache::VFSDbh();

  my $HomeDir = $Self->HomeDir();
  my ($Used, $Total) = $HomeDir->GetQuota();

  my $NewUsed = $HomeDir->ContentSize(IncludeSubDirs => 1);

  my $User = $Self->User();
  my $ServiceLevelObj = $User->ServiceLevel();
  my $NewTotal = ($ServiceLevelObj->{FileStorageQuota} * 1024) +
                 ($User->{ExtraFileQuota} * 1024 * 1024);

  if ($Used != $NewUsed or $Total != $NewTotal) {
    warn "Quota mismatch: fixing ($Used/$Total) => ($NewUsed/$NewTotal) (U: " . $Self->UserId() . ")\n";
    $HomeDir->SetQuota($NewUsed, $NewTotal, SkipCommit => 1, SkipQuotaCheck => 1);
  }

  ME::VFS::Cache::Commit() unless $Opts{SkipCommit};

  return ($NewUsed, $NewTotal);
}

sub CalculateQuotaTotal {
  my $Self = shift;
  my $User = shift || $Self->User();

  my $ServiceLevelObj = $User->ServiceLevel();
  my $FileQuota = ($ServiceLevelObj->{FileStorageQuota} * 1024) +
                  ($User->{ExtraFileQuota} * 1024 * 1024);

  return $FileQuota;
}

# Information methods

=item I<MaxFileSize()

Returns the maximum size in bytes for any single file in the filestorage.

=cut
sub MaxFileSize {
  my $Self = shift;

  return $Self->ServiceLevel(@_)->{MaxFileSize} || (10 * 1024 * 1024);
}

sub ServiceLevel {
  my $Self = shift;

  return $Self->User()->ServiceLevel();
}

sub Mkdir {
  my $Self = shift;
  my $Path = shift;
  my %Opts = @_;

  my ($parent, $this) = $Self->ParentAndFile($Path);

  my $ParentNode = $Self->PathToNode($parent, %Opts);
  if ($Opts{Deep} and not $ParentNode) {
    $ParentNode = $Self->Mkdir($parent, %Opts, SkipCommit => 1);
  }
  die "No such dir $parent" unless $ParentNode;

  my $ThisNode = $ParentNode->Mkdir($this,
    %Opts,
  );

  return $ThisNode;
}

sub DirectoryTree {
  my $Self = shift;
  my %Opts = @_;
  my $Node = $Opts{BaseNode} || $Self->RootNode();

  # XXX: Cache!

  return ()
    if ($Opts{IgnoreTemp} and $Node->Depth() == 2 and $Node->NodeName() eq 'temp');

  my @SubDirs = $Node->SubDirs();

  if ($Opts{HomeFirst} and $Node->IsRootNode()) {
    my $HomeDir = $Self->HomeDir();
    my $ParentNode = $HomeDir->ParentNode();
    @SubDirs = ($ParentNode, grep { not $_->Equals($ParentNode) } @SubDirs);
  }

  return ($Node)
    if ($Opts{OnlyExpanded} and $Node->Depth() >= 2 and not $Node->GetUserProperty('IsExpanded'));

  return (
    $Node,
    map { $Self->DirectoryTree(%Opts, BaseNode => $_) } @SubDirs,
  );
}

# Tool to make simple easy-to-read directory selector!

sub RecentDirs {
  my $Self = shift;
  my %Opts = @_;

  my $RecentText = $Self->HomeDir()->GetUserProperty('RecentDirs');
  my @NodeIds = split(',', $RecentText || '');

  # Catch if a dir has been deleted, or not a dir
  my @Nodes = grep { !$_->IsDeleted() && $_->IsDir() }
              map { eval { $Self->Node($_) } or () } @NodeIds;

  # If so, remove from recent nodeid list
  if (scalar(@NodeIds) != scalar(@Nodes)) {
    my $NewText = join(',', map { $_->NodeId() } @Nodes);
    $Self->HomeDir()->SetUserProperty('RecentDirs', $NewText, %Opts);
  }

  if ($Opts{OnlyWritable}) {
    @Nodes = grep { $_->CanWrite() } @Nodes;
  }

  return @Nodes;
}

sub RecentSelect {
  my $Self = shift;
  my %Opts = @_;

  my @Dirs = ($Self->HomeDir(),
              sort { $a->Path() cmp $b->Path() } $Self->RecentDirs(%Opts));

  my @Keys = map { $_->NodeId() } @Dirs;
  my %Vals = map { $_->NodeId() => $_->ShortPath() } @Dirs;

  push @Keys, -1;
  $Vals{-1} = "More ...";

  return (\@Keys, \%Vals);
}

sub RecentSet {
  my $Self = shift;
  my $AddNode = shift;
  my %Opts = @_;

  return unless $AddNode;
  return if $AddNode->Equals($Self->HomeDir()); # always in the list..
  return if !$AddNode->IsDir(); # Only dirs in recent list

  my @Nodes = $Self->RecentDirs(%Opts);

  unless (@Nodes and $Nodes[0]->Equals($AddNode)) {
    # need to update property...
    @Nodes = ($AddNode, grep { not $_->Equals($AddNode) } @Nodes);
    $#Nodes = 9 if $#Nodes > 9;
    my $NewText = join(',', map { $_->NodeId() } @Nodes);
    $Self->HomeDir()->SetUserProperty('RecentDirs', $NewText, %Opts);
  }
}

sub DirectorySelect {
  my $Self = shift;
  my %Opts = @_;

  # XXX: implement on top of DirectoryTree, use cache!

  my @Keys;
  my %Names;

  # don't both if they don't have a VFS!
  if ($Self->VFSExists()) {
    my $Node = $Opts{Base} || $Self->RootNode();
    $Self->_DirectorySelect($Node, \@Keys, \%Names, 0, %Opts);
  }

  return(\@Keys, \%Names);
}

sub _DirectorySelect {
  my $Self = shift;
  my $Node = shift;
  my $Keys = shift;
  my $Names = shift;
  my $Depth = shift;
  my %Opts = @_;

  return
    if ($Opts{SkipPrivate} and $Node->Depth() == 3 and $Node->NodeName() eq 'private');

  return
    if ($Opts{IgnoreTemp} and $Node->Depth() == 2 and $Node->NodeName() eq 'temp');

  my $dothis;
  if ($Node->Depth() == 2) {
    $dothis = 1; # always display level 2 nodes (home directories)..
  } elsif (not $Node->CanRead()) {
    $dothis = 1 if $Opts{IncludeUnreadable};
  } elsif ($Opts{OnlyWritable}) {
    $dothis = 1 if $Node->CanWrite();
  } else {
    $dothis = 1; # Readable :)
  }

  # show this and go mark more depth..
  if ($dothis) {
    push @$Keys, $Node->NodeId();
    $Names->{$Node->NodeId()} = (". " x $Depth) . $Node->ShortName();
    ++$Depth;
  }

  # and recurse
  if ($Node->IsRootNode()) {
    my $PNode = $Self->HomeDir->ParentNode();
    $Self->_DirectorySelect($PNode, $Keys, ,$Names, $Depth, %Opts);
    foreach my $SubDir ($Node->SubDirs()) {
      next if $SubDir->NodeId() == $PNode->NodeId();
      $Self->_DirectorySelect($SubDir, $Keys, $Names, $Depth, %Opts);
    }
  } else {
    foreach my $SubDir ($Node->SubDirs()) {
      $Self->_DirectorySelect($SubDir, $Keys, $Names, $Depth, %Opts);
    }
  }
}

sub OpenFile {
  my $Self = shift;
  my $Path = shift;
  my $Mode = shift;
  my %Opts = @_;
  my ($parent, $this) = $Self->ParentAndFile($Path);
  my $ParentNode = $Self->PathToNode($parent, %Opts);
  die "No such dir $parent" unless $ParentNode and $ParentNode->IsDir();

  return $ParentNode->OpenFile($this, $Mode, @_);
}

sub OpenTempFile {
  my $Self = shift;
  my $Name = shift;
  my $Mode = shift;

  my $TempDir = $Self->TempDir();

  return $TempDir->OpenTempFile($Name, $Mode, @_);
}

sub ImportFile {
  my $Self = shift;
  my $File = shift;
  my $Path = shift;
  my %Opts = @_;

  my ($parent, $this) = $Self->ParentAndFile($Path);

  my $ParentNode = $Self->PathToNode($parent, %Opts);

  die "No such directory $parent" unless $ParentNode;

  my $Node = $ParentNode->ImportFile($File, $this,
    %Opts,
  );

  return $Node;
}

sub ImportTempFile {
  my $Self = shift;
  my $File = shift;
  my $Name = shift;
  my %Opts = @_;

  my $TempDir = $Self->TempDir();

  my $Node = $TempDir->ImportTempFile($File, $Name, %Opts);

  return $Node;
}

sub _GenerateName {
  my $Name = shift;
  my $Num = shift;

  if ($Name =~ m/^(.+)\.(.+)$/) {
    return "$1\[$Num\].$2";
  }
  else {
    return "$Name\[$Num\]";
  }
}

sub Delete {
  my $Self = shift;
  my $Path = shift;

  my $Node = $Self->PathToNode($Path);

  die "No such file $Path" unless $Node;

  $Node->Delete();
}

sub Copy {
  my $Self = shift;
  my $Source = shift;
  my $Path = shift;
  my %Opts = @_;

  my $Node = $Self->PathToNode($Source);
  die "No such file or directory $Source" unless $Node;

  my ($parent, $this) = $Self->ParentAndFile($Path);

  my $ParentNode = $Self->PathToNode($parent, %Opts);
  die "No parent directory $parent" unless ($ParentNode and $ParentNode->IsDir());

  $Node->CopyTo(%Opts,
    ParentNode => $ParentNode,
    NodeName => $this
  );
}

sub Move {
  my $Self = shift;
  my $Source = shift;
  my $Path = shift;
  my %Opts = @_;

  my $Node = $Self->PathToNode($Source);
  die "No such file or directory $Source" unless $Node;

  my ($parent, $this) = $Self->ParentAndFile($Path);

  my $ParentNode = $Self->PathToNode($parent, %Opts);
  die "No parent directory $parent" unless ($ParentNode and $ParentNode->IsDir());

  $Node->MoveTo(%Opts,
    ParentNode => $ParentNode,
    NodeName => $this
  );
}

sub GetFileList {
  my $Self = shift;
  my $Path = shift;

  my $Node = $Self->PathToNode($Path);

  die "No such directory $Path" unless ($Node and $Node->IsDir());
  die "No permission to read $Path" unless $Node->CanRead();

  return $Node->Files();
}

sub GetSubDirs {
  my $Self = shift;
  my $Path = shift;

  my $Node = $Self->PathToNode($Path);

  die "No such directory $Path" unless ($Node and $Node->IsDir());
  die "No permission to read $Path" unless $Node->CanRead();

  return $Node->SubDirs();
}

sub CreateZipFile {
  my $Self = shift;

  $Self->{TZDelta} = $Self->User()->TZSeconds();

  my $ZipFile = Archive::Zip->new();
  foreach my $Node (@_) {
    $Self->_AddToZipFile($ZipFile, $Node, "");
  }
  my $TempFh = File::Temp::tempfile();
  $ZipFile->writeToFileHandle($TempFh);
  seek($TempFh, 0, 0);
  return $TempFh;
}

sub _AddToZipFile {
  my $Self = shift;
  my $ZipFile = shift;
  my $Node = shift;
  my $BasePath = shift;

  return $ZipFile unless $Node->Exists();

  my $Name = $Node->NodeName();
  if ($Node->IsDir()) {
    foreach my $SubDir ($Node->SubDirs()) {
      $Self->_AddToZipFile($ZipFile, $SubDir, $BasePath . $Name . "/");
    }
    foreach my $File ($Node->Files()) {
      $Self->_AddToZipFile($ZipFile, $File, $BasePath . $Name . "/");
    }
  } else {
    my $Member = ME::Zip::AddUnicodeFile($ZipFile, $Node->DiskPath(1), $BasePath . $Name);
    ME::Zip::SetMemberTime($Member, $Node->Modified(), $Self->{TZDelta});
  }

  return $ZipFile;
}

sub ExtractZipFile {
  my $Self = shift;
  my $Target = shift;
  my %Opts = @_;

  my $Arc = Archive::Zip->new();

  my $TmpDir = $ENV{TMPDIR} || $ME::VFS::IOHandle::VFSTmpDir;
  my $tempdir = File::Temp::tempdir(DIR => $TmpDir, CLEANUP => 1);

  if (my $File = delete $Opts{File}) {
    die "failed to read as a zip file: $::GlobalZipError\n" unless $Arc->read($File) == AZ_OK;
  } 
  elsif (my $Fh = delete $Opts{Fh}) {
    File::Copy::copy( $Fh, "$tempdir/file.zip");
    die "failed to read as a zip file: $::GlobalZipError\n" unless $Arc->read("$tempdir/file.zip") == AZ_OK;
  }
  else {
    die "Need a File or Fh\n";
  }

  if (my $Rename = delete $Opts{Rename}) {
    $Target = $Target->ChildByName($Rename) || $Target->Mkdir($Rename, SkipCommit => 1);
  }

  my @Members = $Arc->members();

  # give ourselves enough time...
  my $OldAlarm = alarm(120 + int(@Members));
  eval {
    foreach my $Member (@Members) {
      next if $Member->isDirectory();
      my $DirNode = $Target;
      my $MemberName = $Member->fileName();
      $MemberName = $Self->FixString($MemberName) if $MemberName =~ /[^\x00-\x7f]/;
      while ($MemberName =~ s{(.*?)/}{}) {
        my $Name = $1;
        $DirNode = $DirNode->ChildByName($Name) || $DirNode->Mkdir($Name, %Opts, SkipCommit => 1);
      }
      my $TempFile = "$tempdir/tempfile";
      { # make taint safe
        local %ENV = ();
        $Member->extractToFileNamed($TempFile);
      }

      # only include real files
      # zip files can inclue symlinks to arbitrary system files, nasty
      if (-f $TempFile && !-l $TempFile) {
        my $NewNode = $DirNode->ImportFile($TempFile, $MemberName,
          %Opts,
          SkipCommit => 1,
        );
      }

      # clean up
      unlink($TempFile);
    }
  };
  my $Error = $@;

  alarm($OldAlarm);
  die $Error if $Error;

  ME::VFS::Cache::Commit() unless $Opts{SkipCommit};
}

sub ParentAndFile {
  my $Self = shift;
  my $Path = shift;
  $Path =~ s{//+}{/}g; # strip multiple slashes
  $Path =~ s{/$}{}; # strip trailing slash

  my @bits = split('/', $Path); # cheap and nasty!
  my $name = pop @bits;
  my $parent = join('/', @bits);
  $parent .= '/'; # always trailing slash the a directory

  return ($parent, $name);
}

sub ParentNodeAndName {
  my $Self = shift;
  my $Path = shift;
  my %Opts = @_;

  my ($parent, $name) = $Self->ParentAndFile($Path);

  my $ParentNode = $Self->PathToNode($parent, %Opts);

  die "No such directory $parent\n"
    if !($ParentNode && $ParentNode->Exists() && $ParentNode->IsDir());

  return ($ParentNode, $name);
}

sub GetWebAccess {
  my $Self = shift;
  my $WebAccessId = shift;

  if ($WebAccessId) {
    my $WA = ME::VFS::Cache::Get('W', $WebAccessId);
    return undef unless $WA->{UserId} eq $Self->UserId(); # no peeking!
    return $WA;
  }

  my $Ids = ME::VFS::Cache::Get('UW', $Self->UserId());

  my @response;
  foreach my $Host (keys %$Ids) {
    push @response, map { ME::VFS::Cache::Get('W', $_) } values %{$Ids->{$Host}};
  }
  return wantarray ? @response : \@response;
}

sub GetWebAccessForHost {
  my $Self = shift;
  my $HostName = shift;
  my $All = shift;
  my $UserId = $Self->UserId(); # avoid peeking again!

  my $Ids = ME::VFS::Cache::Get('HW', $HostName);

  return grep { $All or $_->{UserId} eq $UserId } map { ME::VFS::Cache::Get('W', $_) } values %$Ids;
}

sub AddWebAccess {
  my $Self = shift;
  my $Data = shift;
  my %Opts = @_;

  # make sure it doesn't clash
  die "A website already exists at http://$Data->{HostName}$Data->{BasePath}\n" if
    grep { $_->{BasePath} eq $Data->{BasePath} } $Self->GetWebAccessForHost($Data->{HostName}, 'ALL');

  $Data->{HostName} = lc($Data->{HostName});
  $Data->{UserId} = $Self->UserId(); # enforce :)
  $Data->{NodeId} ||= undef; # not zero!

  my $CreateOpts = {};
  my $DisplayMode = $Data->{DisplayMode};
  if ($DisplayMode == ME::VFS::DisplayModeRedirect
      || $DisplayMode == ME::VFS::DisplayModeCloaked
      || $DisplayMode == ME::VFS::DisplayModeRedirectInternal
      || !$Data->{NodeId}) {
    $CreateOpts->{IgnoreForeign} = 1;
  }
  # Create it
  ME::VFS::Cache::Create('W', $Data, $CreateOpts);

  ME::VFS::Cache::Commit() unless $Opts{SkipCommit};
}

sub DeleteWebAccess {
  my $Self = shift;
  my $WebAccessId = shift;
  my %Opts = @_;

  # no deleting other people's!
  my $WA = ME::VFS::Cache::Get('W', $WebAccessId);
  die "Not your web access $WebAccessId" if $WA && $WA->{UserId} ne $Self->UserId();

  # Delete it
  ME::VFS::Cache::Delete('W', $WebAccessId);

  ME::VFS::Cache::Commit() unless $Opts{SkipCommit};
}

sub ChangeWebAccess {
  my $Self = shift;
  my $WebAccessId = shift;
  my $Data = shift;
  my %Opts = @_;

  my $OldData = ME::VFS::Cache::Get('W', $WebAccessId);
  die "Web access doesn't exist $WebAccessId" unless $OldData;
  die "Not your web access $WebAccessId" if $OldData->{UserId} ne $Self->UserId();

  # NOTE: can't change user or host, so no need to invalidate keys...
  my %Change;
  foreach my $key (qw(DisplayMode Password NodeId RedirectTarget RedirectTitle IsActive ThrottleMode)) {
    $Change{$key} = $Data->{$key} if exists $Data->{$key};
  }
  $Change{NodeId} ||= undef if exists $Change{NodeId}; # not zero!

  ME::VFS::Cache::Update('W', $WebAccessId, \%Change);

  # no need to change caching, enforced limited things you can do!

  ME::VFS::Cache::Commit() unless $Opts{SkipCommit};
}

# these functions allow us to put flag values in the MMAP cache

sub UserFlagGet {
  my $Self = shift;
  my $Flag = shift;

  return $Self->FlagGet($Self->UserId() . ":" . $Flag);
}

sub UserFlagSet {
  my $Self = shift;
  my $Flag = shift;
  my $Value = shift;

  return $Self->FlagSet($Self->UserId() . ":" . $Flag, $Value);
}

sub UserFlagDelete {
  my $Self = shift;
  my $Flag = shift;

  return $Self->FlagDelete($Self->UserId() . ":" . $Flag);
}

sub FlagGet {
  my $Self = shift;
  my $Flag = shift;

  my $Res = $ME::VFS::Cache::MMAPCache->get("_flag:$Flag");

  return undef unless $Res;
  return $$Res;
}

sub FlagSet {
  my $Self = shift;
  my $Flag = shift;
  my $Value = shift;

  return $ME::VFS::Cache::MMAPCache->set("_flag:$Flag", \$Value);
}

sub FlagDelete {
  my $Self = shift;
  my $Flag = shift;

  return $ME::VFS::Cache::MMAPCache->remove("_flag:$Flag");
}

# UTF8 encoding helpers

sub FixUriString {
  my $Self = shift;
  my $Path = shift;
  my $ClientQuirks = shift; # optional quirks handling...

  # trivial case
  return '/' unless defined $Path;
  
  # If any high-byte characters, assume in client encoding!
  # XXX - check that the client is a known-naughty one from $ClientQuirks
  if ($Path =~ m{ [\x80-\xFF] }xms) {
    my $Charset = $Self->User->{DefCharSet};
    return decode($Charset, $Path, Encode::FB_CROAK);
  }

  # otherwise, convert % encoded characters - see URI::Escape manpage 
  $Path =~ s{ % ([0-9A-Fa-f]{2}) }{ chr(hex($1)) }xmseg;

  return $Self->FixString($Path);
}

sub FixString {
  my $Self = shift;
  my $String = shift;

  # already fixed?
  if (Encode::is_utf8($String)) {
    Carp::cluck("Bogus already Perl-UTF8 encoded string found");
    return $String;
  }

  # Attempt to decode with UTF8 first
  if (defined(my $perlstring
       = eval { decode('utf8', $String, Encode::FB_CROAK) })) {
    return $perlstring;
  }

  # Then the user's charset
  my $Charset = $Self->User->{DefCharSet};
  if (defined(my $perlstring 
       = eval { decode($Charset, $String, Encode::FB_CROAK) })) {
    return $perlstring;
  }

  # finally try iso-8859-1
  if (defined(my $perlstring 
       = eval { decode('iso-8859-1', $String, Encode::FB_CROAK)
 })) {
    warn "DECODED ISO $perlstring";
    return $perlstring;
  }

  warn "Octets = " . join(' ', map { ord($_) } split (//, $String)) . "\n";
  die "Unable to decode string $String!"; # should never happen
  
}

sub FixWebAccess {
  my $Self = shift;
  my %Opts = @_;

  # get values
  my $Domains = $Self->GetWebDomains();
  my $WebHosts = ME::VFS::Cache::Get('UW', $Self->UserId());

  my @DeleteIds;

  # not the most efficient, but we don't care here
  HOST:
  foreach my $HostName (keys %{$WebHosts}) {
    foreach my $Domain (@{$Domains}) {
      next HOST if $HostName eq $Domain;
      next HOST if $HostName =~ m{ [.] $Domain \z}xms;
    }

    # fallthrough, didn't match any domain
    push @DeleteIds, values %{$WebHosts->{$HostName}};
  }

  foreach my $Id (@DeleteIds) {
    my $Item = ME::VFS::Cache::Get('W', $Id);

    my $UnqName = $Self->User()->UnqName();
    my $Msg = "Removing no longer allowed website for $UnqName: http://$Item->{HostName}$Item->{BasePath}\n";
    if ($::action_log_fh) {
      print $::action_log_fh $Msg;
    } else {
      warn $Msg;
    }

    $Self->DeleteWebAccess($Id, SkipCommit => 1);
  }

  ME::VFS::Cache::Commit() unless $Opts{SkipCommit};
}

sub GetWebDomains {
  my $Self = shift;

  my $Dbh = ME::DB->GetDbh( Name => 'main' );

  my $UnqName = $Self->User->UnqName(Separator => '.');
  $UnqName =~ s/@/./;

  # XXX - should we check d.BrandId too?
  my $Aliases = $Dbh->selectall_arrayref(<<EOF, {}, $Self->UserId);
  SELECT a.Name, d.Domain
    FROM Aliases a, Domains d
   WHERE d.DomainId = a.DomainId
     AND a.UserId = ?
     AND a.Active = 1
     AND d.UserId = 0
EOF

  # Get domains (any state is fine, it'll only work if the DNS points to us anyway)
  my $OwnDomains = $Dbh->selectall_arrayref(<<EOF, {}, $Self->UserId);
  SELECT d.Domain
    FROM Domains d
   WHERE d.UserId = ?
EOF

  my @Domains = ($UnqName,
     (map { $_->[0] . '.' . $_->[1] } grep { $_->[0] =~ m/^[a-zA-Z0-9_-]+$/ } @$Aliases), # only valid hostname chars...
     (map { $_->[0] } @$OwnDomains),
  );

  return wantarray ? @Domains : \@Domains;
}

# XXX: we need to support this one
sub IsReadOnly {
  my $Self = shift;
  ME::VFS::Cache::IsReadOnly($Self->UserId());
}

sub PrepareForFreeze {
  ME::VFS::Cache::Finish();
}

sub Cleanup {
  ME::VFS::Cache::Finish();
}

sub DESTROY {
  ME::VFS::Cache::Finish();
}

sub get_ratetrack {
  my $Self = shift;

  my $RateTrack = $Self->{'RateTrack'};
  if (!$RateTrack) {
    $RateTrack = $Self->{RateTrack} = ME::RateTrack->new(
      'vfs', 1
    );
  }

  return $RateTrack;
}

1;
