package ME::VFS::Cache;

=pod

=head1 NAME

ME::VFS::Cache - The caching and database management layer for
ME::VFS.  Designed to be used only within the VFS code, but to
handle all database requests in a transaction safe manner.

Most of the caching has been ripped out for now, instead it
pre-fetches interesting data.

=head1 DESCRIPTION

Cache values are named according to the their Cat (Category)
and Item, e.g. the Node with NodeId 100 is called 'N:100'.
In the database they are stored with a ':' separator, but
most functions take the values separately for easier API
usage, so you can just call:

ME::VFS::Cache::Update('N', $Self->NodeId(), { Modified => time() });

if you want to 'touch()' the Node.

=cut

# Use Modules {{{

use strict;
use warnings;
use Carp;


use GlobalUtils;
use ME::Machine;
use ME::BasicCache;
use Data::Dumper;
use ME::DB;
use Encode qw/encode_utf8 decode decode_utf8/;
use IO::LockedFile;
use Storable;

# }}}

# Setup Variables {{{

our $_VFSDbh;

our $FSDBSpec = 'dbi:mysql:filestorage:mysql-filestorage.internal';
our $MMAPFile = -d('/tmpfs-nocleanup') ? '/tmpfs-nocleanup/vfscache' : '/tmpfs/vfscache';
our $MMAPLockFile = "$MMAPFile.lock";
our $NukeDirty = 0;
our %MEMCACHE;
our $IN_TRANSACTION = 0;
our $LAST_EXPIRY;
our %NEWIds;
our %DELETED_SHA1s;
our $HostName = ME::Machine->Name();

our $MMAPCache = CreateCache($MMAPFile, '1h', '64m', compress => 1);

our $DEBUG_CACHE = $ENV{DEBUG_VFS} || 0;

# }}}

# Configuration {{{

# DataSets are the actual database entries.  These are the only things
# that can have updates made to them.  Each lists a Table in the
# filestorage database, and the key field on that table which we will
# refer to each item by.  This in not necessarily the same as the primary
# key, but it must have a unique key defined on it.
my %DataSets = (
  N => {
    Table => 'Nodes',
    Key => 'NodeId',
    # NO PK here, we'll handle it up a layer
  },
  A => {
    Table => 'UserACLs',
    Key => 'UserACLId',
    Foreign => 'N',
    PK => ['NodeId', 'UserId', 'ACLName'],
  },
  P => {
    Table => 'Properties',
    Key => 'PropertyId',
    Foreign => 'N',
    PK => ['NodeId', 'NamespaceURI', 'PropName'],
  },
  L => {
    Table => 'Locks',
    Key => 'LockId',
    Foreign => 'N',
    PK => ['NodeId', 'NodeName', 'LockName'],
  },
  W => {
    Table => 'WebAccess',
    Key => 'WebAccessId',
    Foreign => 'N',
    PK => ['HostName', 'BasePath'],
  },
  NQ => {
    Table => 'NodeQuotas',
    Key => 'NodeId',
    Foreign => 'N',
    # doesn't have its own key - this one is messy because of usage
  },
);

# lookups are views on the table generated by fetching a set of rows
# by some condition.  They are hashes mapping from non-key attributes
# back to keys, so for example the ND "Node Dirs" attribute on a
# directory node will contain a hash mapping sub directory names to
# their nodeids,
#
# "NC:99" =>
# {
#    public => 100,
#    private => 101,
#    www => 102,
# };
#
# If it has a 'DataSet' option, then the data fetched will also be used
# to "populate" that dataset, so for example the above would create values
# for "N:100", "N:101" and "N:102" as well.
#
# Keys:
#  Table -> The table on which to query
#  Key -> The field from that table which matches
#  DataSet -> The dataset to write this value to, and also to get the
#             key value for HashKey below, see example...
#  HashKey -> The fields from that table to use for hash key items
#        (so: for NP below,
#           { $v{NamespaceUri} => { $v{PropName} => $v{$Tables{P}{Key}} } }
#         where %v is the fetched row from Properties table)
#  Where -> When looking this value up from the DB (not editing it in
#           memory) apply this clause as well as restricting on
#           $Lookup{Key} = $Item.
#  Fields -> Which rows to fetch from the DB (must be compatible with
#            DataSet above!)  Defaults to '*' if not given.
#
#  PerlWhere -> If considering a Table value being Created/Updated/Deleted
#               then test against this to see if it affects a value of
#               this lookup.
#  PerlKeys -> Rather than using the 'Key' value above, use this function
#              to work out the key value(s) based on the requested $Item
#              and don't add anything to $Lookup{Where} as it will already
#              have a ? in it.  Tricky.
#  NukeDirty -> Boolean, if set then don't bother updating the memory
#               cache, just ditch the cached data and let it re-fetch.
#
# In particular for updates - these lookups are applied both to the old AND
# new states of this item, and the difference between the two states is used
# to make changes to the stored value for the lookup.
my %Lookups = (
  NP => {            # Node Properties
    Table => 'Properties',
    Key => 'NodeId',
    DataSet => 'P',
    HashKey => ['NamespaceURI', 'PropName'],
  },
  NA => {            # Node ACLs
    Table => 'UserACLs',
    Key => 'NodeId',
    DataSet => 'A',
    HashKey => ['UserId', 'ACLName'],
  },
  NL => {            # Node Locks
    Table => 'Locks',
    Key => 'NodeId',
    DataSet => 'L',
    HashKey => ['NodeName', 'LockName'],
  },
  ND => {            # Node (sub)Directories
    Table => 'Nodes',
    Key => 'ParentNodeId',
    DataSet => 'N',
    HashKey => ['IsDeleted', 'NodeName'],
                 # We don't want children of the root node, only deleted      and only dirs
    PerlWhere => sub { $_[0]{ParentNodeId} != 1 and $_[0]{NodeType} eq 'd' },
    Where => "NodeType = 'd'",
  },
  NF => {            # Node (sub)Files
    Table => 'Nodes',
    Key => 'ParentNodeId',
    DataSet => 'N',
    HashKey => ['IsDeleted', 'NodeName'],
                 # We don't want children of the root node, only deleted      and only files
    PerlWhere => sub { $_[0]{ParentNodeId} != 1 and $_[0]{NodeType} eq 'f' },
    Where => "NodeType = 'f'",
  },
  UA => {            # Per-User ACLs
    Table => 'UserACLs',
    Key => 'UserId',
    DataSet => 'A',
    HashKey => ['NodeId', 'ACLName'],
  },
  UW => {            # Per-User Web Access
    Table => 'WebAccess',
    Key => 'UserId',
    DataSet => 'W',
    HashKey => ['HostName', 'BasePath'],
  },
  HW => {            # Per-Host Web Access
    Table => 'WebAccess',
    Key => 'HostName',
    DataSet => 'W',
    HashKey => 'BasePath',
  },
);

# special case Node sets
my %NodeSets = (
  'NDZ' => {
    Key => 'ParentNodeId',
    Where => "NodeType = 'd'",
  },
  'NFZ' => {
    Key => 'ParentNodeId',
    Where => "NodeType = 'f'",
  },
  'UT' => { # user tree
    Key => 'UserId',
    Where => "NodeType = 'd'",
  },
);

# }}}

=item I<VFSDbh()>

Connect to the 'filestorage' database if not already connected.

Returns a connected database handle.
=cut
sub VFSDbh {
  # if we were in a transaction and lost the database connection,
  # then it's all bad, need to fail back up to the user!
  if ($_VFSDbh && $IN_TRANSACTION && ! $_VFSDbh->{Active}) {
    die "Database connection disappeared in the middle of a transaction, please try again\n";
  }

  # we use transactions, can't have automatic reconnects
  return $_VFSDbh = ME::DB->GetDbh( Name => 'file', NoReconnect => 1 );
}

=item I<Init()>

Make sure the database is in a consistent state, and junk left over
from a previously failed transaction (but I hope that doesn't occur)
is cleaned up.

No defined return value.
=cut
sub Init {
  Rollback(); # make sure nothing left unfinished!

  # regardless, clean up
  warn "INIT\n" if $DEBUG_CACHE;
  %DELETED_SHA1s = ();
  %MEMCACHE = ();

  return undef;
}

=item I<Finish()>

This function should be called when you've finished using the VFS layer,
and just rolls back any changes that are not yet committed, and emits
and error into the error log as well.  Should be a NOOP if your code
isn't buggy.

No defined return value.

=cut
sub Finish {
  # these things can happen if you get a die.  Until we have proper
  # exceptions (i.e. never with Perl, *sigh*) it's easier to just roll
  # back out here.
  Rollback();
}

=item I<Begin()>

Start a transaction.  Should never be called directly, but it will be
called by anything which wishes to modify the database.

No defined return value.

=cut
sub Begin {
  return if $IN_TRANSACTION;
  VFSDbh()->begin_work();
  warn "BEGIN\n" if $DEBUG_CACHE;
  $IN_TRANSACTION = 1;
}

=item I<Commit()>

Commit all the changes made so far to the database.

No defined return value.
=cut
sub Commit {
  my $VFSDbh = VFSDbh();

  return unless $IN_TRANSACTION;

  $VFSDbh->commit();
  warn "COMMIT\n" if $DEBUG_CACHE;

  if (%DELETED_SHA1s) {
    eval "use ME::VFS::IOHandle";
    foreach my $sha1 (keys %DELETED_SHA1s) {
      ME::VFS::IOHandle::DeleteSha($sha1, $DELETED_SHA1s{$sha1});
    }
    %DELETED_SHA1s = ();
  }
  %MEMCACHE = (); # just FTP really

  $IN_TRANSACTION = 0;
}

=item I<Rollback()>

Roll back all changes, delete the cached changes in memory and cancel
them in the database.  Is a NOOP if not in a transaction.

No defined return value.
=cut
sub Rollback {
  return unless $IN_TRANSACTION;
  VFSDbh()->rollback();
  warn "ROLLBACK\n" if $DEBUG_CACHE;
  %DELETED_SHA1s = ();
  %MEMCACHE = ();
  $IN_TRANSACTION = 0;
}

=item I<Create($Cat, $Val)>

ONLY FOR USE WITH TABLES WITH AN auto_increment KEY!!!

Given a category (which must index into %Tables) and a
hash of values, create that record in the database.

Updates all Lookups appropriately.

Will start a transaction if necessary.

Returns the auto created value of $Tables{$Cat}{Key} for
the new record.

=cut
sub Create {
  my $Cat = shift;
  my $Val = shift;
  my $InOpts = shift;
  my $DataSet = $DataSets{$Cat} or die "Not a table $Cat\n";

  Begin();

  # Create the record
  my %Opts = (InsertIgnore => 1);
  if ($DataSet->{Foreign} && !$InOpts->{IgnoreForeign}) {
    # always Nodes, NodeId :)
    my $FDS = $DataSets{$DataSet->{Foreign}} or die "INTERNAL: broken foreign";
    $Opts{UseSelect} = [$FDS->{Table}, $FDS->{Key}];
  }
  my $Item = VFSDbh()->DHashCreateUnicode('UTF-8', $DataSet->{Table}, $Val, %Opts);
  return unless $Item;

  my $Name = "$Cat:$Item";
  $Val->{$DataSet->{Key}} = $Item;

  warn "CREATE $Name\n" if $DEBUG_CACHE;

  Dirty($Cat, $Item, $Val);
  FixLookups($Cat, $Item, undef, $Val);

  return $Item;
}

=item I<Update($Cat, $Item, $Val)>

ALSO USED TO 'Create' VALUES WHERE THE KEY IS PRE-KNOWN!!!

Given a category (which must index into %Tables) update the
record for $Item in the table.  Only the keys of $Val (a
hash) which are different from the old value are used to
update, and nothing is done if all keys are the same.

Updates all Lookups appropriately.

Will start a transaction if necessary.

No defined return value.

=cut
sub Update {
  my $Cat = shift;
  my $Item = shift;
  my $Val = shift;

  my $Name = "$Cat:$Item";
  my $DataSet = $DataSets{$Cat} or die "Not a table $Cat\n";

  Begin();

  warn "UPDATE $Name\n" if $DEBUG_CACHE;

  my $OldVal = Get($Cat, $Item);
  if ($OldVal) {
    my %Update;
    foreach my $key (keys %$Val) {
      next if ((not defined $OldVal->{$key} and not defined $Val->{$key})
            or (defined $OldVal->{$key} and defined $Val->{$key} and $OldVal->{$key} eq $Val->{$key}));
      $Update{$key} = $Val->{$key};
    }
    return unless keys %Update; # nothing changed!
    unless ($DataSet->{ReadOnly}) {
      VFSDbh()->DHashUpdateUnicode( 'UTF-8', $DataSet->{Table}, "$DataSet->{Key} = ?", \%Update, $Item);
    }
    my $NewVal = {%$OldVal, %Update};
    Dirty($Cat, $Item, $NewVal);
    FixLookups($Cat, $Item, $OldVal, $NewVal);
  } else {
    unless ($DataSet->{ReadOnly}) {
      my %Opts;
      if ($DataSet->{Foreign}) {
        # always Nodes, NodeId :)
        my $FDS = $DataSets{$DataSet->{Foreign}} or die "INTERNAL: broken foreign";
        $Opts{UseSelect} = [$FDS->{Table}, $FDS->{Key}];
      }
      VFSDbh()->DHashCreateUnicode( 'UTF-8', $DataSet->{Table}, $Val, %Opts);
    }
    Dirty($Cat, $Item, $Val);
    FixLookups($Cat, $Item, undef, $Val);
  }
}

=item I<Delete($Cat, $Item)>

Given a category (which must index into %Tables) delete the
record for $Item in the table.

Updates all Lookups appropriately.

Will start a transaction if necessary.

No defined return value.

=cut
sub Delete {
  my $Cat = shift;
  my $Item = shift;

  my $Name = "$Cat:$Item";
  my $DataSet = $DataSets{$Cat} or die "Not a table $Cat\n";

  Begin();

  warn "DELETE $Name\n" if $DEBUG_CACHE;

  my $OldVal = Get($Cat, $Item);

  VFSDbh()->do("DELETE FROM $DataSet->{Table} WHERE $DataSet->{Key} = ?", {}, $Item);

  Dirty($Cat, $Item, undef);
  FixLookups($Cat, $Item, $OldVal, undef);
}

=item I<FixLookups($Cat, $Item, $OldVal, $NewVal)>

Called whenever an item is Created, Updated or Deleted
(with $OldVal or $NewVal as 'undef' if it doesn't exists
at either end of the operation)

Calcuates which lookups are affected by each of $OldVal
and $NewVal, and if any, what the values of those lookups
are at that time.

Compares the results, and then if there are any changes
between the old and new values of this particular lookup
after the change, marks the lookup 'Dirty' and makes the
appropriate changes to the in memory copy of the lookup.

NOTE: this could just mark the lookup 'Dirty' and delete
our cached copy, but then operations like unzipping
thousands of files into a directory would make thousands
of queries to get NF of the parent dir, so this is more
efficient in that case.

No defined return value.

=cut
sub FixLookups {
  my $Cat = shift;
  my $Item = shift;
  my $OldVal = shift;
  my $NewVal = shift;

  warn "Fixing lookups $Cat:$Item\n" if $DEBUG_CACHE;

  # we just scan every single lookup to see if it might be
  # a hit!
  foreach my $Lookup (keys %Lookups) {
    my $OldData = $OldVal;
    my $NewData = $NewVal;
    my $Expire = $Lookups{$Lookup};
    next unless $Expire->{DataSet};
    my $KeyField = $DataSets{$Expire->{DataSet}}{Key};

    my %OldKeys;
    my %NewKeys;

    # so we get the keys from this - but only if there is no PerlWhere or
    # we match the PerlWhere.
    if ($Expire->{DataSet} eq $Cat) {
      if ($Expire->{Key}) {
        if ($OldVal and (not exists $Expire->{PerlWhere} or $Expire->{PerlWhere}->($OldVal))) {
          $OldKeys{$OldVal->{$Expire->{Key}}} = 1;
        }
        if ($NewVal and (not exists $Expire->{PerlWhere} or $Expire->{PerlWhere}->($NewVal))) {
          $NewKeys{$NewVal->{$Expire->{Key}}} = 1;
        }
      }
      if ($Expire->{PerlKeys}) {
        if ($OldVal) {
          $OldKeys{$_} = 1 for $Expire->{PerlKeys}->($OldVal);
        }
        if ($NewVal) {
          $NewKeys{$_} = 1 for $Expire->{PerlKeys}->($NewVal);
        }
      }
    }

    # At this point we have the keys that matched the old items, and the keys that matched the
    # new items.  We work out which are deletes and which are Adds, and also any that 'may' have
    # to be altered.
    my @Deletes = grep { not $NewKeys{$_} } keys %OldKeys;
    my @Adds = grep { not $OldKeys {$_} } keys %NewKeys;
    my @Alters = grep { $NewKeys{$_} } keys %OldKeys;

    # if all the hashkeys are equal, then the Lookup hasn't changed even if the value has, so we
    # can skip those "Alters", otherwise we put in both a "Delete" and an "Add" for this item.
    foreach my $Alter (@Alters) {
      foreach my $HashKey (::AsList($Expire->{HashKey})) {
        unless ($OldVal->{$HashKey} eq $NewVal->{$HashKey}) {
          push @Deletes, $Alter;
          push @Adds, $Alter;
          last;
        }
      }
    }

    # Nothing has changed, yay!
    next unless (@Adds or @Deletes);

    # If there's special flag "NukeDirty", then don't bother 
    # trying to update the cache, just ditch all the data and 
    # let it re-fetch from the DB.
    if ($NukeDirty or $Expire->{NukeDirty}) {
      my %All = map { $_ => 1 } (@Adds, @Deletes);
      Dirty($Lookup, $_) for keys %All;
    } else {
      foreach my $Key (@Deletes) {
        my @l = ::AsList($Expire->{HashKey});
        # XXX: we could just mark dirty rather than fetching if it's not already in memory!
        my $Data = Get($Lookup, $Key);
        my $hash = $Data;
        my @seen;
        # walk down the hash keys, stepping into each hash of the value.
        while (@l > 1) {
          my $this = shift @l;
          push @seen, [$hash, $OldVal->{$this}];
          $hash->{$OldVal->{$this}} ||= {};
          $hash = $hash->{$OldVal->{$this}};
        }
        # and delete the actual value.
        delete $hash->{$OldVal->{$l[0]}};

        # then if this leaves an empty hash at each parent level, remove that
        # hash from the data!
        foreach my $item (reverse @seen) {
          unless (keys %{$item->[0]->{$item->[1]}}) {
            delete $item->[0]->{$item->[1]};
          }
        }

        # and commit it back to memory.
        Dirty($Lookup, $Key, $Data);
      }

      foreach my $Key (@Adds) {
        my @l = ::AsList($Expire->{HashKey});
        # XXX: we could just mark dirty rather than fetching if it's not already in memory!
        my $Data = Get($Lookup, $Key);
        my $hash = $Data;
        # walk down the hash keys, stepping into each hash of the value, and creating that
        # hash first if needed.
        while (@l > 1) {
          my $this = shift @l;
          $hash->{$NewData->{$this}} ||= {};
          $hash = $hash->{$NewData->{$this}};
        }
        # then create the value.
        $hash->{$NewData->{$l[0]}} = $NewVal->{$KeyField};

        # and commit it back to memory.
        Dirty($Lookup, $Key, $Data);
      }
    }
  }

  # special case all the Node things
  if ($Cat eq 'N') {
    if ($OldVal and not $OldVal->{IsDeleted}) {
      if ($OldVal->{NodeType} eq 'd') {
        if ($MEMCACHE{"NDZ:$OldVal->{ParentNodeId}"}) {
          delete $MEMCACHE{"NDZ:$OldVal->{ParentNodeId}"}{$OldVal->{NodeName}};
        }
      }
      if ($OldVal->{NodeType} eq 'f') {
        if ($MEMCACHE{"NFZ:$OldVal->{ParentNodeId}"}) {
          delete $MEMCACHE{"NFZ:$OldVal->{ParentNodeId}"}{$OldVal->{NodeName}};
        }
        if ($MEMCACHE{"NFILES:$OldVal->{ParentNodeId}"}) {
          $MEMCACHE{"NFILES:$OldVal->{ParentNodeId}"}--;
        }
      }
    }
    if ($NewVal and not $NewVal->{IsDeleted}) {
      if ($NewVal->{NodeType} eq 'd') {
        if ($MEMCACHE{"NDZ:$NewVal->{ParentNodeId}"}) {
          $MEMCACHE{"NDZ:$NewVal->{ParentNodeId}"}{$NewVal->{NodeName}} = $Item;
        }
      }
      if ($NewVal->{NodeType} eq 'f') {
        if ($MEMCACHE{"NFZ:$NewVal->{ParentNodeId}"}) {
          $MEMCACHE{"NFZ:$NewVal->{ParentNodeId}"}{$NewVal->{NodeName}} = $Item;
        }
        if ($MEMCACHE{"NFILES:$NewVal->{ParentNodeId}"}) {
          $MEMCACHE{"NFILES:$NewVal->{ParentNodeId}"}++;
        }
      }
    }
  }
}

=item I<NodeQuotaDiff($QuotaNodeId, $Diff)>

Sets a change to be made to a node's quota at commit time.

Updates the "NQ" value for that NodeId as well, and returns
the new NQ value.

=cut
sub NodeQuotaDiff {
  my $NodeId = shift;
  my $Change = shift;
  my $Val = Get('NQ', $NodeId);
  if ($Change) {
    Begin(); # we need a transaction if we change any quotas
    VFSDbh()->do("UPDATE NodeQuotas SET Used = Used + ? WHERE NodeId = ?", {}, $Change, $NodeId);
    $Val->{Used} += $Change;
  }
  return $Val;
}

=item I<Dirty($Cat, $Item, $NewVal)>

Mark an item as being "Dirty" in the cache.  This means that its
value has been changed, so every other cache must invalidate any
copy it might have.

If $NewVal is passed, copies that value into local memory to avoid
having to re-fetch it.  If it's not passed the the memory copy is
deleted and it must be fetched again.

Returns $NewVal

=cut
sub Dirty {
  my $Cat = shift;
  my $Item = shift;
  my $NewVal = shift;

  my $Name = "$Cat:$Item";

  warn "DIRTY $Name\n" if $DEBUG_CACHE;
  delete $MEMCACHE{$Name};

  # allow fixup caching...
  $MEMCACHE{$Name} = $NewVal if $NewVal;

  return $NewVal;
}

=item I<Get($Cat, $Item)>

Fetch the value of an item.

Returns a hashref containing the item, or undef if there is no such value
in the DB (note: all Lookups will return an empty hash anyway, since
that just means there were no matching records at this point in time!)

It will read from the MEMCACHE if possible, or otherwise call 'Fetch'.

=cut
sub Get {
  my $Cat = shift;
  my $Item = shift;
  Carp::confess("No item") unless $Item;
  my $Name = "$Cat:$Item";

  return $MEMCACHE{$Name} if exists $MEMCACHE{$Name};

  if ($Cat eq 'NFILES') {
    # REALLY SPECIAL VALUE!
    return scalar (keys %{$MEMCACHE{"NFZ:$Item"}}) if exists $MEMCACHE{"NFZ:$Item"};
    # special case, fetch the tree
    my $Node = Get('N', $Item);
    my $UNFNAME = "UNFILES:$Node->{UserId}";
    Fetch('UNFILES', $Node->{UserId}) unless $MEMCACHE{$UNFNAME};
    # if there were no matches, we'll have to fill it in
    $MEMCACHE{$Name} = 0 unless exists $MEMCACHE{$Name};
    $MEMCACHE{$UNFNAME} = {};
  }

  if ($Cat eq 'NDZ' and $Item != 1) {
    # special case, fetch the tree
    my $Node = Get('N', $Item);
    my $UTNAME = "UT:$Node->{UserId}";
    Fetch('UT', $Node->{UserId}) unless $MEMCACHE{$UTNAME};
    $MEMCACHE{$UTNAME} = {};
  }

  return $MEMCACHE{$Name} if exists $MEMCACHE{$Name};

  my $Val = Fetch($Cat, $Item);
  return undef unless $Val;
  $MEMCACHE{$Name} = $Val;

  return $MEMCACHE{$Name};
}

=item I<Fetch($Cat, $Item)>

Fetch the value of an item from the MMAPCache or DB.

Will always skip the MMAPCache if the value is "Dirty" because the
MMAPCache will be out of date then, but the DB will be up to date
because within the transaction we have committed the changes back.

For %Tables - just fetch the row from the DB.

For %Lookups - fetch all rows that match by the 'Key' and 'Where'
values, and also store the values of each maching DataSet row
into the cache for faster access.

Returns a hashref being the value requested or undef if it's
a Table and the value doesn't exist.

=cut
sub Fetch {
  my $Cat = shift;
  my $Item = shift;

  my $Name = "$Cat:$Item";

  warn "Fetching $Name\n" if $DEBUG_CACHE;

  # otherwise, fetch from the DB
  my $VFSDbh = VFSDbh();

  # fetch the contents of a table if it's a table
  my $L;
  if ($Cat eq 'UNFILES') {
    my $sql = "SELECT ParentNodeId,COUNT(*) FROM Nodes " .
              "WHERE IsDeleted = 0 AND NodeType = 'f' AND UserId = ? " .
              "GROUP BY ParentNodeId";
    my $sth = $VFSDbh->prepare($sql);
    $sth->execute(encode_utf8($Item));
    while (my $Row = $sth->fetchrow_arrayref()) {
      $MEMCACHE{"NFILES:$Row->[0]"} = int($Row->[1]);
    }
  } elsif ($L = $NodeSets{$Cat}) {
    my @Where = ("$L->{Key} = ?");
    if ($L->{Where}) {
      push @Where, "$L->{Where}";
    }
    my $Fields = $L->{Fields} || '*';
    my $sql = "SELECT NodeId,UserId,ParentNodeId,NodeName," .
              "NodeType,Created,Modified,TempExpires,Sha1Hex,FileSize,"  .
              "PropertyId,NamespaceURI,PropName,PropValue,IsXML" .
              " FROM Nodes LEFT JOIN Properties USING (NodeId)" .
              " WHERE IsDeleted = 0 AND " . join(" AND ", @Where) .
              " ORDER BY NodeId ASC, PropertyId ASC";
    my $sth = $VFSDbh->prepare($sql);
    $sth->execute(encode_utf8($Item));
    while (my $Row = $sth->fetchrow_arrayref()) {
      for (@$Row) {
        next unless defined $_;
        utf8::decode($_) or die "Failed to decode $_ for $Name";
      }
      # got a node (may be repeated, but that's OK)
      $MEMCACHE{"N:$Row->[0]"} = {
        NodeId => $Row->[0],
        UserId => $Row->[1],
        ParentNodeId => $Row->[2],
        NodeName => $Row->[3],
        NodeType => $Row->[4],
        Created => $Row->[5],
        Modified => $Row->[6],
        TempExpires => $Row->[7],
        Sha1Hex => $Row->[8],
        FileSize => $Row->[9],
        IsDeleted => 0,
      };
      if ($Row->[4] eq 'f') {
        $MEMCACHE{"NFZ:$Row->[2]"}{$Row->[3]} = $Row->[0];
      }
      if ($Row->[4] eq 'd') {
        $MEMCACHE{"NDZ:$Row->[0]"} ||= {}; # in case it has none
        $MEMCACHE{"NDZ:$Row->[2]"}{$Row->[3]} = $Row->[0];
      }

      # got a property?
      if ($Row->[10]) {
        $MEMCACHE{"P:$Row->[10]"} = {
          PropertyId => $Row->[10],
          NodeId => $Row->[0],
          NamespaceURI => $Row->[11],
          PropName => $Row->[12],
          PropValue => $Row->[13],
          IsXML => $Row->[14],
        };
        $MEMCACHE{"NP:$Row->[0]"}{$Row->[11]}{$Row->[12]} = $Row->[10];
      }
      else {
        $MEMCACHE{"NP:$Row->[0]"} = {}; # cache negative lookup too
      }
    }
    $sth->finish();

    return $MEMCACHE{$Name} || {};

  } elsif ($L = $DataSets{$Cat}) {
    my @Where;
    if ($L->{Key}) {
      push @Where, "$L->{Key} = ?";
    }
    if ($L->{Where}) {
      push @Where, "$L->{Where}";
    }
    my $Fields = $L->{Fields} || '*';
    my $sql = "SELECT $Fields FROM $L->{Table} WHERE " . join(" AND ", @Where) . " ORDER BY $L->{Key} ASC";
    my $Data = $VFSDbh->selectrow_hashref($sql, {}, encode_utf8($Item));
    return undef unless $Data;
    for (values %$Data) {
      next unless defined $_;
      utf8::decode($_) or die "Failed to decode $_ for $Name";
    }
    return $Data;

  # otherwise it's an ID lookup
  } elsif ($L = $Lookups{$Cat}) {
    my $DataSet = $DataSets{$L->{DataSet}};
    my @Where;
    if ($L->{Key}) {
      push @Where, "$L->{Key} = ?";
    }
    if ($L->{Where}) {
      push @Where, "$L->{Where}";
    }
    die "No clauses" unless @Where;
    my $Fields = $L->{Fields} || '*';
    my $sql = "SELECT $Fields FROM $L->{Table} WHERE " . join(" AND ", @Where) . " ORDER BY $L->{Key} ASC";
    my $sth = $VFSDbh->prepare($sql);
    $sth->execute(encode_utf8($Item));
    my $Data = {};
    while (my $Row = $sth->fetchrow_hashref()) {
      for (values %$Row) {
        next unless defined $_;
        utf8::decode($_) or die "Failed to decode $_ ($L->{Table} $Item)";
      }
      # Create the lookup item by following the hashkeys
      my $Key = $Row->{$DataSet->{Key}};
      my @l = ::AsList($L->{HashKey});
      my $hash = $Data;
      while (@l > 1) {
        my $this = shift @l;
        $hash->{$Row->{$this}} ||= {};
        $hash = $hash->{$Row->{$this}};
      }
      unless (defined $Row->{$l[0]}) {
        warn "UNDEFINED KEY ERROR: " . Dumper($Row, $l[0], $Key, $L);
      }
      $hash->{$Row->{$l[0]}} = $Key;
      # and cache the fetched table data too!
      if ($L->{Table} ne 'Nodes' or $Row->{IsDeleted} == 0) {
        my $Name = "$L->{DataSet}:$Key";
        $MEMCACHE{$Name} = $Row;
      }
    }
    return $Data;
  } else {
    die "Unknown category $Cat\n";
  }
}

# GENERIC LOOKUPS

=item I<IsReadOnly($UserId)>

XXX - not implemented yet, always says "No".

Returns a boolean to say if the filestorage is read only for a
particular user.

=cut
sub IsReadOnly {
  my $UserId = shift;
  return 0;
}

sub RemoveDuplicates {
  my $verbose = shift;

  my $VFSDbh = VFSDbh();

  foreach my $DS (values %DataSets) {
    next unless $DS->{PK};
    print "doing $DS->{Table}\n" if $verbose;
    $VFSDbh->begin_work();
    my $Ids = $VFSDbh->selectcol_arrayref("SELECT A.$DS->{Key} FROM $DS->{Table} A INNER JOIN $DS->{Table} B USING (" . join(', ', @{$DS->{PK}}) . ") WHERE A.$DS->{Key} < B.$DS->{Key}");
    foreach my $Id (@$Ids) {
      print "removing duplicate $Id from $DS->{Table}\n" if $verbose;
      $VFSDbh->do("DELETE FROM $DS->{Table} WHERE $DS->{Key} = ?", {}, $Id);
    }
    $VFSDbh->commit();
  }
}

=item I<CleanupFiles()>

Designed to be called from cron - this function deletes records from
the Nodes table which are over 7 days deleted.

No defined return value.
=cut
sub CleanupFiles {
  my $VFSDbh = VFSDbh();
  my $verbose = shift;
  my $time = time();

  # get a list of nodes to delete
  my $NodeIds = $VFSDbh->selectcol_arrayref("SELECT NodeId FROM Nodes WHERE TempExpires < ? AND TempExpires > 0 ORDER BY TempExpires", {}, $time);

  my $total = @$NodeIds;
  my $n = 0;
  while (my @list = splice(@$NodeIds, 0, 100)) {
    DeleteNodes(\@list, $verbose);
    Commit();
    $n += @list;
    print "$n..$total\n" if $verbose;
  }
}

sub DeleteNodes {
  my $VFSDbh = VFSDbh();
  my $NodeIds = shift || [];
  my $verbose = shift || 0;

  # loop over, deleting each
  foreach my $NodeId (@{$NodeIds}) {
    # first, delete the node, but get a copy of its contents
    my $Node = Get('N', $NodeId);
    die "Missing node $NodeId\n"
      unless $Node;
    if ($verbose > 1) {
      my $Name = encode_utf8($Node->{NodeName});
      print "deleting node $NodeId ($Name, $Node->{ParentNodeId})\n";
    }
    Delete('N', $NodeId);

    # and if it's a file, see if we can delete the detail record
    if ($Node->{NodeType} eq 'f') {
      my ($Count) = $VFSDbh->selectrow_array("SELECT COUNT(*) FROM Nodes WHERE Sha1Hex = ? AND NodeId <> ?", {}, $Node->{Sha1Hex}, $NodeId);

      # delete if this is the only reference
      if (!$Count) {
        print "Deleting sha1 file $Node->{Sha1Hex} ($NodeId)\n" if $verbose > 1;
        # don't delete until after we commit the DB cleanup!
        $DELETED_SHA1s{$Node->{Sha1Hex}} = $Node->{FileSize};
      } else {
        print "Keeping sha1 file $Node->{Sha1Hex} ($Count additional copies)\n" if $verbose > 1;
      }
    }
  }
}

1;
