package ME::VFS::DirNode;

use strict;
use warnings;
use Carp;

use base qw(ME::VFS::Node);

sub NodeType { 'd' }

sub Create {
  my $class = shift;
  my %Opts = @_;

  my $Self = $class->SUPER::Create(%Opts, NodeType => 'd', SkipCommit => 1);

  # copy the view mode from the parent if one exists (e.g. photo gallery)
  # and any other properties :)
  my $ParentNode = $Self->ParentNode();
  foreach my $PropName (qw{ ViewMode }) {
    if (my $Value = $ParentNode->GetUserProperty($PropName)) {
      $Self->SetUserProperty($PropName, $Value, SkipCommit => 1);
    }
  }

  # no need to copy ACLs any more, Deep is automatic :)

  ME::VFS::Cache::Commit() unless $Opts{SkipCommit};

  return $Self;
}

sub Delete {
  my $Self = shift;
  my %Opts = @_;

  $Opts{IsDeleted} ||= $Self->NodeId();

  foreach my $SubDir ($Self->SubDirs()) {
    $SubDir->Delete(%Opts, SkipCommit => 1);
  }

  foreach my $File ($Self->Files()) {
    $File->Delete(%Opts, SkipCommit => 1);
  }

  $Self->SUPER::Delete(%Opts, SkipCommit => 1);

  ME::VFS::Cache::Commit() unless $Opts{SkipCommit};
}

sub UnDelete {
  my $Self = shift;
  my %Opts = @_;

  $Opts{IsDeleted} ||= $Self->NodeId();

  foreach my $SubDir ($Self->SubDirs(IsDeleted => $Opts{IsDeleted})) {
    $SubDir->UnDelete(%Opts, SkipCommit => 1);
  }

  foreach my $File ($Self->Files(IsDeleted => $Opts{IsDeleted})) {
    $File->UnDelete(%Opts, SkipCommit => 1);
  }

  $Self->SUPER::UnDelete(%Opts, SkipCommit => 1);

  ME::VFS::Cache::Commit() unless $Opts{SkipCommit};
}

sub Chown {
  my $Self = shift;
  my %Opts = @_;

  if ($Opts{Recurse}) {
    foreach my $SubDir ($Self->SubDirs()) {
      $SubDir->Chown(%Opts, SkipCommit => 1);
    }
    foreach my $File ($Self->Files()) {
      $File->Chown(%Opts, SkipCommit => 1);
    }
  }
  
  $Self->SUPER::Chown(%Opts, SkipCommit => 1);
}

sub SubDirs {
  my $Self = shift;
  my %Opts = @_;

  my @SubDirs = $Self->CanRead() ?
                $Self->SubDirsByLookup(%Opts) :
                $Self->SubDirsByACL(%Opts);

  return @SubDirs;
}

# Lookup Sub Directories from database directly.  Use this
# where you already have a parent with an ACL, so you know
# you can see everything in this directory.
sub SubDirsByLookup {
  my $Self = shift;
  my %Opts = @_;

  my $did = $Opts{IsDeleted} || 0;

  unless ($did) {
    my $Data = ME::VFS::Cache::Get('NDZ', $Self->NodeId());
    if ($Data) {
      return map { $Self->SubNode($Data->{$_}, 'ME::VFS::DirNode') }
             sort keys %$Data;
    }
  }

  my $NodeIds = ME::VFS::Cache::Get('ND', $Self->NodeId());

  my @SubDirs;
  if ($did == -1) {
    @SubDirs = map { $Self->SubNode($_, 'ME::VFS::DirNode') }
               map { values %$_ }
               values %{$NodeIds};
  } else {
    @SubDirs = map { $Self->SubNode($_, 'ME::VFS::DirNode') }
               values %{$NodeIds->{$did}};
  }

  my %SubDirs = map { $_->NodeName() => $_ } @SubDirs;
  return map { $SubDirs{$_} } sort keys %SubDirs;
}

# Get all the subdirectories which have children that
# have an ACL for this user.
#
# Use this where you DON'T have a parent with an ACL,
# so you can only see items which are on the path to
# "drill down" to something you _can_ see.
sub SubDirsByACL {
  my $Self = shift;
  my %Opts = @_;

  my $did = $Opts{IsDeleted} || 0;

  my $ACLNodeIds = ME::VFS::Cache::Get('UA', $Self->VFS->UserId());
  my @Nodes = map { $Self->VFS->Node($_) } keys %$ACLNodeIds;

  # implicit ACLs for HomeDir and TempDir
  push @Nodes, $Self->VFS->HomeDir(), $Self->VFS->TempDir();

  my $Depth = $Self->Depth();

  my %ById;
  foreach my $Node (@Nodes) {
    next unless ($did == -1 or $Node->IsDeleted() == $did);
    while ($Node->Depth() > $Depth) {
      my $ParentNode = $Node->ParentNode();
      if ($ParentNode->Equals($Self)) {
        # hash for duplicate elimination
        $ById{$Node->NodeId()} = $Node;
        last;
      }
      $Node = $ParentNode;
    }
  }

  my %SubDirs = map { $_->NodeName() => $_ } values %ById;
  return map { $SubDirs{$_} } sort keys %SubDirs;
}

sub Files {
  my $Self = shift;
  my %Opts = @_;

  # files are only visible if you have an ACL
  return unless $Self->CanRead();

  my $did = $Opts{IsDeleted} || 0;

  unless ($did) {
    my $Data = ME::VFS::Cache::Get('NFZ', $Self->NodeId());
    if ($Data) {
      return map { $Self->SubNode($Data->{$_}, 'ME::VFS::FileNode') }
             sort keys %$Data;
    }
  }

  my $NodeIds = ME::VFS::Cache::Get('NF', $Self->NodeId());

  my @Files;
  if ($did == -1) {
    @Files = map { $Self->SubNode($_, 'ME::VFS::FileNode') }
             map { values %$_ }
             values %{$NodeIds};
  } else {
    @Files = map { $Self->SubNode($_, 'ME::VFS::FileNode') }
             values %{$NodeIds->{$did}};
  }

  my %Files = map { $_->NodeName() => $_ } @Files;
  return map { $Files{$_} } sort keys %Files;
}

sub NumFiles {
  my $Self = shift;
  my %Opts = @_;

  # files are only visible if you have an ACL
  return unless $Self->CanRead();

  return ME::VFS::Cache::Get('NFILES', $Self->NodeId());
}

sub Search {
  my $Self = shift;
  my $Pattern = shift;
  my $Limit = shift;

  my $SearchRE = qr/\Q$Pattern\E/i;

  my @ResultNodes;

  # Include files at current level first, then sub-levels
  my @SearchNodes = ($Self->Files(), $Self->SubDirs());

  # Keep going until we reach limit or run out of nodes to check
  while (my $Node = shift @SearchNodes) {
    push @ResultNodes, $Node if $Node->NodeName() =~ $SearchRE;
    last if $Limit && @ResultNodes == $Limit;
    push @SearchNodes, ($Node->Files(), $Node->SubDirs()) if $Node->IsDir();
  }

  return @ResultNodes;
}

sub PathToNode {
  my $Self = shift;
  my $Path = shift;
  my %Opts = @_;

  if ($Path =~ s:(.*?)/::) {
    my $Name = $1;
    my $SubDir = $Self->DirByName($Name, %Opts);
    return undef unless $SubDir;
    return $SubDir if $Path eq ''; # trailing slash was the end
    return $SubDir->PathToNode($Path, %Opts);
  }

  return $Self->ChildByName($Path);
}

sub ChildByName {
  my $Self = shift;
  my $Name = shift;

  # workaround for DAV stuff mainly...
  return $Self unless defined $Name;

  return $Self->DirByName($Name) || $Self->FileByName($Name);
}

sub DirByName {
  my $Self = shift;
  my $Name = shift;
  my %Opts = @_;

  unless (defined $Name) {
    Carp::cluck("DirByName called with undef value for " . $Self->UserId() . "\n" . Data::Dumper::Dumper($Self));
    return $Self;
  }

  return $Self if $Name eq '';
  return $Self if $Name eq '.';
  return $Self->ParentNode() if $Name eq '..';

  if ($Self->CanRead()) {
    my $DirIds = ME::VFS::Cache::Get('NDZ', $Self->NodeId());

    if ($DirIds->{$Name}) {
      return $Self->SubNode($DirIds->{$Name}, 'ME::VFS::DirNode');
    }
    elsif ($Opts{CreateMissingDirs} && $Self->CanWrite()) {
      return $Self->Mkdir($Name);
    }
  }
  else {
    foreach my $Node ($Self->SubDirsByACL(%Opts)) {
      next unless $Node->NodeName() eq $Name;
      return $Node;
    }
  }

  return;
}

sub FileByName {
  my $Self = shift;
  my $Name = shift;

  unless (defined $Name) {
    Carp::cluck("FileByName called with undef value for " . $Self->UserId() . "\n" . Data::Dumper::Dumper($Self));
    return $Self;
  }

  if ($Self->CanRead()) {
    my $FileIds = ME::VFS::Cache::Get('NFZ', $Self->NodeId());

    if ($FileIds->{$Name}) {
      return $Self->SubNode($FileIds->{$Name}, 'ME::VFS::FileNode');
    }
  }

  return;
}

sub Mkdir {
  my $Self = shift;
  my $NodeName = shift;
  return ME::VFS::DirNode->Create(@_, NodeName => $NodeName, ParentNode => $Self);
}

sub OpenFile {
  my $Self = shift;
  my $NodeName = shift;
  my $Mode = shift;
  return ME::VFS::IOHandle->open(@_, NodeName => $NodeName, Mode => $Mode, ParentNode => $Self);
}

sub OpenTempFile {
  my $Self = shift;
  my $NodeName = shift;
  my $Mode = shift; # bit pointless to be 'r', but works symetrically if file exists
  return ME::VFS::IOHandle->open(@_,
    NodeName => $NodeName,
    Mode => $Mode,
    ParentNode => $Self,
    TempExpires => time() + 86400,
    TryRename => 1, 
    SkipQuotaCheck => 1, 
  );
}

sub ImportFile {
  my $Self = shift;
  my $File = shift;
  my $NodeName = shift;
  return ME::VFS::FileNode->CreateFromFile(@_, File => $File, NodeName => $NodeName, ParentNode => $Self);
}

sub ImportTempFile {
  my $Self = shift;
  my $File = shift;
  my $NodeName = shift;
  return ME::VFS::FileNode->CreateFromFile(@_, 
    File => $File, 
    NodeName => $NodeName,
    ParentNode => $Self,
    TempExpires => time() + 86400,
    TryRename => 1, 
    SkipQuotaCheck => 1, 
  );
}

sub MoveTo {
  my $Self = shift;
  my %Opts = @_;

  my $OldParentNode = $Self->ParentNode();
  $Self->SUPER::MoveTo(
    %Opts,
    SkipCommit => 1,
  );

  # remove cached paths
  $Self->_PostMoveDeep();

  ME::VFS::Cache::Commit() unless $Opts{SkipCommit};

  return $Self;
}

sub _PostMoveDeep {
  my $Self = shift;

  # recurse
  foreach my $SubNode ($Self->SubDirs(), $Self->Files()) {
    # Path will have changed
    delete $SubNode->{Data};
    $SubNode->_PostMoveDeep();
  }

  $Self->SUPER::_PostMoveDeep();
}

sub CopyTo {
  my $Self = shift;
  my %Opts = @_;

  my $NewNode = $Self->SUPER::CopyTo(%Opts, SkipCommit => 1);
  croak "Didn't get a new node" unless $NewNode;

  # ACLs are handled automatically by the create, so it's all good
  # Properties are handled by the ME::VFS::Node function called above.

  foreach my $SubDir ($Self->SubDirs()) {
    $SubDir->CopyTo(
      %Opts,
      ParentNode => $NewNode,
      NodeName => $SubDir->NodeName(),
      SkipCommit => 1,
    );
  }
  foreach my $File ($Self->Files()) {
    $File->CopyTo(
      %Opts,
      ParentNode => $NewNode,
      NodeName => $File->NodeName(),
      SkipCommit => 1,
    );
  }

  ME::VFS::Cache::Commit() unless $Opts{SkipCommit};

  return $NewNode;
}

sub FileSize {
  my $Self = shift;
  return 0;
}

sub ContentSize {
  my $Self = shift;
  my %Opts = @_;

  my $Size = 0;
  # add all the files
  foreach my $File ($Self->Files()) {
    $Size += $File->FileSize();
  }

  if ($Opts{IncludeSubDirs}) {
    # and potential files in each subdir
    foreach my $Dir ($Self->SubDirs()) {
      $Size += $Dir->ContentSize(%Opts);
    }
  }

  return $Size;
}

# Short form :)
sub Exists {
  my $Self = shift;
  return 0 if $Self->IsDeleted();
  return 0 unless $Self->ParentNode->Exists();
  my $Node = $Self->ParentNode->DirByName($Self->NodeName());
  return 0 unless $Node;
  return $Node->NodeId() == $Self->NodeId();
}

# ripe for optimisation
sub IsHomeDir {
  my $Self = shift;
  return $Self->VFS->HomeDir->Equals($Self);
}

sub IsTempDir {
  my $Self = shift;
  return $Self->VFS->TempDir->Equals($Self);
}

sub HasACL {
  my $Self = shift;
  my $ACLName = shift;
  my $UserId = shift || $Self->VFS->UserId();
  unless (exists $Self->{Data}{ACLs}{"$ACLName:$UserId"}) {
    my $UserACLs = shift || ME::VFS::Cache::Get('UA', $UserId);
    my $NodeId = $Self->{NodeId};
    my $HasACL = 0;
    if (exists $UserACLs->{$NodeId} and exists $UserACLs->{$NodeId}{$ACLName}) {
      $HasACL = 1;
    }
    elsif ($Self->Depth() == 2 and ($Self->IsHomeDir() or ($ACLName ne 'a' and $Self->IsTempDir()))) {
      $HasACL = 1;
    }
    else {
      $HasACL = $Self->ParentNode->HasACL($ACLName, $UserId, $UserACLs);
    }
    $Self->{Data}{ACLs}{"$ACLName:$UserId"} = $HasACL;
  }
  return $Self->{Data}{ACLs}{"$ACLName:$UserId"};
}

sub GetWebLink {
  my $Self = shift;
  my %Opts = @_;

  unless (exists $Self->{Data}{WebPath}) {
    foreach my $Access (sort { (defined $b->{Password}) <=> (defined $a->{Password}) } $Self->VFS->GetWebAccess()) {
      # I don't really care what order they're in.. so long as there's something there!
      next unless $Access->{NodeId}; # needs to have a node!
      next unless $ME::VFS::DisplayModes{$Access->{DisplayMode}}->[1] eq 'F';
      next unless $Access->{IsActive};
      my $Node = $Self->VFS->Node($Access->{NodeId});
      next unless ($Node and $Node->Exists()); # obviously has to BE there!
      next unless $Self->HasAncestor($Node); # can be reached from here...
      $Self->{Data}{WebPath} = [$Access->{HostName} . $Access->{BasePath} . $Self->PathRelativeTo($Node), $Access->{WebAccessId}, $Self->NodeId() == $Node->NodeId()];
      last;
    }
    $Self->{Data}{WebPath} = undef unless exists $Self->{Data}{WebPath};
  }

  if ($Self->{Data}{WebPath}) {
    $Self->{Data}{WebPath}[0] =~ s{/+$}{};
    if ( $Opts{SSL} ) {
      return ["https://$Opts{SSL}/userdirs/$Self->{Data}{WebPath}[0]/", $Self->{Data}{WebPath}[1], $Self->{Data}{WebPath}[2]];
    } else {
      return ["http://$Self->{Data}{WebPath}[0]/", $Self->{Data}{WebPath}[1], $Self->{Data}{WebPath}[2]];
    }
  }

  return undef;
}

# Quota stuff

sub IsQuotaNode {
  my $Self = shift;
  return ME::VFS::Cache::Get('NQ', $Self->NodeId()) ? 1 : 0;
}

sub GetQuota {
  my $Self = shift;

  my $Data = ME::VFS::Cache::Get('NQ', $Self->NodeId());

  $Data = { Used => 0, Total => 0 } unless $Data;

  return ($Data->{Used}, $Data->{Total});
}

sub GetQuotaUsed {
  my $Self = shift;
  my ($Used, $Total) = $Self->GetQuota();
  return $Used;
}

sub GetQuotaTotal {
  my $Self = shift;
  my ($Used, $Total) = $Self->GetQuota();
  return $Total;
}

sub _CheckQuotaFields {
  my $NodeId = shift;
  my $Used = shift;
  my $Total = shift;
  my %Opts = @_;

  if ($Total <= 0) {
    die "Unable to set quota to zero - delete the quota node manually\n";
  }
  if ($Used < 0) {
    warn "Setting used quota $Used back to 0 for node $NodeId";
    $Used = 0;
  }
  if ($Used > $Total) {
    unless ($Opts{SkipQuotaCheck}) {
      die "Can't go over quota $Used > $Total\n";
    }
  }

  return $Used;
}

sub SetQuota {
  my $Self = shift;
  my $Used = shift;
  my $Total = shift;
  my %Opts = @_;

  $Used = _CheckQuotaFields($Self->NodeId(), $Used, $Total, %Opts);

  ME::VFS::Cache::Update('NQ', $Self->NodeId(), { NodeId => $Self->NodeId(), Used => $Used, Total => $Total });
  ME::VFS::Cache::Commit() unless $Opts{SkipCommit};

  return 1;
}

sub SetQuotaUsed {
  my $Self = shift;
  my $Value = shift;
  my %Opts = @_;

  my ($Used, $Total) = $Self->GetQuota();

  return $Self->SetQuota($Value, $Total, %Opts);
}

sub SetQuotaTotal {
  my $Self = shift;
  my $Value = shift;
  my %Opts = @_;

  my ($Used, $Total) = $Self->GetQuota();

  return $Self->SetQuota($Used, $Value, %Opts);
}

sub AddQuotaUsed {
  my $Self = shift;
  my $Extra = shift;
  my %Opts = @_;

  my ($Used, $Total) = $Self->GetQuota();

  my $NewUsed = _CheckQuotaFields($Self->NodeId(), $Used + $Extra, $Total, %Opts);

  ME::VFS::Cache::NodeQuotaDiff($Self->NodeId(), $NewUsed - $Used);
  ME::VFS::Cache::Commit() unless $Opts{SkipCommit};

  return 1;
}

1;
