package ME::VFS::FileNode;

use strict;
use warnings;
use Carp;
use File::Path;
use File::Copy;
use File::Temp;
use MIME::Types;
use ME::Constants qw($ImageMimeTypeRE);
use Digest::SHA qw(sha1_hex);

use base qw(ME::VFS::Node);

use ME::VFS::IOHandle;

$::MimeTypes ||= MIME::Types->new;

sub NodeType { 'f' }

=pod

=head1 NAME

ME::VFS::FileNode - The interface for File specific components
of an ME::VFS::Node.

=head1 EXAMPLES

See also the ME::VFS::Node documentation for more details.

  my $Node = ME::VFS::FileNode->new($VFS, $NodeId);

  # NOTE: generally only called by ME::VFS::IOHandle when
  # creating new files, or CopyTo when creating a duplicate
  # node, but can be accessed directly to do strange things
  # if required.
  # See also ME::VFS::DirNode->OpenFile();
  my $NewNode = ME::VFS::FileNode->Create($VFS, {
    ParentNode => $ParentNode,
    NodeName => 'file.txt',
    Sha1Hex => ...,
    FileSize => ...,
  });

  my $Sha1Hex = $Node->Sha1Hex();
  my $FileSize = $Node->FileSize();

  my $NewNode = $Node->CopyTo(
    ParentNode => $NewParentDir,
    NodeName => $NewFileName,
  );
  $Node->MoveTo(
    ParentNode => $NewParentDir,
    NodeName => $NewFileName,
  );

  my $Path = $Node->Path();


=head1 DESCRIPTION

This module inherits from ME::VFS::Node

This works by creating a temporary file for 'w' or 'a', and copy the
content into the temp file if 'a'.  On calling of 'close' the contents
of the file are copied back into the file storage and the quota updated.

=head1 API

All functions take a hash of options.  If SkipCommit is passed then they
won't commit changes to the cache, speeding up operations like unzipping
hundreds of files into a single directory!

=over

=cut

=item I<Create(%Opts)>

Creates a new Node of type 'f' (File) in the database, and returns
the node object that was created.

=cut

sub Create {
  my $class = shift;
  my %Opts = @_;

  my $MimeType = delete $Opts{MimeType};

  my $VFS = $Opts{VFS} || $Opts{ParentNode}->VFS();
  my $ParentNode = $Opts{ParentNode} || $VFS->Node($Opts{ParentNodeId}, IgnoreACLs => 1);

  if ($VFS->IsReadOnly()) {
    die "ReadOnly Filesystem\n"; # get in before the FileDetail create
  }

  my $ThisNode = $ParentNode->ChildByName($Opts{NodeName});

  my $Self;
  if ($ThisNode and $ThisNode->Sha1Hex() eq $Opts{Sha1Hex}) {
    # don't create new node the file hasn't changed, some tools
    # upload the same file over and over again and we don't want
    # to blat the real change history.
    my $now = time();
    ME::VFS::Cache::Update('N', $ThisNode->NodeId(), {
      Modified => $Opts{Modified} || $now,
      Created => $Opts{Created} || $now,
      TempExpires => $Opts{TempExpires} || undef,
    });

    $Self = $VFS->Node($ThisNode->NodeId);
  }
  else {
    $Self = $class->SUPER::Create(%Opts, NodeType => 'f', SkipCommit => 1);
    return unless $Self; # Handle DB collision gracefully

    my $QuotaNode = $Self->QuotaNode();
    if ($QuotaNode) {
      $QuotaNode->AddQuotaUsed($Opts{FileSize}, 
                               SkipQuotaCheck => $Opts{SkipQuotaCheck}, 
                               SkipCommit => 1); # it's created!
    }

    $Self->_CleanupDeletedHistory(SkipCommit => 1);
  }

  if ($MimeType) {
    $Self->SetSysProperty('MimeType', $MimeType, SkipCommit => 1);
  }

  ME::VFS::Cache::Commit() unless $Opts{SkipCommit};

  return $Self;
}

sub MoveTo {
  my $Self = shift;
  my %Opts = @_;

  # normalise options
  unless ($Opts{TryRename}) {
    my $OldParent = $Self->ParentNode();
    my $OldName = $Self->NodeName();
    my $NewParent = $Opts{ParentNode} || $OldParent;
    my $NewName = $Opts{NodeName} || $OldName;
    my $Dest = $Opts{Node};
    if ($Dest) {
      $NewParent = $Dest->ParentNode();
      $NewName = $Dest->NodeName();
    } else {
      $Dest = $NewParent->ChildByName($NewName);
    }

    if ($Dest and $Dest->IsFile() and $Dest->Sha1Hex() eq $Self->Sha1Hex()) {
      # don't create new node the file hasn't changed, some tools
      # upload the same file over and over again and we don't want
      # to blat the real change history.
      ME::VFS::Cache::Update('N', $Dest->NodeId(), {
        Modified => $Self->Modified(),
        Created => $Self->Created(),
        TempExpires => $Self->TempExpires(),
      });
      $Self->Delete(SkipCommit => 1); # still got moved!
      ME::VFS::Cache::Commit() unless $Opts{SkipCommit};
      return $NewParent->VFS->Node($Dest->NodeId());
    }
  }

  $Self = $Self->SUPER::MoveTo(%Opts, SkipCommit => 1);
  $Self->_CleanupDeletedHistory(SkipCommit => 1);

  ME::VFS::Cache::Commit() unless $Opts{SkipCommit};

  return $Self;
}

sub _CleanupDeletedHistory {
  my $Self = shift;
  my %Opts = @_;

  # we don't want to keep more than 30 deleted copies of a file
  # because things like a webcam uploader cause thousands of tiny
  # files and bog the directory loading down!
  my $NodeFiles = ME::VFS::Cache::Get('NF', $Self->ParentNode->NodeId());
  my $Name = $Self->NodeName();
  my @Ids = map { $NodeFiles->{$_}{$Name} or () } # convert to NodeId
            sort { $b <=> $a } # sort by IsDeleted - with the newest first
            grep { $_ } # skip the non deleted one
            keys %$NodeFiles;
  if (@Ids > 30) {
    ME::VFS::Cache::DeleteNodes([splice(@Ids, 25)]);
    ME::VFS::Cache::Commit() unless $Opts{SkipCommit};
  }
  # Don't want a huge number of deleted files collecting on a
  # node because it causes excess load on the servers.  We want
  # to be able to keep 30 old copies on a single node still (above)

  # these are active files
  my $Alive = keys(%{$NodeFiles->{'0'}});
  # there's only one IsDeleted per deleted file, so keys of the top
  # level is an accurate proxy for count of nodes.  Remove one for
  # the alive key (0)
  my $Deleted = keys(%$NodeFiles) - 1;
  # add 30 to the denominotor.  Not only safe against div#0, but
  # makes sure this only kicks in when Deleted is high.
  if (($Deleted / ($Alive + 30)) > 2) {
    # find the 30 oldest.  This saves having to run this code path
    # EVERY upload when the folder is "on the edge"... take larger
    # chunks instead.
    my $NodeIds = ME::VFS::Cache::VFSDbh->selectcol_arrayref(<<EOF, {}, $Self->ParentNodeId());
SELECT NodeId
FROM Nodes
WHERE IsDeleted > 0 AND ParentNodeId = ?
ORDER BY TempExpires ASC
LIMIT 30
EOF
    ME::VFS::Cache::DeleteNodes($NodeIds);
    ME::VFS::Cache::Commit() unless $Opts{SkipCommit};
  }
} 

=item I<CreateFromFile(%Opts)>

Create a node from a new file.  Opens an IO::VFS::IOHandle in write mode
and copies $Opts{File} into it.

=cut

sub CreateFromFile {
  my $class = shift;
  my %Opts = @_;

  my $VFS = $Opts{VFS} || $Opts{ParentNode}->VFS();
  if ($VFS->IsReadOnly()) {
    die "ReadOnly Filesystem\n";
  }

  my $File = delete $Opts{File};

  my $fh = ME::VFS::IOHandle->open(%Opts, Mode => 'w', SkipCommit => 1);
  unless ($fh) {
    my $Error = "$!";
    my $Path = $Opts{ParentNode}->Path();
    ME::VFS::Cache::Rollback();
    die "Failed to create $Path/$Opts{NodeName}: $Error\n";
  }
  File::Copy::copy($File, $fh);

  my $NewNode = $fh->close();
  unless ($NewNode and $NewNode->IsFile()) {
    my $Error = "$!";
    my $Path = $Opts{ParentNode}->Path();
    ME::VFS::Cache::Rollback();
    die "Failed to create $Path/$Opts{NodeName}: $Error\n";
  }

  ME::VFS::Cache::Commit() unless $Opts{SkipCommit};

  return $NewNode;
}

=item I<Delete(%Opts)>

Delete this node.  See the description in ME::VFS::Node

=cut

sub Delete {
  my $Self = shift;
  my %Opts = @_;

  if ($Self->VFS->IsReadOnly()) {
    die "ReadOnly Filesystem\n";
  }

  my $QuotaNode = $Self->QuotaNode();
  if ($QuotaNode) {
    $QuotaNode->AddQuotaUsed(- $Self->FileSize(), 
                             SkipQuotaCheck => 1,
                             SkipCommit => 1); # it's being deleted!
  }

  $Self->SUPER::Delete(%Opts, SkipCommit => 1);

  ME::VFS::Cache::Commit() unless $Opts{SkipCommit};
}

sub UnDelete {
  my $Self = shift;
  my %Opts = @_;

  if ($Self->VFS->IsReadOnly()) {
    die "ReadOnly Filesystem\n";
  }

  $Self->SUPER::UnDelete(%Opts, SkipCommit => 1);

  # no free lunch
  my $QuotaNode = $Self->QuotaNode();
  if ($QuotaNode) {
    $QuotaNode->AddQuotaUsed($Self->FileSize(), 
                             SkipQuotaCheck => $Opts{SkipQuotaCheck},
                             SkipCommit => 1);
  }

  ME::VFS::Cache::Commit() unless $Opts{SkipCommit};
}

=item I<Open($Mode, %Opts)>

Open this file.  $Mode defaults to 'r' but can be 'r' (read),
'w' (write) or 'a' (append).  See ME::VFS::IOHandle for details on
how append works.

=cut

sub Open {
  my $Self = shift;
  my $Mode = shift;
  return ME::VFS::IOHandle->open(@_, Mode => $Mode, Node => $Self);
}

=item I<Slurp()>

Read the file into a scalar and return the content

=cut

sub Slurp {
  my $Self = shift;
  my $Fh = $Self->Open('r');
  local $/ = undef;
  my $Scalar = <$Fh>;
  close($Fh);
  return $Scalar;
}

=item I<FileSize()>

Returns the size of this file in bytes.

=cut

sub FileSize {
  my $Self = shift;
  return int($Self->NodeData->{FileSize});
}

=item I<MimeType()>

Returns the mime type of the file (guessed by the file(1) utility
in the current implementation, but may also be passed to the
Create function)

=cut

sub MimeType {
  my $Self = shift;
  my %Opts = @_;
  my $Prop = $Self->GetSysProperty('MimeType');
  if (not $Opts{ForceAuto} and $Prop) {
    return $Prop;
  } 
  elsif ($Opts{NoAuto}) {
    return undef;
  }
  else {
    my $MimeTypeObj = $::MimeTypes->mimeTypeOf($Self->NodeName());
    return $MimeTypeObj ? $MimeTypeObj->type() : 'application/octet-stream';
  }
}

sub SetMimeType {
  my $Self = shift;
  my $MimeType = shift;
  $Self->SetSysProperty('MimeType', $MimeType, @_);
}

=item I<Sha1Hex()>

Returns the hexadecimal encoding of the SHA1 (Digest::SHA)
hash of the bytes in the file.  This is used to uniquely
identify the file contents, both for storage on disk and
possibly for ETag and other applications.

=cut

sub Sha1Hex {
  my $Self = shift;
  return $Self->NodeData->{Sha1Hex};
}

=item I<DiskPath($Fetch)>

Returns the path on disk to the underlying file.  If $Fetch is true
then fetch the file from a backend server it isn't already cached.

=cut

sub DiskPath {
  my $Self = shift;
  my $Fetch = shift;
  return ME::VFS::IOHandle::ShaDiskPath($Self->Sha1Hex(), $Fetch);
}

=item I<FileToken($Version)>

Returns a token by which the file can be fetched from VFS without authentication

=cut
sub FileToken {
  my $Self = shift;
  my $Version = shift // 'v1';

  die "Invalid token version" unless $Version eq 'v1';

  my $UnqId = ME::User::UserIdToUnqId($Self->UserId());
  my $NodeId = $Self->NodeId();
  my $Unique = sha1_hex($Self->NodeId() . $Self->FileSize() . $Self->Sha1Hex());
  return "v1-$UnqId-$NodeId-$Unique";
}

sub IsPicture {
  my $Self = shift;
  my $NodeName = $Self->NodeName();
  return 0 if ( $NodeName =~ m/^\./i);
  my $MimeType = $Self->MimeType();
  return 1 if ($MimeType =~ $ImageMimeTypeRE);
  return 1 if ($NodeName =~ m/\.(?:jpe?g|gif|bmp|png)$/i);
  return 0;
}

# Short form :)
sub Exists {
  my $Self = shift;
  return 0 if $Self->IsDeleted();
  my $Node = $Self->ParentNode->FileByName($Self->NodeName());
  return 0 unless $Node;
  return 0 unless $Self->ParentNode->Exists();
  return $Node->NodeId() == $Self->NodeId();
}

sub GetWebLink {
  my $Self = shift;
  my %Opts = @_;

  my $ParentWebLink = $Self->ParentNode->GetWebLink(%Opts);

  if ($ParentWebLink) {
    $ParentWebLink->[0] =~ s{/+$}{};
    return [$ParentWebLink->[0] . '/' . $Self->NodeName(), $ParentWebLink->[1], 0];
  }

  return undef;
}

# stub ends

sub IsQuotaNode { 0; }

sub HasACL { 
  my $Self = shift;
  return $Self->ParentNode->HasACL(@_);
}

1;
