package ME::VFS::IOHandle;

# Use Modules {{{

use strict;
use warnings;

use Digest::SHA;
use Errno qw(ENOSPC EFBIG EBADF EACCES EEXIST EISDIR EIO ENOENT EROFS);
use File::Copy;
use File::Path;
use File::Temp;
use IO::File;
use IO::Socket;
use Carp;

use base qw/IO::File/;

our $Sha1Dir = "/var/fastmail/userdata/sha1";
our $VFSTmpDir = "/var/fastmail/userdata/tmp";

our $sha1 = Digest::SHA->new();

our $ShaDepth = 2;

# }}}

=pod

=head1 NAME

ME::VFS::IOHandle - An IO::Handle interface to ME::VFS

=head1 EXAMPLES

  my $io = ME::VFS::IOHandle->open(
    VFS => $FS,     # ME::VFS object
    Dir => $Dir,            # /public/ - must have trailing slash
    NodeName => $NodeName,  # example.txt
    Mode => $Mode,          # 'r', 'w', or 'a'
  );
  or IOHandle->open(...);

  $io->syswrite(...);

  $io->close();

=head1 DESCRIPTION

This module wraps IO::File to support all the quota and copy-on-complete
functionality required by ME::VFS.

This works by creating a temporary file for 'w' or 'a', and copy the
content into the temp file if 'a'.  On calling of 'close' the contents
of the file are copied back into the file storage and the quota updated.

=head1 API

There is an AUTOLOAD function which maps back to the underlying module
API for IO::File, with the following exceptions.

=over

=cut

=item I<open(%Opts)>

Sets $! and dies for the following errors:
 EISDIR : Attempt to open a directory node
 EACCES : No read permission on file for 'r'
          No write permission on file for 'w|a'
 EEXIST : File already exists for write, overwrite is not enabled
 ENOENT : File does not exist for read or append without Create => 1
 EIO    : Underlying SHA1 file missing for read or append (may occur
          if file storage NFS mount is offline!)
 EROFS  : File storage is in READ ONLY mode for user or global and
          you have attempted to write a file.

Takes the following parameters:

 Node => ME::VFS::FileNode OR
 ParentNode => ME::VFS::DirNode AND
 NodeName => "string"
 Mode => 'r', 'w', 'a' (default 'r')
 Overwrite => BOOLEAN (default false) (ONLY FOR 'w')
 Create => BOOLEAN (default true) (ONLY FOR 'a)

 Will be created with default ACLs - deep copy of parent node.

=cut
sub open {
  my $class = shift;
  my %Opts = @_;

  # Calculate the node we're working with
  my $Node;
  my $ParentNode;
  my $NodeName;
  if (exists $Opts{Node}) {
    $Node = delete $Opts{Node};
    $Opts{ParentNode} = $ParentNode = $Node->ParentNode();
    $Opts{NodeName} = $NodeName = $Node->NodeName();
  } else {
    $ParentNode = $Opts{ParentNode};
    $NodeName = $Opts{NodeName};
    $Node = $ParentNode->ChildByName($NodeName);
  }

  unless ($ParentNode and $ParentNode->IsDir()) {
    $! = ENOENT;
    return undef;
  }

  my $Mode = $Opts{Mode} || 'r';
  my $Overwrite = $Opts{Overwrite} || 0;
  my $Create = exists $Opts{Create} ? $Opts{Create} : 1;
  my $TryRename = $Opts{TryRename} || 0;

  my $Self;

  # Attempt to open the file for reading
  if ($Mode eq 'r') {
    unless ($Node) {
      $! = ENOENT;
      return undef;
    }

    if ($Node->IsDir()) {
      $! = EISDIR;
      return undef;
    }

    unless ($Node->CanRead()) {
      $! = EACCES;
      return undef;
    }

    # open the underlying filestorage file
    my $Path = ShaDiskPath($Node->Sha1Hex(), 1);
    unless ($Path) {
      warn "NO PATH ON VFS " . $Node->Sha1Hex() . "\n";
      $! = EIO;
      # underlying file doesn't exist, this should never happen!
      return undef;
    }
    $Self = IO::File->new($Path, 'r');

    unless ($Self) {
      warn "MISSING FILE ON VFS " . $Node->Sha1Hex() . "\n";
      $! = EIO;
      # underlying file doesn't exist, this should never happen!
      return undef;
    }

    # bless down into this class
    $Self = bless $Self, ref($class) || $class;

    # tell the syswrite function we can't write to this file!
    ${*$Self}{me_filestorage_iohandle_Writable} = 0;
    ${*$Self}{me_filestorage_iohandle_Path} = $Path;

  # We want to write to the file!
  } elsif ($Mode eq 'w' or $Mode eq 'a') {
    if ($ParentNode->VFS->IsReadOnly()) {
      $! = EROFS;
      return undef;
    }

    if ($Node and $Node->IsDir()) {
      $! = EISDIR;
      return undef;
    }

    if ($Mode eq 'w' and not $TryRename and not $Overwrite and $Node) {
      $! = EEXIST;
      return undef;
    }

    if ($Mode eq 'a' and not $Create and not $Node) {
      $! = ENOENT;
      return undef;
    }

    # you must be able to write to the parent node for any operation, including
    # changing a file that already exists because a write is a delete and create.
    unless ($ParentNode->CanWrite()) {
      $! = EACCES;
      return undef;
    }

    # don't set unlink here - we'll handle it later if needed, or 'mv' the
    # file into its final position.
    my $TmpDir = $ENV{TMPDIR} || $VFSTmpDir;
    unless (-d $TmpDir) {
      File::Path::mkpath($TmpDir);
      chown(99, 99, $TmpDir);
    }
    $Self = File::Temp->new(UNLINK => 0, DIR => $TmpDir);
    my $TempFile = $Self->filename();
    $Self = bless $Self, ref($class) || $class;

    my ($Used, $Total) = (0, 0);
    my $QuotaNode = $ParentNode->QuotaNode();
    if ($QuotaNode) {
       ($Used, $Total) = $QuotaNode->GetQuota();
    }

    # set up all the variables we need to check how big the file can get!
    ${*$Self}{me_filestorage_iohandle_Writable} = 1; # open for writing!
    ${*$Self}{me_filestorage_iohandle_MaxFileSize} = $ParentNode->VFS->MaxFileSize();
    ${*$Self}{me_filestorage_iohandle_Quota} = $Total;
    ${*$Self}{me_filestorage_iohandle_RemainingSpace} = $Total - $Used;
    ${*$Self}{me_filestorage_iohandle_CurrSize} = 0;
    ${*$Self}{me_filestorage_iohandle_TempFile} = $TempFile; # make sure we have a copy of the filename
    ${*$Self}{me_filestorage_iohandle_Opts} = \%Opts;
    ${*$Self}{me_filestorage_iohandle_FileNode} = $Node;
    ${*$Self}{me_filestorage_iohandle_Path} = $TempFile;

    if ($Node) {
      # we will be deleting what's currently there before we write
      # the new file in, so this space is also free.
      ${*$Self}{me_filestorage_iohandle_RemainingSpace} += $Node->FileSize();

      # but if it's append we also need to copy in what exists of
      # the file.  The extra amount in CurrSize will exactly balance
      # out the extra RemainingSpace.
      if ($Opts{Mode} eq 'a') {
        my $FSPath = ShaDiskPath($Node->Sha1Hex(), 1);
        my $oldfh = IO::File->new($FSPath, 'r');
        unless ($oldfh) {
          $Self->close();
          unlink($TempFile); # don't leak tempfiles!
          $! = EIO;
          return undef;
        }
        $Self->seek(0, 0); # start of file
        File::Copy::copy($oldfh, $Self);
        $Self->seek(0, 2); # end of file
        $oldfh->close();
        # will be subtracted before quota calculated

        ${*$Self}{me_filestorage_iohandle_CurrSize} += $Node->FileSize();
      }
    }
  } else {
    return "Invalid mode $Mode!";
  }

  return $Self;
}

=item I<syswrite(...)>

Checks that there is sufficient space left to fit within both the
RemainingSpace and MaxFileSize values.  If over RemainingSpace
returns ENOSPC, if over MaxFileSize return EFBIG.

If the file was opened Read Only, return

=cut
sub syswrite {
  my $Self = shift;

  unless (${*$Self}{me_filestorage_iohandle_Writable}) {
    # don't set a permanant error, they can still read ok :)
    $! = EBADF;
    return undef;
  }

  # already failed for some reason!
  if (exists ${*$Self}{me_filestorage_iohandle_Errno}) {
    $! = ${*$Self}{me_filestorage_iohandle_Errno};
    if (exists ${*$Self}{me_filestorage_iohandle_TempFile}) {
      unlink(${*$Self}{me_filestorage_iohandle_TempFile});
      delete ${*$Self}{me_filestorage_iohandle_TempFile};
    }
    return undef;
  }

  # calculate how many characters to write;
  my $buflen = length($_[0]);
  my $writelen = $buflen;
  if (defined $_[2]) {
    $writelen -= $_[2];
  }
  if (defined $_[1] and $_[1] < $writelen) {
    $writelen = $_[1];
  }

  # Over maximum file size for the user, abort immediately
  if (not ${*$Self}{me_filestorage_iohandle_Opts}{SkipSizeCheck} and
      (${*$Self}{me_filestorage_iohandle_CurrSize} + $writelen) 
         > ${*$Self}{me_filestorage_iohandle_MaxFileSize}
     ) {
    ${*$Self}{me_filestorage_iohandle_Errno} = $! = EFBIG;
    if (exists ${*$Self}{me_filestorage_iohandle_TempFile}) {
      unlink(${*$Self}{me_filestorage_iohandle_TempFile});
      delete ${*$Self}{me_filestorage_iohandle_TempFile};
    }
    return undef;
  }
  
  # Over quota for user, abort immediately
  elsif (not ${*$Self}{me_filestorage_iohandle_Opts}{SkipQuotaCheck} and 
         ${*$Self}{me_filestorage_iohandle_Quota} and 
         (${*$Self}{me_filestorage_iohandle_CurrSize} + $writelen)
             > ${*$Self}{me_filestorage_iohandle_RemainingSpace}
        ) {
    ${*$Self}{me_filestorage_iohandle_Errno} = $! = ENOSPC;
    if (exists ${*$Self}{me_filestorage_iohandle_TempFile}) {
      unlink(${*$Self}{me_filestorage_iohandle_TempFile});
      delete ${*$Self}{me_filestorage_iohandle_TempFile};
    }
    return undef;
  }

  # Otherwise try to do the real write and catch errors.
  else {
    my $res = $Self->SUPER::syswrite(@_);
    unless (defined $res) {
      ${*$Self}{me_filestorage_iohandle_Errno} = $!;
      if (exists ${*$Self}{me_filestorage_iohandle_TempFile}) {
        unlink(${*$Self}{me_filestorage_iohandle_TempFile});
        delete ${*$Self}{me_filestorage_iohandle_TempFile};
      }
      return undef;
    }
    ${*$Self}{me_filestorage_iohandle_CurrSize} += $res;
    return $res;
  }

  # never get here
}

=item I<close()>

If there is a TempNodeName, then check if there has been an error.
If no error, then copy the contents back to the original NodeName
and Dir in the VFS.

Unlink the TempNodeName if it exists.

=cut
sub close {
  my $Self = shift;

  my $NewNode;

  # Only copy the file if the TempFile exists.  We delete it in every
  # other case!
  if (exists ${*$Self}{me_filestorage_iohandle_TempFile}) {

    # get info from the hash
    my $TempFile = delete ${*$Self}{me_filestorage_iohandle_TempFile};
    my $Opts = ${*$Self}{me_filestorage_iohandle_Opts};

    # Caller wants to throw it away, just delete it and move on
    if (exists ${*$Self}{me_filestorage_iohandle_DeleteOnClose}) {
      unlink($TempFile);
    }

    else {
      # remove the old one if it exists.  We must be allowed to, because
      # we checked for that in ->open(...);
      # SKIP, breaks TryRename - Overwrite should catch this later anyway
      #if (${*$Self}{me_filestorage_iohandle_FileNode}) {
        #${*$Self}{me_filestorage_iohandle_FileNode}->Delete(%$Opts);
        #delete ${*$Self}{me_filestorage_iohandle_FileNode};
      #}

      $NewNode = CreateFromTempFile($TempFile, %$Opts, Fh => $Self);
    }
  }

  my $res = $Self->SUPER::close();

  if ($NewNode) {
    return $NewNode;
  } else {
    return $res;
  }
}

=item I<filename()>

Get pathname to underlying file. This is just to match
the File::Temp interface to make passing a either a
File::Temp object or a $Node->Open() filehandle work

=cut
sub filename {
  my $Self = shift;
  return ${*$Self}{me_filestorage_iohandle_Path};
}

=item I<DeleteOnClose()>

If called, file will be deleted when close and not submitted to VFS. Obviously
this only makes sense for writing, never for reading.

=cut

sub DeleteOnClose {
  my $Self = shift;
  if (exists ${*$Self}{me_filestorage_iohandle_TempFile}) {
    ${*$Self}{me_filestorage_iohandle_DeleteOnClose} = 1;
  }
  else {
    carp "Can't set DeleteOnClose flag on a read handle";
  }
}

=item I<DESTROY()>

Destructor handle calls ->close() explicity to ensure data is copied
correctly in write case.

=cut
sub DESTROY {
  my $Self = shift;
  local $@ = undef;
  $Self->close();
}

=back

=cut

#################
# helper methods, non OO

sub CreateFromTempFile {
  my $TempFile = shift;
  my %Opts = @_;

  my $fh = delete $Opts{Fh} || IO::File->new($TempFile);
  $fh->seek(0, 0);
  $sha1->reset();
  $sha1->addfile($fh);
  my $FileSize = -s $fh;
  $fh->close();
  my $Sha1Hex = $sha1->hexdigest();

  # move the temp file to its final destination
  # XXX: assumes sha1 is 100% shiny-safe
  my $Path = ShaDiskPath($Sha1Hex);
  File::Copy::move($TempFile, $Path);
  chown(99, 99, $Path);
  chmod 0644, $Path;

  unless (-f $Path) {
    warn "Failed to copy file in to place $TempFile => $Path\n";
    die "Failed to copy file in to place $TempFile => $Path\n";
  }

  my $res = '';
  eval {
    my $s = IO::Socket::UNIX->new('/var/run/vfsproxy/sock');
    $s->print("PUT $Path $Sha1Hex $FileSize\n");
    $res = $s->getline();
  };

  unless ($res =~ m/OK $Sha1Hex $FileSize/) {
    warn "Failed to upload file $res\n";
    die "Failed to upload file $res\n";
  }

  # finally, the file is in place so let's create the node
  my $Name = delete $Opts{NodeName};
  if ($Opts{TryRename}) {
    my $BaseName = $Name;
    my $ParentNode = $Opts{ParentNode};
    my $n = 2;

    # loop until we find a name that's not in use!
    while ($n < 1000) {
      my $Node = $ParentNode->PathToNode($Name);
      if ($Node) {
        # exit if we've already found ourselves!
        if ($Node->IsFile() and $Node->Sha1Hex eq $Sha1Hex) {
          $Node->SetExpires($Opts{TempExpires}, %Opts) 
            if ($Opts{TempExpires} and $Node->TempExpires() and
                $Opts{TempExpires}  >  $Node->TempExpires());
          return $Node;
        }
      }
      else {
        # ok, we can try this filename
        my $NewNode = ME::VFS::FileNode->Create(
          %Opts,
          NodeName => $Name,
          Sha1Hex => $Sha1Hex,
          FileSize => $FileSize,
          IgnoreErrors => 1,
        );
        return $NewNode if $NewNode;
      }

      # generate a new name
      $Name = _GenerateName($BaseName, $n++);

      # reasonable limits apply...
    }
    die "Unable to find unused temp file name based on " . $ParentNode->Path() . "/$BaseName\n";
  }

  my $NewNode = ME::VFS::FileNode->Create(
    %Opts,
    NodeName => $Name,
    Sha1Hex => $Sha1Hex,
    FileSize => $FileSize,
  );

  return $NewNode;
}

sub _GenerateName {
  my $Name = shift;
  my $Num = shift;

  if ($Name =~ m/^(.+)\.(.+)$/) {
    return "$1\[$Num\].$2";
  }
  else {
    return "$Name\[$Num\]";
  }
}

sub ShaDiskPath {
  my $Sha1Hex = shift;
  my $DoFetch = shift;
  my @SubDirs = split //, $Sha1Hex, ($ShaDepth+1);
  pop @SubDirs; # Last item is remainder
  my $dir = join "/", $Sha1Dir, @SubDirs;
  unless (-d $dir) {
    File::Path::mkpath($dir);
    chown(99, 99, $dir);
  }
  my $file = "$dir/$Sha1Hex";
  if ($DoFetch) {
    foreach my $num (1..3) {
      last if -f $file;
      eval {
        my $s = IO::Socket::UNIX->new('/var/run/vfsproxy/sock');
        die "UNABLE TO CONNECT\n" unless $s;
        $s->print("GET $file.tmp $Sha1Hex\n");
        my $response = $s->getline();
        die "no response" unless defined $response;
        die "$response\n" unless $response =~ m/^OK $Sha1Hex/;
      };
      if ($@) {
        unlink("$file.tmp");
        warn "vfsproxy fetch error for $Sha1Hex: $@\n";
        sleep 1;
      }
      else {
        rename("$file.tmp", $file);
      }
    }
  }
  return $file;
}

sub DeleteSha {
  my $Sha1Hex = shift;
  my $Size = shift;
  my $response;
  foreach my $num (1..3) {
    eval {
      my $s = IO::Socket::UNIX->new('/var/run/vfsproxy/sock');
      die "UNABLE TO CONNECT\n" unless $s;
      $s->print("DEL $Sha1Hex $Size\n");
      $response = $s->getline();
      die "no response" unless defined $response;
      die "$response\n" unless $response =~ m/^OK $Sha1Hex/;
    };
    if ($@) {
      warn "vfsproxy failed to delete $Sha1Hex: $@\n";
      sleep 1;
    }
    else {
      my $file = ShaDiskPath($Sha1Hex);
      unlink($file) if -f $file;
      return $response;
    }
  }
}

# the end
1;
