package ME::VFS::Interface::DAV;

=head1 ME::VFS::Interface::DAV

An implementation of: http://www.ietf.org/rfc/rfc2518.txt
for MessagingEngine's Filestorage and Apache mod_perl.

=cut

# Use modules {{{
use strict;
use warnings;
use ME::ApacheRequest;
use Apache2::Const qw(:common);
use File::Temp;
use Socket;
use ME::VFS;
use ME::User;
use Encode;
use HTTP::Date;
use XML::LibXML;
use HTML::GenerateUtil qw(escape_html escape_uri);
use ME::VFS::Interface::Base;
use ME::VFS::Interface::HTTP::DirListingRenderer;

# see Apache2::Const manpage
our %methods = (
  DELETE    => \&DELETE_handler,
  GET       => \&GET_handler,
  HEAD      => sub { GET_handler(@_, 1) },
  OPTIONS   => \&OPTIONS_handler,
  PUT       => \&PUT_handler,
  PROPFIND  => \&PROPFIND_handler,
  PROPPATCH  => \&PROPPATCH_handler,
  MKCOL     => \&MKCOL_handler,
  COPY      => \&COPY_handler,
  MOVE      => \&MOVE_handler,
  LOCK      => \&LOCK_handler,
  UNLOCK    => \&UNLOCK_handler,
);

our $PARSER = XML::LibXML->new(
            no_network => 1,
            load_ext_dtd => 0,
            expand_entities => 0,
            expand_xinclude => 0,
            ext_ent_handler => sub {
                # my ($sys_id, $pub_id) = @_;
                # warn "Received external entity: $sys_id:$pub_id";
                "";
            },
);

# }}}

# DAV users always have to be authenticated, there's no such thing
# as an anonymous DAV session on our server.
#
# This is the standard Apache AuthenHandler function.
sub authen_handler {
  my $R = ME::ApacheRequest->new(shift);

  my $Host = $R->hostname() || "dav.messagingengine.com";
  $R->auth_name($Host);

  # grab details
  my $Password = $R->get_basic_auth_pw();
  my $Username = $R->user;
  my $Ip = $R->ip();
  my $SSL = $R->https();

  my $User = ME::User->new_login($Username, $Password, 'dav', $Ip, $SSL);

  # error if it isn't.
  unless ($User) {
    $R->note_basic_auth_failure();
    return AUTH_REQUIRED;
  }

  # set the username for logging (so if they used username#domain
  # it still comes out the same)
  $R->user(scalar($User->UnqName()));

  # set user object for use in the main request
  $R->pnotes('User', $User);

  return OK;
}

# This is the main Apache ContentHandler function - everything
# happens through here.
sub content_handler {
  my $R = ME::ApacheRequest->new(shift);

  if (my $FN = $methods{$R->method}) {
    my $User = $R->pnotes('User');
    my $VFS = $User->VFS(Init => 1);

    # Collect information from the headers set by the proxy
    my $Hostname = $R->hostname() || 'dav.messagingengine.com';
    my $SSL = $R->https();
    my $S = $SSL ? 's' : '';

    if ($Hostname =~ /^myfiles\./) {
      $VFS->Chroot($VFS->HomeDir());
    }

    # Generate the Base URI for responses
    my $Base = "http$S://$Hostname";

    # Reset the URI to be the one that was originally requested
    my $Uri;

    # Includes hostname component, remove it
    if ($Uri = $R->header_in('X-URL')) {
      $Uri =~ s{^.*?/}{/};
    }

    else {
      $Uri = $R->unparsed_uri();
      if ($Uri =~ m/\#/) {
        return 424; # ERROR - fragment not allowed
      }
    }

    # Tidy up the Destination header as well if it exists.
    # get rewritten.
    my $Destination = $R->header_in('Destination');
    if ($Destination) {
      # because of the broken-ness of the proxy rewriter only working on
      # non-SSL connections for the Destination header, we try both!
      $Destination =~ s{^$Base}{}; # if that works, otherwise
      $Destination =~ s{^https?://[^/]+/dav/}{/}; # there's always this way
      $Destination =~ s{^https?://[^/]+/}{/}; # finally...
    }

    my $ClientQuirks = _DetectClientQuirks($R, $User);

    my %Details = (
      User => $User,
      FS => $VFS,
      Base => $Base,
      SSL => $SSL,
      Path => $VFS->FixUriString($Uri, $ClientQuirks),
      Destination => $VFS->FixUriString($Destination, $ClientQuirks),
      ClientQuirks => $ClientQuirks,
    );

   #if (my $v = $R->header_in('X-Litmus')) {
     #my $m = $R->method();
     #my $d = $Destination ? " ($Destination)" : '';
     #warn "X-Litmus: $v - $m $Uri$d \n";
   #}

   #if (my $v = $R->header_in('X-Litmus-second')) {
     #my $m = $R->method();
     #my $d = $Destination ? " ($Destination)" : '';
     #warn "X-Litmus-second: $v - $m $Uri$d \n";
   #}

    my $RC;
    eval {
      $RC = $FN->($R, \%Details);
    };
    if ($@) {
      warn "Overall Die $@";
      return 500; # internal server error
    }
    if (0) { # sucks
      my %resp = (
        404 => 'SC_NOT_FOUND',
        201 => 'SC_CREATED',
        204 => 'SC_NO_CONTENT',
      );
      if ($resp{$RC}) {
        my $str = "$resp{$RC}\n";
        $R->status($RC);
        $R->set_content_length(do { use bytes; length($str) } + 1);
        $R->content_type('text/plain; charset=utf-8');
        $R->print($str);
        return OK;
      }
    }
    $VFS->Cleanup();
    my $UnqName = $User->UnqName();
    if (-d "/var/log/dav/$UnqName") {
      if (open(FH, ">>/var/log/dav/$UnqName/log.$$")) {
        my $headers_in = $R->{r}->headers_in();
        my $headers_out = $R->{r}->headers_out();
        print FH Data::Dumper::Dumper({
          method => $R->method, 
          rc => $RC, 
          uri => $Uri,
          headers_in => {%$headers_in}, 
          headers_out => {%$headers_out}, 
          path => $Details{Path}, 
          destination => $Details{Destination}, 
          body_in => $Details{body_in}, 
          body_out => $Details{body_out}
        });
        close(FH);
      }
    }
    return $RC;
  }

  # return Unimplemented if it's not one of the supported methods
  return 501;
}

# doc?
sub OPTIONS_handler {
  my($R, $Opts) = @_;
  $R->status(200);
  $R->header_out('DAV' => '1, 2');
  $R->header_out('MS-Author-Via' => 'DAV');
  $R->header_out('Allow' => join(', ', sort keys %methods));
  $R->header_out('Public' => join(', ', sort keys %methods));
  $R->header_out('Keep-Alive' => 'timeout=15, max=96');
  $R->set_content_length(0);
  $R->content_type('httpd/unix-directory');
  return OK;
}

=head2 8.1 PROPFIND

   The PROPFIND method retrieves properties defined on the resource
   identified by the Request-URI, if the resource does not have any
   internal members, or on the resource identified by the Request-URI
   and potentially its member resources, if the resource is a collection
   that has internal member URIs.  All DAV compliant resources MUST
   support the PROPFIND method and the propfind XML element (section
   12.14) along with all XML elements defined for use with that element.

   A client may submit a Depth header with a value of "0", "1", or
   "infinity" with a PROPFIND on a collection resource with internal
   member URIs.  DAV compliant servers MUST support the "0", "1" and
   "infinity" behaviors. By default, the PROPFIND method without a Depth
   header MUST act as if a "Depth: infinity" header was included.

   A client may submit a propfind XML element in the body of the request
   method describing what information is being requested.  It is
   possible to request particular property values, all property values,
   or a list of the names of the resource's properties.  A client may
   choose not to submit a request body.  An empty PROPFIND request body
   MUST be treated as a request for the names and values of all
   properties.

   All servers MUST support returning a response of content type
   text/xml or application/xml that contains a multistatus XML element
   that describes the results of the attempts to retrieve the various
   properties.

   If there is an error retrieving a property then a proper error result
   MUST be included in the response.  A request to retrieve the value of
   a property which does not exist is an error and MUST be noted, if the
   response uses a multistatus XML element, with a response XML element
   which contains a 404 (Not Found) status value.

   Consequently, the multistatus XML element for a collection resource
   with member URIs MUST include a response XML element for each member
   URI of the collection, to whatever depth was requested. Each response
   XML element MUST contain an href XML element that gives the URI of
   the resource on which the properties in the prop XML element are
   defined.  Results for a PROPFIND on a collection resource with
   internal member URIs are returned as a flat list whose order of
   entries is not significant.

   In the case of allprop and propname, if a principal does not have the
   right to know whether a particular property exists then the property
   should be silently excluded from the response.

   The results of this method SHOULD NOT be cached.

=item 8.1.1 Example - Retrieving Named Properties

   >>Request

   PROPFIND  /file HTTP/1.1
   Host: www.foo.bar
   Content-type: text/xml; charset="utf-8"
   Content-Length: xxxx

   <?xml version="1.0" encoding="utf-8" ?>
   <D:propfind xmlns:D="DAV:">
     <D:prop xmlns:R="http://www.foo.bar/boxschema/">
          <R:bigbox/>
          <R:author/>
          <R:DingALing/>
          <R:Random/>
     </D:prop>
   </D:propfind>

   >>Response

   HTTP/1.1 207 Multi-Status
   Content-Type: text/xml; charset="utf-8"
   Content-Length: xxxx

   <?xml version="1.0" encoding="utf-8" ?>
   <D:multistatus xmlns:D="DAV:">
     <D:response>
          <D:href>http://www.foo.bar/file</D:href>
          <D:propstat>
               <D:prop xmlns:R="http://www.foo.bar/boxschema/">
                    <R:bigbox>
                         <R:BoxType>Box type A</R:BoxType>
                    </R:bigbox>
                    <R:author>
                         <R:Name>J.J. Johnson</R:Name>
                    </R:author>
               </D:prop>
               <D:status>HTTP/1.1 200 OK</D:status>
          </D:propstat>
          <D:propstat>
               <D:prop><R:DingALing/><R:Random/></D:prop>
               <D:status>HTTP/1.1 403 Forbidden</D:status>
               <D:responsedescription> The user does not have access to
   the DingALing property.
               </D:responsedescription>
          </D:propstat>
     </D:response>
     <D:responsedescription> There has been an access violation error.
     </D:responsedescription>
   </D:multistatus>

   In this example, PROPFIND is executed on a non-collection resource
   http://www.foo.bar/file.  The propfind XML element specifies the name
   of four properties whose values are being requested. In this case
   only two properties were returned, since the principal issuing the
   request did not have sufficient access rights to see the third and
   fourth properties.

=item 8.1.2 Example - Using allprop to Retrieve All Properties

   >>Request

   PROPFIND  /container/ HTTP/1.1
   Host: www.foo.bar
   Depth: 1
   Content-Type: text/xml; charset="utf-8"
   Content-Length: xxxx

   <?xml version="1.0" encoding="utf-8" ?>
   <D:propfind xmlns:D="DAV:">
     <D:allprop/>
   </D:propfind>

   >>Response

   HTTP/1.1 207 Multi-Status
   Content-Type: text/xml; charset="utf-8"
   Content-Length: xxxx

   <?xml version="1.0" encoding="utf-8" ?>
   <D:multistatus xmlns:D="DAV:">
     <D:response>
          <D:href>http://www.foo.bar/container/</D:href>
          <D:propstat>
               <D:prop xmlns:R="http://www.foo.bar/boxschema/">
                    <R:bigbox>
                         <R:BoxType>Box type A</R:BoxType>
                    </R:bigbox>
                    <R:author>
                         <R:Name>Hadrian</R:Name>
                    </R:author>
                    <D:creationdate>
                         1997-12-01T17:42:21-08:00
                    </D:creationdate>
                    <D:displayname>
                         Example collection
                    </D:displayname>
                    <D:resourcetype><D:collection/></D:resourcetype>
                    <D:supportedlock>
                         <D:lockentry>
                              <D:lockscope><D:exclusive/></D:lockscope>
                              <D:locktype><D:write/></D:locktype>
                         </D:lockentry>
                         <D:lockentry>
                              <D:lockscope><D:shared/></D:lockscope>
                              <D:locktype><D:write/></D:locktype>
                         </D:lockentry>
                    </D:supportedlock>
               </D:prop>
               <D:status>HTTP/1.1 200 OK</D:status>
          </D:propstat>
     </D:response>
     <D:response>
          <D:href>http://www.foo.bar/container/front.html</D:href>
          <D:propstat>
               <D:prop xmlns:R="http://www.foo.bar/boxschema/">
                    <R:bigbox>
                         <R:BoxType>Box type B</R:BoxType>
                    </R:bigbox>
                    <D:creationdate>
                         1997-12-01T18:27:21-08:00
                    </D:creationdate>
                    <D:displayname>
                         Example HTML resource
                    </D:displayname>
                    <D:getcontentlength>
                         4525
                    </D:getcontentlength>
                    <D:getcontenttype>
                         text/html
                    </D:getcontenttype>
                    <D:getetag>
                         zzyzx
                    </D:getetag>
                    <D:getlastmodified>
                         Monday, 12-Jan-98 09:25:56 GMT
                    </D:getlastmodified>
                    <D:resourcetype/>
                    <D:supportedlock>
                         <D:lockentry>
                              <D:lockscope><D:exclusive/></D:lockscope>
                              <D:locktype><D:write/></D:locktype>
                         </D:lockentry>
                         <D:lockentry>
                              <D:lockscope><D:shared/></D:lockscope>
                              <D:locktype><D:write/></D:locktype>
                         </D:lockentry>
                    </D:supportedlock>
               </D:prop>
               <D:status>HTTP/1.1 200 OK</D:status>
          </D:propstat>
     </D:response>
   </D:multistatus>

   In this example, PROPFIND was invoked on the resource
   http://www.foo.bar/container/ with a Depth header of 1, meaning the
   request applies to the resource and its children, and a propfind XML
   element containing the allprop XML element, meaning the request
   should return the name and value of all properties defined on each
   resource.

   The resource http://www.foo.bar/container/ has six properties defined
   on it:

   http://www.foo.bar/boxschema/bigbox,
   http://www.foo.bar/boxschema/author, DAV:creationdate,
   DAV:displayname, DAV:resourcetype, and DAV:supportedlock.

   The last four properties are WebDAV-specific, defined in section 13.
   Since GET is not supported on this resource, the get* properties
   (e.g., getcontentlength) are not defined on this resource. The DAV-
   specific properties assert that "container" was created on December
   1, 1997, at 5:42:21PM, in a time zone 8 hours west of GMT
   (creationdate), has a name of "Example collection" (displayname), a
   collection resource type (resourcetype), and supports exclusive write
   and shared write locks (supportedlock).

   The resource http://www.foo.bar/container/front.html has nine
   properties defined on it:

   http://www.foo.bar/boxschema/bigbox (another instance of the "bigbox"
   property type), DAV:creationdate, DAV:displayname,
   DAV:getcontentlength, DAV:getcontenttype, DAV:getetag,
   DAV:getlastmodified, DAV:resourcetype, and DAV:supportedlock.

   The DAV-specific properties assert that "front.html" was created on
   December 1, 1997, at 6:27:21PM, in a time zone 8 hours west of GMT
   (creationdate), has a name of "Example HTML resource" (displayname),
   a content length of 4525 bytes (getcontentlength), a MIME type of
   "text/html" (getcontenttype), an entity tag of "zzyzx" (getetag), was
   last modified on Monday, January 12, 1998, at 09:25:56 GMT
   (getlastmodified), has an empty resource type, meaning that it is not
   a collection (resourcetype), and supports both exclusive write and
   shared write locks (supportedlock).

=item 8.1.3 Example - Using propname to Retrieve all Property Names

   >>Request

   PROPFIND  /container/ HTTP/1.1
   Host: www.foo.bar
   Content-Type: text/xml; charset="utf-8"
   Content-Length: xxxx

   <?xml version="1.0" encoding="utf-8" ?>
   <propfind xmlns="DAV:">
     <propname/>
   </propfind>

   >>Response

   HTTP/1.1 207 Multi-Status
   Content-Type: text/xml; charset="utf-8"
   Content-Length: xxxx

   <?xml version="1.0" encoding="utf-8" ?>
   <multistatus xmlns="DAV:">
     <response>
          <href>http://www.foo.bar/container/</href>
          <propstat>
               <prop xmlns:R="http://www.foo.bar/boxschema/">
                    <R:bigbox/>
                    <R:author/>
                    <creationdate/>
                    <displayname/>
                    <resourcetype/>
                    <supportedlock/>
               </prop>
               <status>HTTP/1.1 200 OK</status>
          </propstat>
     </response>
     <response>
          <href>http://www.foo.bar/container/front.html</href>
          <propstat>
               <prop xmlns:R="http://www.foo.bar/boxschema/">
                    <R:bigbox/>
                    <creationdate/>
                    <displayname/>
                    <getcontentlength/>
                    <getcontenttype/>
                    <getetag/>
                    <getlastmodified/>
                    <resourcetype/>
                    <supportedlock/>
               </prop>
               <status>HTTP/1.1 200 OK</status>
          </propstat>
     </response>
   </multistatus>


   In this example, PROPFIND is invoked on the collection resource
   http://www.foo.bar/container/, with a propfind XML element containing
   the propname XML element, meaning the name of all properties should
   be returned.  Since no Depth header is present, it assumes its
   default value of "infinity", meaning the name of the properties on
   the collection and all its progeny should be returned.

   Consistent with the previous example, resource
   http://www.foo.bar/container/ has six properties defined on it,
   http://www.foo.bar/boxschema/bigbox,
   http://www.foo.bar/boxschema/author, DAV:creationdate,
   DAV:displayname, DAV:resourcetype, and DAV:supportedlock.

   The resource http://www.foo.bar/container/index.html, a member of the
   "container" collection, has nine properties defined on it,
   http://www.foo.bar/boxschema/bigbox, DAV:creationdate,
   DAV:displayname, DAV:getcontentlength, DAV:getcontenttype,
   DAV:getetag, DAV:getlastmodified, DAV:resourcetype, and
   DAV:supportedlock.

   This example also demonstrates the use of XML namespace scoping, and
   the default namespace.  Since the "xmlns" attribute does not contain
   an explicit "shorthand name" (prefix) letter, the namespace applies
   by default to all enclosed elements.  Hence, all elements which do
   not explicitly state the namespace to which they belong are members
   of the "DAV:" namespace schema.

=cut
sub PROPFIND_handler {
  my($R, $Opts) = @_;
  my $Path = $Opts->{Path};
  my $FS = $Opts->{FS};

  my ($ParentNode, $name) = eval { $FS->ParentNodeAndName($Path) };
  return 404 if $@ || !$ParentNode;

  # breaks lock discovery, hmm
  #my @locks = _ParseIfHeader($R->header_in('If'));
  #return 423 if $ParentNode->NameLocked($name, \@locks, 'r');

  my $FileInfo = $ParentNode->ChildByName($name);
  return 404 unless $FileInfo;

  my $Depth = _ParseDepthHeader($R->header_in('Depth'));

  my $reqinfo = 'allprop';
  my @reqprops;
  if (my $len = $R->header_in('Content-length')) {
    my $content;
    $R->read($content, $len);
    $Opts->{body_in} = $content;
    my $doc;
    eval {
      $doc = $PARSER->parse_string($content);
    };
    # Litmus: props.3 says no prefix with a blank namespace please!
    if ($@ or not $doc or _HasNULLNamespace($doc)) {
      return 400;
    }

    # if there's content it has to be a propfind request with a node
    my ($typenode) = _GetNSNodes($doc, 'DAV:', 'propfind');
    return 400 unless $typenode;

    # if it's a prop request then we need to read the list of props
    $reqinfo = $typenode->localname();
    if ($reqinfo eq 'prop') {
      foreach my $node (_GetNSNodes($doc, 'DAV:', 'propfind', 'prop')) {
        push @reqprops, [$node->namespaceURI || '', $node->localname];
      }
    }
  }

  my $doc = XML::LibXML::Document->new('1.0', 'utf-8');
  my $multistat = $doc->createElement('multistatus');
  $multistat->setNamespace('DAV:', 'D'); # apply it at once!
  $doc->setDocumentElement($multistat);

  # find the relevant files to return data on.
  my @infos = ($FileInfo);
  if ($Depth eq 'one' and $FileInfo->IsDir()) {
    push @infos, $FileInfo->SubDirs();
    push @infos, $FileInfo->Files();
  } elsif ($Depth eq 'infinity' and $FileInfo->IsDir()) {
    _GetAllChildren($FileInfo, \@infos);
  }

  my @response;
  my ($QuotaUsed, $QuotaTotal) = $FS->GetQuota();
  for my $info (@infos) {
    my $itempath = $info->Path();
    $itempath =~ s{/*$}{/} if $info->IsDir(); # trailing slash on dirs
    my $hreftext;

    # special case for Konqueror - can't handle URI encoded UTF-8.
    if ($Opts->{ClientQuirks}{HasUTF8Path}) {
      my $hrefpath = encode_utf8($itempath);
      $hreftext = $hrefpath
    } else {
      my $hrefpath = ME::VFS::Interface::Base::GenerateUriPath($itempath);
      $hreftext = $hrefpath;
    }

    my $mtime = HTTP::Date::time2str($info->Modified());
    my $ctime = HTTP::Date::time2isoz($info->Created());
    $ctime =~ s/ /T/; # proper ISO
    my $type = $info->IsDir() ? 'httpd/unix-directory' : $info->MimeType();
    my $displayname = encode_utf8($info->NodeName());

=over 4

13 DAV Properties

   For DAV properties, the name of the property is also the same as the
   name of the XML element that contains its value. In the section
   below, the final line of each section gives the element type
   declaration using the format defined in [REC-XML]. The "Value" field,
   where present, specifies further restrictions on the allowable
   contents of the XML element using BNF (i.e., to further restrict the
   values of a PCDATA element).

13.1 creationdate Property

   Name:       creationdate
   Namespace:  DAV:
   Purpose:    Records the time and date the resource was created.
   Value:      date-time ; See Appendix 2
   Description: The creationdate property should be defined on all DAV
   compliant resources.  If present, it contains a timestamp of the
   moment when the resource was created (i.e., the moment it had non-
   null state).

   <!ELEMENT creationdate (#PCDATA) >

13.2 displayname Property

   Name:       displayname
   Namespace:  DAV:
   Purpose:    Provides a name for the resource that is suitable for
   presentation to a user.
   Description: The displayname property should be defined on all DAV
   compliant resources.  If present, the property contains a description
   of the resource that is suitable for presentation to a user.

   <!ELEMENT displayname (#PCDATA) >

13.3 getcontentlanguage Property

   Name:       getcontentlanguage
   Namespace:  DAV:
   Purpose:    Contains the Content-Language header returned by a GET
   without accept headers
   Description: The getcontentlanguage property MUST be defined on any
   DAV compliant resource that returns the Content-Language header on a
   GET.
   Value:      language-tag   ;language-tag is defined in section 14.13
   of [RFC2068]

   <!ELEMENT getcontentlanguage (#PCDATA) >

13.4 getcontentlength Property

   Name:       getcontentlength
   Namespace:  DAV:
   Purpose:    Contains the Content-Length header returned by a GET
   without accept headers.
   Description: The getcontentlength property MUST be defined on any
   DAV compliant resource that returns the Content-Length header in
   response to a GET.

   Value:      content-length ; see section 14.14 of [RFC2068]

   <!ELEMENT getcontentlength (#PCDATA) >

13.5 getcontenttype Property

   Name:       getcontenttype
   Namespace:  DAV:
   Purpose:    Contains the Content-Type header returned by a GET
   without accept headers.
   Description: This getcontenttype property MUST be defined on any DAV
   compliant resource that returns the Content-Type header in response
   to a GET.
   Value:      media-type   ; defined in section 3.7 of [RFC2068]

   <!ELEMENT getcontenttype (#PCDATA) >

13.6 getetag Property

   Name:       getetag
   Namespace:  DAV:
   Purpose:    Contains the ETag header returned by a GET without
   accept headers.
   Description: The getetag property MUST be defined on any DAV
   compliant resource that returns the Etag header.
   Value:      entity-tag  ; defined in section 3.11 of [RFC2068]

   <!ELEMENT getetag (#PCDATA) >

13.7 getlastmodified Property

   Name:       getlastmodified
   Namespace:  DAV:
   Purpose:    Contains the Last-Modified header returned by a GET
   method without accept headers.
   Description: Note that the last-modified date on a resource may
   reflect changes in any part of the state of the resource, not
   necessarily just a change to the response to the GET method.  For
   example, a change in a property may cause the last-modified date to
   change. The getlastmodified property MUST be defined on any DAV
   compliant resource that returns the Last-Modified header in response
   to a GET.
   Value:      HTTP-date  ; defined in section 3.3.1 of [RFC2068]

   <!ELEMENT getlastmodified (#PCDATA) >

13.8 lockdiscovery Property

   Name:       lockdiscovery
   Namespace:  DAV:
   Purpose:    Describes the active locks on a resource
   Description: The lockdiscovery property returns a listing of who has
   a lock, what type of lock he has, the timeout type and the time
   remaining on the timeout, and the associated lock token.  The server
   is free to withhold any or all of this information if the requesting
   principal does not have sufficient access rights to see the requested
   data.

   <!ELEMENT lockdiscovery (activelock)* >

13.8.1 Example - Retrieving the lockdiscovery Property

   >>Request

   PROPFIND /container/ HTTP/1.1
   Host: www.foo.bar
   Content-Length: xxxx
   Content-Type: text/xml; charset="utf-8"

   <?xml version="1.0" encoding="utf-8" ?>
   <D:propfind xmlns:D='DAV:'>
     <D:prop><D:lockdiscovery/></D:prop>
   </D:propfind>

   >>Response

   HTTP/1.1 207 Multi-Status
   Content-Type: text/xml; charset="utf-8"
   Content-Length: xxxx

   <?xml version="1.0" encoding="utf-8" ?>
   <D:multistatus xmlns:D='DAV:'>
     <D:response>
          <D:href>http://www.foo.bar/container/</D:href>
          <D:propstat>
               <D:prop>
                    <D:lockdiscovery>
                         <D:activelock>
                              <D:locktype><D:write/></D:locktype>
                              <D:lockscope><D:exclusive/></D:lockscope>
                              <D:depth>0</D:depth>
                              <D:owner>Jane Smith</D:owner>
                              <D:timeout>Infinite</D:timeout>
                              <D:locktoken>
                                   <D:href>
               opaquelocktoken:f81de2ad-7f3d-a1b2-4f3c-00a0c91a9d76
                                   </D:href>
                              </D:locktoken>
                         </D:activelock>
                    </D:lockdiscovery>
               </D:prop>
               <D:status>HTTP/1.1 200 OK</D:status>
          </D:propstat>
     </D:response>
   </D:multistatus>

   This resource has a single exclusive write lock on it, with an
   infinite timeout.

13.9 resourcetype Property

   Name:       resourcetype
   Namespace:  DAV:
   Purpose:    Specifies the nature of the resource.
   Description: The resourcetype property MUST be defined on all DAV
   compliant resources.  The default value is empty.

   <!ELEMENT resourcetype ANY >

13.10 source Property

   Name:       source
   Namespace:  DAV:
   Purpose:    The destination of the source link identifies the
   resource that contains the unprocessed source of the link's source.
   Description: The source of the link (src) is typically the URI of the
   output resource on which the link is defined, and there is typically
   only one destination (dst) of the link, which is the URI where the
   unprocessed source of the resource may be accessed.  When more than
   one link destination exists, this specification asserts no policy on
   ordering.

   <!ELEMENT source (link)* >

13.10.1 Example - A source Property

   <?xml version="1.0" encoding="utf-8" ?>
   <D:prop xmlns:D="DAV:" xmlns:F="http://www.foocorp.com/Project/">
     <D:source>
          <D:link>
               <F:projfiles>Source</F:projfiles>
               <D:src>http://foo.bar/program</D:src>
               <D:dst>http://foo.bar/src/main.c</D:dst>
          </D:link>
          <D:link>
               <F:projfiles>Library</F:projfiles>
               <D:src>http://foo.bar/program</D:src>
               <D:dst>http://foo.bar/src/main.lib</D:dst>
          </D:link>
          <D:link>
               <F:projfiles>Makefile</F:projfiles>
               <D:src>http://foo.bar/program</D:src>
               <D:dst>http://foo.bar/src/makefile</D:dst>
          </D:link>
     </D:source>
   </D:prop>

   In this example the resource http://foo.bar/program has a source
   property that contains three links.  Each link contains three
   elements, two of which, src and dst, are part of the DAV schema
   defined in this document, and one which is defined by the schema
   http://www.foocorp.com/project/ (Source, Library, and Makefile).  A
   client which only implements the elements in the DAV spec will not
   understand the foocorp elements and will ignore them, thus seeing the
   expected source and destination links.  An enhanced client may know
   about the foocorp elements and be able to present the user with
   additional information about the links.  This example demonstrates
   the power of XML markup, allowing element values to be enhanced
   without breaking older clients.

13.11 supportedlock Property

   Name:       supportedlock
   Namespace:  DAV:
   Purpose:    To provide a listing of the lock capabilities supported
   by the resource.
   Description: The supportedlock property of a resource returns a
   listing of the combinations of scope and access types which may be
   specified in a lock request on the resource.  Note that the actual
   contents are themselves controlled by access controls so a server is
   not required to provide information the client is not authorized to
   see.

   <!ELEMENT supportedlock (lockentry)* >

13.11.1 Example - Retrieving the supportedlock Property

   >>Request

   PROPFIND  /container/ HTTP/1.1
   Host: www.foo.bar
   Content-Length: xxxx
   Content-Type: text/xml; charset="utf-8"

   <?xml version="1.0" encoding="utf-8" ?>
   <D:propfind xmlns:D="DAV:">
     <D:prop><D:supportedlock/></D:prop>
   </D:propfind>

   >>Response

   HTTP/1.1 207 Multi-Status
   Content-Type: text/xml; charset="utf-8"
   Content-Length: xxxx

   <?xml version="1.0" encoding="utf-8" ?>
   <D:multistatus xmlns:D="DAV:">
     <D:response>
          <D:href>http://www.foo.bar/container/</D:href>
          <D:propstat>
               <D:prop>
                    <D:supportedlock>
                         <D:lockentry>
                              <D:lockscope><D:exclusive/></D:lockscope>
                              <D:locktype><D:write/></D:locktype>
                         </D:lockentry>
                         <D:lockentry>
                              <D:lockscope><D:shared/></D:lockscope>
                              <D:locktype><D:write/></D:locktype>
                         </D:lockentry>
                    </D:supportedlock>
               </D:prop>
               <D:status>HTTP/1.1 200 OK</D:status>
          </D:propstat>
     </D:response>
   </D:multistatus>

=back

=cut
    my $LiveProps = {
      'DAV:' => {
        # RFC 2518
        creationdate => $ctime, # 13.1
        displayname => $displayname, # 13.2
        # XXX - from Apache layer?
        #getcontentlanguage => undef, # 13.3
        getcontentlength => $info->FileSize(), # 13.4
        getcontenttype => $type, # 13.5
        getetag => $info->GetETag(), # 13.6
        getlastmodified => $mtime, # 13.7
        # XXX - need to write _MakeLockDiscovery
        # to return a sub which looks them up so we don't pay the cost of
        # building this property if we don't need it
        lockdiscovery => _MakeLockDiscovery($ParentNode, $name), # 13.8
        resourcetype => $info->IsDir() ? { collection => undef } : undef, # 13.9
        # XXX - we don't do "source"
        #source => undef, # 13.10
        supportedlock => _MakeSupportedLock(), # 13.11

        # quota - RFC4331
        'quota-used-bytes' => $QuotaUsed,
        'quota-available-bytes' => ($QuotaTotal - $QuotaUsed),
      },
    };

    if ($Opts->{ClientQuirks}{HasWindowsProperties}) {
      $LiveProps->{'DAV:'}{'isreadonly'} = $info->CanWrite() ? 0 : 1;
      $LiveProps->{'DAV:'}{'ishidden'} = $displayname =~ m{\A \. }xms ? 1 : 0;
      $LiveProps->{'DAV:'}{'iscollection'} = $info->IsDir() ? 1 : 0;
      $LiveProps->{'DAV:'}{'isroot'} = $info->IsRootNode() ? 1 : 0;
      $LiveProps->{'DAV:'}{'href'} = $hreftext;
      $LiveProps->{'DAV:'}{'name'} = $info->NodeName();
    }

    if ($Opts->{ClientQuirks}{HasQuota512}) {
      $LiveProps->{'DAV:'}{'quotaused'} = int($QuotaUsed / 512);
      $LiveProps->{'DAV:'}{'quota'} = int($QuotaTotal / 512);
    }

    if ($info->IsFile()) {
      unless ($reqinfo eq 'allprop' and $Opts->{ClientQuirks}{WindowsVistaAllprop}) {
        $LiveProps->{'ME:'}{'sha1hex'} = $info->Sha1Hex();
      }
    }

    my $DeadProps = $info->GetProperties();

    # find the properties
    my @wantprops;
    if ($reqinfo eq 'allprop' or $reqinfo eq 'propname') {
      foreach my $ns (keys %$LiveProps) {
        foreach my $propname (keys %{$LiveProps->{$ns}}) {
          push @wantprops, [$ns, $propname];
        }
      }
      # skip dead properties for Windows DAV
      unless ($Opts->{ClientQuirks}{WindowsVistaAllprop}) {
        foreach my $ns (keys %$DeadProps) {
          foreach my $propname (keys %{$DeadProps->{$ns}}) {
            push @wantprops, [$ns, $propname];
          }
        }
      }
    } else {
      @wantprops = @reqprops;
    }

    my @ok;
    my @missing;
    if ($reqinfo eq 'propname') {
      # empty elements, but with namespace
      @ok = map { _MakeElement($doc, @$_) } @wantprops;
    } else {
      foreach my $wantprop (@wantprops) {
        my ($ns, $name) = @$wantprop;
        if (exists $LiveProps->{$ns}->{$name}) {
          my $item = _MakeElement($doc, $ns, $name);
          _MakeNSNodes($doc, $item, $ns, $LiveProps->{$ns}->{$name});
          push @ok, $item;
        } elsif (exists $DeadProps->{$ns}->{$name}) {
          if ($DeadProps->{$ns}->{$name}->{IsXML}) { # IsXML
            my $xml = $DeadProps->{$ns}->{$name}->{PropValue};
            # workaround for broken attributes from webdrive
            if ($Opts->{ClientQuirks}{WebDriveNS}) {
              $xml =~ s{^<srtns:(srt_[a-z]+)>}{<srtns:$1 xmlns:srtns="http://www.southrivertech.com/">};
            }
            # it's not the setting, it's the getting.  FOR NOW.  Maybe hack setting
            # instead with a client quirk?
            if ($xml =~ m{^\<Z:Win32}) {
              $xml =~ s{\>}{ xmlns:Z="$ns">};
            }
            my $subdoc = $PARSER->parse_balanced_chunk($xml);
            my $item = $subdoc->cloneNode(1);
            $doc->adoptNode($item);
            push @ok, $item;
          } else {
            my $item = _MakeElement($doc, $ns, $name);
            _MakeNSNodes($doc, $item, $ns, $DeadProps->{$ns}->{$name}->{PropValue});
            push @ok, $item;
          }
        } else {
          my $item = _MakeElement($doc, $ns, $name);
          push @missing, $item;
        }
      }
    }
    my @propstat;
    if (@ok) {
      push @propstat, {
        status => 'HTTP/1.1 200 OK',
        prop => [\@ok],
      };
    }
    if (@missing and $reqinfo ne 'allprop') {
      push @propstat, {
        status => 'HTTP/1.1 404 Not Found',
        prop => [\@missing],
      };
    }
    push @response, {
      href => $hreftext,
      propstat => \@propstat,
    }
  }

  _MakeNSNodes($doc, $multistat, 'DAV:', { response => \@response });

  my $Body = qq{<?xml version="1.0" encoding="utf-8"?>\n} .
             $doc->documentElement->toString(1);

  $Opts->{body_out} = $Body;

  $R->status(207);
  $R->set_content_length(do { use bytes; length($Body) });
  $R->header_out('Content-Location' => _EscapeURI("$Opts->{Base}$Path"));
  $R->header_out('Pragma' => 'no-cache');
  $R->header_out('Cache-Control' => 'no-cache');
  $R->content_type('text/xml; charset=utf-8');
  $R->print($Body);

  return OK;
}

=head2 8.2 PROPPATCH

   The PROPPATCH method processes instructions specified in the request
   body to set and/or remove properties defined on the resource
   identified by the Request-URI.

   All DAV compliant resources MUST support the PROPPATCH method and
   MUST process instructions that are specified using the
   propertyupdate, set, and remove XML elements of the DAV schema.
   Execution of the directives in this method is, of course, subject to
   access control constraints.  DAV compliant resources SHOULD support
   the setting of arbitrary dead properties.

   The request message body of a PROPPATCH method MUST contain the
   propertyupdate XML element.  Instruction processing MUST occur in the
   order instructions are received (i.e., from top to bottom).
   Instructions MUST either all be executed or none executed. Thus if
   any error occurs during processing all executed instructions MUST be
   undone and a proper error result returned. Instruction processing
   details can be found in the definition of the set and remove
   instructions in section 12.13.

=item 8.2.1 Status Codes for use with 207 (Multi-Status)

   The following are examples of response codes one would expect to be
   used in a 207 (Multi-Status) response for this method.  Note,
   however, that unless explicitly prohibited any 2/3/4/5xx series
   response code may be used in a 207 (Multi-Status) response.

   200 (OK) - The command succeeded.  As there can be a mixture of sets
   and removes in a body, a 201 (Created) seems inappropriate.

   403 (Forbidden) - The client, for reasons the server chooses not to
   specify, cannot alter one of the properties.

   409 (Conflict) - The client has provided a value whose semantics are
   not appropriate for the property.  This includes trying to set read-
   only properties.

   423 (Locked) - The specified resource is locked and the client either
   is not a lock owner or the lock type requires a lock token to be
   submitted and the client did not submit it.

   507 (Insufficient Storage) - The server did not have sufficient space
   to record the property.

=item 8.2.2 Example - PROPPATCH

   >>Request

   PROPPATCH /bar.html HTTP/1.1
   Host: www.foo.com
   Content-Type: text/xml; charset="utf-8"
   Content-Length: xxxx

   <?xml version="1.0" encoding="utf-8" ?>
   <D:propertyupdate xmlns:D="DAV:"
   xmlns:Z="http://www.w3.com/standards/z39.50/">
     <D:set>
          <D:prop>
               <Z:authors>
                    <Z:Author>Jim Whitehead</Z:Author>
                    <Z:Author>Roy Fielding</Z:Author>
               </Z:authors>
          </D:prop>
     </D:set>
     <D:remove>
          <D:prop><Z:Copyright-Owner/></D:prop>
     </D:remove>
   </D:propertyupdate>

   >>Response

   HTTP/1.1 207 Multi-Status
   Content-Type: text/xml; charset="utf-8"
   Content-Length: xxxx

   <?xml version="1.0" encoding="utf-8" ?>
   <D:multistatus xmlns:D="DAV:"
   xmlns:Z="http://www.w3.com/standards/z39.50">
     <D:response>
          <D:href>http://www.foo.com/bar.html</D:href>
          <D:propstat>
               <D:prop><Z:Authors/></D:prop>
               <D:status>HTTP/1.1 424 Failed Dependency</D:status>
          </D:propstat>
          <D:propstat>
               <D:prop><Z:Copyright-Owner/></D:prop>
               <D:status>HTTP/1.1 409 Conflict</D:status>
          </D:propstat>
          <D:responsedescription> Copyright Owner can not be deleted or
   altered.</D:responsedescription>
     </D:response>
   </D:multistatus>

   In this example, the client requests the server to set the value of
   the http://www.w3.com/standards/z39.50/Authors property, and to
   remove the property http://www.w3.com/standards/z39.50/Copyright-
   Owner.  Since the Copyright-Owner property could not be removed, no
   property modifications occur.  The 424 (Failed Dependency) status
   code for the Authors property indicates this action would have
   succeeded if it were not for the conflict with removing the
   Copyright-Owner property.

=cut
sub PROPPATCH_handler {
  my($R, $Opts) = @_;
  my $Path = $Opts->{Path};
  my $FS = $Opts->{FS};

  return 403 if $FS->IsReadOnly();

  my ($ParentNode, $name) = eval { $FS->ParentNodeAndName($Path) };
  return 404 if $@ || !$ParentNode;

  my @locks = _ParseIfHeader($R->header_in('If'));
  return 423 if $ParentNode->NameLocked($name, \@locks, 'w');

  my $FileInfo = $ParentNode->ChildByName($name);
  return 404 unless $FileInfo;

  my $Depth = _ParseDepthHeader($R->header_in('Depth'));

  my @props;
  my $willfail = 0;
  if (my $len = $R->header_in('Content-length')) {
    my $content;
    $R->read($content, $len);
    $Opts->{body_in} = $content;
    my $doc;
    eval {
      $doc = $PARSER->parse_string($content);
    };
    if ($@ or not $doc or _HasNULLNamespace($doc)) {
      return 400;
    }

    foreach my $node (_GetNSNodes($doc, 'DAV:', 'propertyupdate', ['set', 'remove'], 'prop')) {
      my $type = $node->parentNode->parentNode->localname;

      # Litmus: props.3 says no prefix with a blank namespace please!
      if ($node->namespaceURI and $node->namespaceURI eq 'DAV:') {
        $willfail = 1 unless ($type eq 'set' and $node->localname() eq 'getlastmodified');
      }

      if ($type eq 'remove') {
        push @props, [$node->namespaceURI || '', $node->localname];
      }
      else {
        push @props, [$node->namespaceURI || '', $node->localname, $node];
      }
    }
  }

  unless (@props) {
    return 409;
  }

  my %props;
  my $Modified;
  foreach my $prop (@props) {
    if ($prop->[0] eq 'DAV:') {
      if ($prop->[1] eq 'getlastmodified') {
        # time in ISO YYYY-MM-DDTHH:MM:SSZ format
        my $TimeStr = $prop->[2]->textContent();
        my $UnixTime = HTTP::Date::str2time($TimeStr);
        $Modified = $UnixTime;
      }
    } else {
      if ($Opts->{ClientQuirks}{WebDriveNS}) {
        if ($prop->[0] eq 'http://www.southrivertech.com/') {
          if ($prop->[1] eq 'srt_modifiedtime') {
            # Let's just define our own property rather than use the standard.
            # Yeah, that sounds like a great idea.
            my $TimeStr = $prop->[2]->textContent();
            my $UnixTime = HTTP::Date::str2time($TimeStr);
            $Modified = $UnixTime;
            # don't skip out, we want to store the WebDrive version too.
          }
        }
      } elsif ($Opts->{ClientQuirks}{HasWindowsProperties}) {
        if ($prop->[0] eq 'urn:schemas-microsoft-com:') {
          if ($prop->[1] eq 'Win32LastModifiedTime') {
            my $TimeStr = $prop->[2]->textContent();
            my $UnixTime = HTTP::Date::str2time($TimeStr);
            $Modified = $UnixTime;
          }
        }
      }
      $props{$prop->[0]}{$prop->[1]} = defined $prop->[2] ? {PropValue => $prop->[2]->toString(), IsXML => 1} : undef;
    }
  }
  if (keys %props) {
    $FileInfo->SetProperties(\%props, IfLocks => \@locks);
  }
  if (defined $Modified) {
    $FileInfo->SetModified($Modified, IfLocks => \@locks);
  }

  my $doc = XML::LibXML::Document->new('1.0', 'utf-8');
  my $multistat = $doc->createElement('multistatus');
  $multistat->setNamespace('DAV:', 'D');
  my $i = 0;
  $doc->setDocumentElement($multistat);

  my $resp = _MakeElement($doc, 'DAV:', 'response');
  $multistat->appendChild($resp);
  my $href = _MakeElement($doc, 'DAV:', 'href');
  my $hreftext = "$Opts->{Base}$Path";
  $href->appendText($hreftext);
  $resp->appendChild($href);

  foreach my $item (@props) {
    my ($ns, $name, $val) = @$item; # blank for del obviously
    my $propstat = _MakeElement($doc, 'DAV:', 'propstat');
    my $status = _MakeElement($doc, 'DAV:', 'status');
    my $prop = _MakeElement($doc, 'DAV:', 'prop');
    my $s = 'HTTP/1.1 200 OK';
    if ($willfail) {
      if ($ns eq 'DAV:') {
        $s = 'HTTP/1.1 409 Conflict';
      } else {
        $s = 'HTTP/1.1 424 Failed Dependency';
      }
    }
    $status->appendText($s);

    my $propname = _MakeElement($doc, $ns, $name);
    $prop->appendChild($propname);

    $propstat->appendChild($status);
    $propstat->appendChild($prop);
    $resp->appendChild($propstat);
  }

  # can't just use $doc->toString() because it breaks on encoded
  # entities for some reason.  Crazy XML::LibXML
  my $Body = qq{<?xml version="1.0" encoding="utf-8"?>\n} .
             $doc->documentElement->toString(1);

  $Opts->{body_out} = $Body;

  $R->status(207);
  $R->header_out('Content-Location' => _EscapeURI("$Opts->{Base}$Path"));
  $R->header_out('Pragma' => 'no-cache');
  $R->header_out('Cache-Control' => 'no-cache');
  $R->set_content_length(do { use bytes; length($Body) });
  $R->content_type('text/xml; charset=utf-8');
  $R->print($Body);

  return OK;
}

=head2 8.3 MKCOL Method

   The MKCOL method is used to create a new collection. All DAV
   compliant resources MUST support the MKCOL method.

=item 8.3.1 Request

   MKCOL creates a new collection resource at the location specified by
   the Request-URI.  If the resource identified by the Request-URI is
   non-null then the MKCOL MUST fail.  During MKCOL processing, a server
   MUST make the Request-URI a member of its parent collection, unless
   the Request-URI is "/".  If no such ancestor exists, the method MUST
   fail.  When the MKCOL operation creates a new collection resource,
   all ancestors MUST already exist, or the method MUST fail with a 409
   (Conflict) status code.  For example, if a request to create
   collection /a/b/c/d/ is made, and neither /a/b/ nor /a/b/c/ exists,
   the request must fail.

   When MKCOL is invoked without a request body, the newly created
   collection SHOULD have no members.

   A MKCOL request message may contain a message body.  The behavior of
   a MKCOL request when the body is present is limited to creating
   collections, members of a collection, bodies of members and
   properties on the collections or members.  If the server receives a
   MKCOL request entity type it does not support or understand it MUST
   respond with a 415 (Unsupported Media Type) status code.  The exact
   behavior of MKCOL for various request media types is undefined in
   this document, and will be specified in separate documents.

=item 8.3.2 Status Codes

   Responses from a MKCOL request MUST NOT be cached as MKCOL has non-
   idempotent semantics.

   201 (Created) - The collection or structured resource was created in
   its entirety.

   403 (Forbidden) - This indicates at least one of two conditions: 1)
   the server does not allow the creation of collections at the given
   location in its namespace, or 2) the parent collection of the
   Request-URI exists but cannot accept members.

   405 (Method Not Allowed) - MKCOL can only be executed on a
   deleted/non-existent resource.

   409 (Conflict) - A collection cannot be made at the Request-URI until
   one or more intermediate collections have been created.

   415 (Unsupported Media Type)- The server does not support the request
   type of the body.

   507 (Insufficient Storage) - The resource does not have sufficient
   space to record the state of the resource after the execution of this
   method.

=item 8.3.3 Example - MKCOL

   This example creates a collection called /webdisc/xfiles/ on the
   server www.server.org.

   >>Request

   MKCOL /webdisc/xfiles/ HTTP/1.1
   Host: www.server.org

   >>Response

   HTTP/1.1 201 Created

=cut
sub MKCOL_handler {
  my($R, $Opts) = @_;
  my $Path = $Opts->{Path};
  my $FS = $Opts->{FS};

  return 403 if $FS->IsReadOnly();

  # we don't know how to create anything with a body of
  # details, so if there's any body then fail
  if ($R->header_in('Content-length')) {
    return 415;
  }

  my ($ParentNode, $name) = eval { $FS->ParentNodeAndName($Path) };
  # can't create something if the parent doesn't exist!
  return 409 if $@ || !$ParentNode;

  my @locks = _ParseIfHeader($R->header_in('If'));
  return 423 if $ParentNode->NameLocked($name, \@locks, 'w');

  my $FileInfo = $ParentNode->ChildByName($name);
  return 405 if $FileInfo;

  # if we can't write then that's a problem
  return 403 if !$ParentNode->CanWrite();

  # we don't get 'out of disk space' for directories, they're
  # free in our system.  It's also never forbidden to create
  # directories within other directories, namespace
  # permitting...
  eval {
    my $FileInfo = $ParentNode->Mkdir($name);
    if (my $modified = $R->header_in('Last-Modified')) {
      $FileInfo->SetModified(HTTP::Date::str2time($modified));
    }
  };

  # any error we haven't handled above
  if ($@) {
    my $Error = $@;
    warn "MKCOL ERROR $Error";
    return 409;
  }

  # ok, created fine
  $R->status(201);
  $R->header_out('Location' => _EscapeURI("$Opts->{Base}$Path"));
  $R->set_content_length(0);
  $R->header_out('Pragma' => 'no-cache');
  $R->header_out('Cache-Control' => 'no-cache');
  $R->content_type('application/octet-stream');
  return OK;
}

=head2 8.4 GET, HEAD for Collections

   The semantics of GET are unchanged when applied to a collection,
   since GET is defined as, "retrieve whatever information (in the form
   of an entity) is identified by the Request-URI" [RFC2068].  GET when
   applied to a collection may return the contents of an "index.html"
   resource, a human-readable view of the contents of the collection, or
   something else altogether. Hence it is possible that the result of a
   GET on a collection will bear no correlation to the membership of the
   collection.

   Similarly, since the definition of HEAD is a GET without a response
   message body, the semantics of HEAD are unmodified when applied to
   collection resources.

=cut
sub GET_handler {
  my($R, $Opts, $IsHEAD) = @_;
  my $Path = $Opts->{Path};
  my $FS = $Opts->{FS};

  my ($ParentNode, $name) = eval { $FS->ParentNodeAndName($Path) };
  return 404 if $@ || !$ParentNode;

  my $FileInfo = $ParentNode->ChildByName($name);
  return 404 unless $FileInfo;

  # support caches
  if (my $ims = $R->header_in('If-Modified-Since')) {
    # if str2time fails then always send contents!
    return 304 if eval { (HTTP::Date::str2time($ims) || 0) >= ($FileInfo->Modified() || 0) };
  }

  # note - we're not handling If-None-Match on not-gets!
  if (my $inm = $R->header_in('If-None-Match')) {
    my $etag = $FileInfo->GetETag();
    foreach my $tag (split /,\s*/, $inm) {
      return 304 if $tag eq $etag;
    }
  }

  my $IP = $R->ip();
  my $RD = {
    OrigHost => 'dav.messagingengine.com',
    Domain => $Opts->{User}->Brand()->{PrimaryDomain},
    Method => 'GET',
    Uri => $Path,
    IP => $IP,
    User => $Opts->{User},
    SSL => $Opts->{SSL},
  };

  # Just send the file or head details
  if ($FileInfo->IsFile()) {

    # Rate Checking
    my $RateCheckResult = $FS->RateCheck(0, "dav.messagingengine.com/$Path", IP => $IP);
    return 413 if $RateCheckResult == 0;

    my $Size = $FileInfo->FileSize();

    # Check for a range request (only simply range requests supported)
    #  Would be nice to pass this to apache to deal with, but need to track BW
    my ($Len, $Start, $End);
    if (my $Range = $R->header_in("Range")) {
      ($Start, $End) = $Range =~ /^bytes[= ](\d*)-(\d*)(?:,|$)/
        or return 501;

      return 416 if $Start eq '';

      # Start and End are inclusive byte values
      $End = $Size-1 if $End eq '' || $End >= $Size;

      if ($Start > $End) {
        return 416;
      } else {
        $R->status(206);
        $R->header_out("Content-Range", "bytes $Start-$End/$Size");
        $R->set_content_length($Len = $End - $Start + 1);
        $RD->{Range} = [ $Start, $End, $Size ];
      }

    # No range request, send entire response
    } else {
      $R->set_content_length($Size);
    }

    $R->header_out('Last-Modified' => HTTP::Date::time2str($FileInfo->Modified()));
    $R->header_out('ETag' => $FileInfo->GetETag());
    $R->header_out('Cache-Control' => 'private');
    $R->header_out('X-Rate-User', scalar($FS->User->UnqName));

    my $Type = $FileInfo->MimeType();
    $R->content_type($Type);

    # Send content
    if (!$IsHEAD) {
      if (my $fh = eval { $FileInfo->Open() }) {

        if (defined $Start) {
          my $Rem = $Len;
          $fh->seek($Start, 0);
          while ($Rem) {
            $fh->read(my $Buf, my $Amount = $Rem > 4096 ? 4096 : $Rem);
            $R->print($Buf);
            $Rem -= $Amount;
          }
        } else {
          $R->send_fd($fh);
        }

        $fh->close();

      # errors worthy of an ugly failure mode!
      } else {
        warn "Failed to load file for " . $FileInfo->Path();
        return 500; # internal server error!
      }
    }

    return OK;
  }

  # use the same listing code as the HTTP version
  if ($FileInfo->IsDir()) {
    my $ThisPath = escape_html("$Opts->{Base}$Opts->{Path}", 0);

    my $Renderer = ME::VFS::Interface::HTTP::DirListingRenderer->new(
      R => $R,
      Node => $FileInfo,
      ThisPath => $ThisPath,
      SSL => $Opts->{SSL},
      RD => $RD,
      Method => 'GET',
      Domain => 'messagingengine.com',
    );
    return ME::VFS::Interface::HTTP::Render($R, $RD, $Renderer, $FS);
  }

  # should never get here!
  return 500;
}

=head2 8.5 POST for Collections

Since by definition the actual function performed by POST is
determined by the server and often depends on the particular
resource, the behavior of POST when applied to collections cannot be
meaningfully modified because it is largely undefined.  Thus the
semantics of POST are unmodified when applied to a collection.

=cut
# POST not implemented

=head2 8.6 DELETE

=item 8.6.1 DELETE for Non-Collection Resources

   If the DELETE method is issued to a non-collection resource whose
   URIs are an internal member of one or more collections, then during
   DELETE processing a server MUST remove any URI for the resource
   identified by the Request-URI from collections which contain it as a
   member.

=item 8.6.2 DELETE for Collections

   The DELETE method on a collection MUST act as if a "Depth: infinity"
   header was used on it.  A client MUST NOT submit a Depth header with
   a DELETE on a collection with any value but infinity.

   DELETE instructs that the collection specified in the Request-URI and
   all resources identified by its internal member URIs are to be
   deleted.

   If any resource identified by a member URI cannot be deleted then all
   of the member's ancestors MUST NOT be deleted, so as to maintain
   namespace consistency.

   Any headers included with DELETE MUST be applied in processing every
   resource to be deleted.

   When the DELETE method has completed processing it MUST result in a
   consistent namespace.

   If an error occurs with a resource other than the resource identified
   in the Request-URI then the response MUST be a 207 (Multi-Status).
   424 (Failed Dependency) errors SHOULD NOT be in the 207 (Multi-
   Status).  They can be safely left out because the client will know
   that the ancestors of a resource could not be deleted when the client
   receives an error for the ancestor's progeny.  Additionally 204 (No
   Content) errors SHOULD NOT be returned in the 207 (Multi-Status).
   The reason for this prohibition is that 204 (No Content) is the
   default success code.

=item 8.6.2.1 Example - DELETE

   >>Request

   DELETE  /container/ HTTP/1.1
   Host: www.foo.bar

   >>Response

   HTTP/1.1 207 Multi-Status
   Content-Type: text/xml; charset="utf-8"
   Content-Length: xxxx

   <?xml version="1.0" encoding="utf-8" ?>
   <d:multistatus xmlns:d="DAV:">
     <d:response>
          <d:href>http://www.foo.bar/container/resource3</d:href>
          <d:status>HTTP/1.1 423 Locked</d:status>
     </d:response>
   </d:multistatus>

   In this example the attempt to delete
   http://www.foo.bar/container/resource3 failed because it is locked,
   and no lock token was submitted with the request. Consequently, the
   attempt to delete http://www.foo.bar/container/ also failed. Thus the
   client knows that the attempt to delete http://www.foo.bar/container/
   must have also failed since the parent can not be deleted unless its
   child has also been deleted.  Even though a Depth header has not been
   included, a depth of infinity is assumed because the method is on a
   collection.

=cut
sub DELETE_handler {
  my($R, $Opts) = @_;
  my $Path = $Opts->{Path};
  my $FS = $Opts->{FS};

  return 403 if $FS->IsReadOnly();

  my ($ParentNode, $name) = eval { $FS->ParentNodeAndName($Path) };
  return 404 if $@ || !$ParentNode;

  my @locks = _ParseIfHeader($R->header_in('If'));
  return 423 if $ParentNode->NameLocked($name, \@locks, 'w');

  my $FileInfo = $ParentNode->ChildByName($name);
  return 404 unless $FileInfo;

  my $Depth = _ParseDepthHeader($R->header_in('Depth'));
  # for files, a depth of zero is created by lftp to avoid deleting directories by mistake
  if ($FileInfo->IsDir()) {
    return 400 unless $Depth eq 'infinity';
  }

  return 403 if !$ParentNode->CanWrite();

  eval {
    $FileInfo->Delete(IfLocks => \@locks);
  };

  if ($@) {
    warn "DELETE ERROR: $@";
    return 409;
  }

  $R->status(204);
  $R->header_out('Location' => _EscapeURI("$Opts->{Base}$Path"));
  $R->header_out('Pragma' => 'no-cache');
  $R->header_out('Cache-Control' => 'no-cache');
  $R->set_content_length(0);
  $R->content_type('application/octet-stream');

  return OK;
}

=head2 8.7 PUT

=item 8.7.1 PUT for Non-Collection Resources

   A PUT performed on an existing resource replaces the GET response
   entity of the resource.  Properties defined on the resource may be
   recomputed during PUT processing but are not otherwise affected.  For
   example, if a server recognizes the content type of the request body,
   it may be able to automatically extract information that could be
   profitably exposed as properties.

   A PUT that would result in the creation of a resource without an
   appropriately scoped parent collection MUST fail with a 409
   (Conflict).

   As defined in the HTTP/1.1 specification [RFC2068], the "PUT method
   requests that the enclosed entity be stored under the supplied
   Request-URI."  Since submission of an entity representing a
   collection would implicitly encode creation and deletion of
   resources, this specification intentionally does not define a
   transmission format for creating a collection using PUT.  Instead,
   the MKCOL method is defined to create collections.

   When the PUT operation creates a new non-collection resource all
   ancestors MUST already exist.  If all ancestors do not exist, the
   method MUST fail with a 409 (Conflict) status code.  For example, if
   resource /a/b/c/d.html is to be created and /a/b/c/ does not exist,
   then the request must fail.

=cut
sub PUT_handler {
  my($R, $Opts) = @_;
  my $Path = $Opts->{Path};
  my $FS = $Opts->{FS};

  my $RateCheckResult = $FS->RateCheck(0, $Path);
  return 413 if $RateCheckResult == 0;

  return 403 if $FS->IsReadOnly();

  my ($ParentNode, $name) = eval { $FS->ParentNodeAndName($Path) };
  return 409 if $@ || !$ParentNode;

  my @locks = _ParseIfHeader($R->header_in('If'));
  return 423 if $ParentNode->NameLocked($name, \@locks, 'w');

  return 403 unless $ParentNode->CanWrite(); # will get error

  my $FileInfo = $ParentNode->ChildByName($name);

  my $RC = $FileInfo ? 204 : 201;

  eval {
    my $HeaderLen = $R->header_in('Content-length') || 0;
    my $Len = 0;
    my $Buf;
    my $FH = File::Temp::tempfile();

    # read data into tempfile
    while(my $Got = $R->read($Buf, 8192)) {
      $FH->print($Buf);
      $Len += $Got;
    }
    $FH->flush();
    $FH->seek(0, 0);

    # move tempfile into filestorage
    my $FileInfo = $ParentNode->ImportFile($FH, $name, Overwrite => 1, IfLocks => \@locks, SkipSizeCheck => 1);
    if (my $modified = $R->header_in('Last-Modified')) {
      $FileInfo->SetModified(HTTP::Date::str2time($modified));
    }

    $FH->close();

    # could die on close, so don't log quota until file safely on disk
    my $RateUpdateResult = $FS->RateUpdate($Len, $FileInfo->Path(), ForceFlag => 1, Source => 'dav') if $Len;
  };

  if (my $Err = $@) {
    # is this defined in the spec?  No, doesn't seem so.  But it should
    # be the same as COPY surely.
    return 507 if $Err =~ m/Can't go over quota/;
    warn "PUT ERROR $Err";
    return 409;
  }

  $R->status($RC);
  $R->header_out('Location' => _EscapeURI("$Opts->{Base}$Path"));
  $R->header_out('Pragma' => 'no-cache');
  $R->header_out('Cache-Control' => 'no-cache');
  $R->set_content_length(0);
  # MS server does this
  $R->header_out('Allow' => join(', ', sort keys %methods));
  $R->content_type('application/octet-stream');

  return OK;
}

=head2 8.8 COPY Method

   The COPY method creates a duplicate of the source resource,
   identified by the Request-URI, in the destination resource,
   identified by the URI in the Destination header.  The Destination
   header MUST be present.  The exact behavior of the COPY method
   depends on the type of the source resource.

   All WebDAV compliant resources MUST support the COPY method.
   However, support for the COPY method does not guarantee the ability
   to copy a resource. For example, separate programs may control
   resources on the same server.  As a result, it may not be possible to
   copy a resource to a location that appears to be on the same server.

=item 8.8.1 COPY for HTTP/1.1 resources

   When the source resource is not a collection the result of the COPY
   method is the creation of a new resource at the destination whose
   state and behavior match that of the source resource as closely as
   possible.  After a successful COPY invocation, all properties on the
   source resource MUST be duplicated on the destination resource,
   subject to modifying headers and XML elements, following the
   definition for copying properties.  Since the environment at the
   destination may be different than at the source due to factors
   outside the scope of control of the server, such as the absence of
   resources required for correct operation, it may not be possible to
   completely duplicate the behavior of the resource at the destination.
   Subsequent alterations to the destination resource will not modify
   the source resource.  Subsequent alterations to the source resource
   will not modify the destination resource.

=item 8.8.2. COPY for Properties

   The following section defines how properties on a resource are
   handled during a COPY operation.

   Live properties SHOULD be duplicated as identically behaving live
   properties at the destination resource.  If a property cannot be
   copied live, then its value MUST be duplicated, octet-for-octet, in
   an identically named, dead property on the destination resource
   subject to the effects of the propertybehavior XML element.

   The propertybehavior XML element can specify that properties are
   copied on best effort, that all live properties must be successfully
   copied or the method must fail, or that a specified list of live
   properties must be successfully copied or the method must fail. The
   propertybehavior XML element is defined in section 12.12.

=item 8.8.3 COPY for Collections

   The COPY method on a collection without a Depth header MUST act as if
   a Depth header with value "infinity" was included.  A client may
   submit a Depth header on a COPY on a collection with a value of "0"
   or "infinity".  DAV compliant servers MUST support the "0" and
   "infinity" Depth header behaviors.

   A COPY of depth infinity instructs that the collection resource
   identified by the Request-URI is to be copied to the location
   identified by the URI in the Destination header, and all its internal
   member resources are to be copied to a location relative to it,
   recursively through all levels of the collection hierarchy.

   A COPY of "Depth: 0" only instructs that the collection and its
   properties but not resources identified by its internal member URIs,
   are to be copied.

   Any headers included with a COPY MUST be applied in processing every
   resource to be copied with the exception of the Destination header.

   The Destination header only specifies the destination URI for the
   Request-URI. When applied to members of the collection identified by
   the Request-URI the value of Destination is to be modified to reflect
   the current location in the hierarchy.  So, if the Request- URI is
   /a/ with Host header value http://fun.com/ and the Destination is
   http://fun.com/b/ then when http://fun.com/a/c/d is processed it must
   use a Destination of http://fun.com/b/c/d.

   When the COPY method has completed processing it MUST have created a
   consistent namespace at the destination (see section 5.1 for the
   definition of namespace consistency).  However, if an error occurs
   while copying an internal collection, the server MUST NOT copy any
   resources identified by members of this collection (i.e., the server
   must skip this subtree), as this would create an inconsistent
   namespace. After detecting an error, the COPY operation SHOULD try to
   finish as much of the original copy operation as possible (i.e., the
   server should still attempt to copy other subtrees and their members,
   that are not descendents of an error-causing collection).  So, for
   example, if an infinite depth copy operation is performed on
   collection /a/, which contains collections /a/b/ and /a/c/, and an
   error occurs copying /a/b/, an attempt should still be made to copy
   /a/c/. Similarly, after encountering an error copying a non-
   collection resource as part of an infinite depth copy, the server
   SHOULD try to finish as much of the original copy operation as
   possible.

   If an error in executing the COPY method occurs with a resource other
   than the resource identified in the Request-URI then the response
   MUST be a 207 (Multi-Status).

   The 424 (Failed Dependency) status code SHOULD NOT be returned in the
   207 (Multi-Status) response from a COPY method.  These responses can
   be safely omitted because the client will know that the progeny of a
   resource could not be copied when the client receives an error for
   the parent.  Additionally 201 (Created)/204 (No Content) status codes
   SHOULD NOT be returned as values in 207 (Multi-Status) responses from
   COPY methods.  They, too, can be safely omitted because they are the
   default success codes.

=item 8.8.4 COPY and the Overwrite Header

   If a resource exists at the destination and the Overwrite header is
   "T" then prior to performing the copy the server MUST perform a
   DELETE with "Depth: infinity" on the destination resource.  If the
   Overwrite header is set to "F" then the operation will fail.

=item 8.8.5 Status Codes

   201 (Created) - The source resource was successfully copied.  The
   copy operation resulted in the creation of a new resource.

   204 (No Content) - The source resource was successfully copied to a
   pre-existing destination resource.

   403 (Forbidden) _ The source and destination URIs are the same.

   409 (Conflict) _ A resource cannot be created at the destination
   until one or more intermediate collections have been created.

   412 (Precondition Failed) - The server was unable to maintain the
   liveness of the properties listed in the propertybehavior XML element
   or the Overwrite header is "F" and the state of the destination
   resource is non-null.

   423 (Locked) - The destination resource was locked.

   502 (Bad Gateway) - This may occur when the destination is on another
   server and the destination server refuses to accept the resource.

   507 (Insufficient Storage) - The destination resource does not have
   sufficient space to record the state of the resource after the
   execution of this method.

=item 8.8.6 Example - COPY with Overwrite

   This example shows resource
   http://www.ics.uci.edu/~fielding/index.html being copied to the
   location http://www.ics.uci.edu/users/f/fielding/index.html.  The 204
   (No Content) status code indicates the existing resource at the
   destination was overwritten.

   >>Request

   COPY /~fielding/index.html HTTP/1.1
   Host: www.ics.uci.edu
   Destination: http://www.ics.uci.edu/users/f/fielding/index.html

   >>Response

   HTTP/1.1 204 No Content

=item 8.8.7 Example - COPY with No Overwrite

   The following example shows the same copy operation being performed,
   but with the Overwrite header set to "F."  A response of 412
   (Precondition Failed) is returned because the destination resource
   has a non-null state.

   >>Request

   COPY /~fielding/index.html HTTP/1.1
   Host: www.ics.uci.edu
   Destination: http://www.ics.uci.edu/users/f/fielding/index.html
   Overwrite: F

   >>Response

   HTTP/1.1 412 Precondition Failed

=item 8.8.8 Example - COPY of a Collection

      >>Request

      COPY /container/ HTTP/1.1
      Host: www.foo.bar
      Destination: http://www.foo.bar/othercontainer/
      Depth: infinity
      Content-Type: text/xml; charset="utf-8"
      Content-Length: xxxx

      <?xml version="1.0" encoding="utf-8" ?>
      <d:propertybehavior xmlns:d="DAV:">
        <d:keepalive>*</d:keepalive>
      </d:propertybehavior>

      >>Response

      HTTP/1.1 207 Multi-Status
      Content-Type: text/xml; charset="utf-8"
      Content-Length: xxxx

      <?xml version="1.0" encoding="utf-8" ?>
      <d:multistatus xmlns:d="DAV:">
        <d:response>
             <d:href>http://www.foo.bar/othercontainer/R2/</d:href>
             <d:status>HTTP/1.1 412 Precondition Failed</d:status>
        </d:response>
      </d:multistatus>

   The Depth header is unnecessary as the default behavior of COPY on a
   collection is to act as if a "Depth: infinity" header had been
   submitted.  In this example most of the resources, along with the
   collection, were copied successfully. However the collection R2
   failed, most likely due to a problem with maintaining the liveness of
   properties (this is specified by the propertybehavior XML element).
   Because there was an error copying R2, none of R2's members were
   copied.  However no errors were listed for those members due to the
   error minimization rules given in section 8.8.3.

=cut
sub COPY_handler {
  my($R, $Opts) = @_;
  my $Path = $Opts->{Path};
  my $FS = $Opts->{FS};

  return 403 if $FS->IsReadOnly();

  my ($ParentNode, $name) = eval { $FS->ParentNodeAndName($Path) };
  return 404 if $@ || !$ParentNode;

  my $FileInfo = $ParentNode->ChildByName($name);
  return 404 if !$FileInfo;

  my $Depth     = $R->header_in('Depth');
  my $Overwrite = $R->header_in('Overwrite');
  if (!defined($Overwrite)) {
    $Overwrite = 'T';
  }

  my $Destination = $Opts->{Destination};
  my ($DestDir, $destname) = eval { $FS->ParentNodeAndName($Destination) };
  return 409 if $@ || !$DestDir;

  my @locks = _ParseIfHeader($R->header_in('If'));
  return 423 if $DestDir->NameLocked($destname, \@locks, 'w');

  return 403 if !$DestDir->CanWrite();

  my $DestFileInfo = $DestDir->ChildByName($destname);
  # notice if it's the same file
  return 403 if $DestFileInfo
             && $DestFileInfo->NodeId() == $FileInfo->NodeId()
             ;

  return 412 if ($DestFileInfo and $Overwrite ne 'T'); # conflict

  my $RC = $DestFileInfo ? 204 : 201; # 'No Content' or 'Created'

  eval {
    if (defined $Depth and $Depth eq '0' and $FileInfo->IsDir()) {
      $DestDir->Mkdir($destname, IfLocks => \@locks);
    } else {
      $FileInfo->CopyTo(
        ParentNode => $DestDir,
        NodeName => $destname,
        Overwrite => ($Overwrite eq 'T'),
        IfLocks => \@locks,
      );
    }
  };

  if (my $Err = $@) {
    return 507 if $Err =~ m/Can't go over quota/;
    warn "COPY ERROR $Err";
    return 409;
  }

  $R->status($RC);
  $R->set_content_length(0);
  $R->header_out('Pragma' => 'no-cache');
  $R->header_out('Cache-Control' => 'no-cache');
  $R->content_type('application/octet-stream');

  return OK;
}

=head2 8.9 MOVE Method

   The MOVE operation on a non-collection resource is the logical
   equivalent of a copy (COPY), followed by consistency maintenance
   processing, followed by a delete of the source, where all three
   actions are performed atomically.  The consistency maintenance step
   allows the server to perform updates caused by the move, such as
   updating all URIs other than the Request-URI which identify the
   source resource, to point to the new destination resource.
   Consequently, the Destination header MUST be present on all MOVE
   methods and MUST follow all COPY requirements for the COPY part of
   the MOVE method.  All DAV compliant resources MUST support the MOVE
   method.  However, support for the MOVE method does not guarantee the
   ability to move a resource to a particular destination.

   For example, separate programs may actually control different sets of
   resources on the same server.  Therefore, it may not be possible to
   move a resource within a namespace that appears to belong to the same
   server.

   If a resource exists at the destination, the destination resource
   will be DELETEd as a side-effect of the MOVE operation, subject to
   the restrictions of the Overwrite header.

=item 8.9.1 MOVE for Properties

   The behavior of properties on a MOVE, including the effects of the
   propertybehavior XML element, MUST be the same as specified in
   section 8.8.2.

=item 8.9.2 MOVE for Collections

   A MOVE with "Depth: infinity" instructs that the collection
   identified by the Request-URI be moved to the URI specified in the
   Destination header, and all resources identified by its internal
   member URIs are to be moved to locations relative to it, recursively
   through all levels of the collection hierarchy.

   The MOVE method on a collection MUST act as if a "Depth: infinity"
   header was used on it.  A client MUST NOT submit a Depth header on a
   MOVE on a collection with any value but "infinity".

   Any headers included with MOVE MUST be applied in processing every
   resource to be moved with the exception of the Destination header.

   The behavior of the Destination header is the same as given for COPY
   on collections.

   When the MOVE method has completed processing it MUST have created a
   consistent namespace at both the source and destination (see section
   5.1 for the definition of namespace consistency). However, if an
   error occurs while moving an internal collection, the server MUST NOT
   move any resources identified by members of the failed collection
   (i.e., the server must skip the error-causing subtree), as this would
   create an inconsistent namespace. In this case, after detecting the
   error, the move operation SHOULD try to finish as much of the
   original move as possible (i.e., the server should still attempt to
   move other subtrees and the resources identified by their members,
   that are not descendents of an error-causing collection).  So, for
   example, if an infinite depth move is performed on collection /a/,
   which contains collections /a/b/ and /a/c/, and an error occurs
   moving /a/b/, an attempt should still be made to try moving /a/c/.
   Similarly, after encountering an error moving a non-collection
   resource as part of an infinite depth move, the server SHOULD try to
   finish as much of the original move operation as possible.

   If an error occurs with a resource other than the resource identified
   in the Request-URI then the response MUST be a 207 (Multi-Status).

   The 424 (Failed Dependency) status code SHOULD NOT be returned in the
   207 (Multi-Status) response from a MOVE method.  These errors can be
   safely omitted because the client will know that the progeny of a
   resource could not be moved when the client receives an error for the
   parent.  Additionally 201 (Created)/204 (No Content) responses SHOULD
   NOT be returned as values in 207 (Multi-Status) responses from a
   MOVE.  These responses can be safely omitted because they are the
   default success codes.

=item 8.9.3 MOVE and the Overwrite Header

   If a resource exists at the destination and the Overwrite header is
   "T" then prior to performing the move the server MUST perform a
   DELETE with "Depth: infinity" on the destination resource.  If the
   Overwrite header is set to "F" then the operation will fail.

=item 8.9.4 Status Codes

   201 (Created) - The source resource was successfully moved, and a new
   resource was created at the destination.

   204 (No Content) - The source resource was successfully moved to a
   pre-existing destination resource.

   403 (Forbidden) _ The source and destination URIs are the same.

   409 (Conflict) _ A resource cannot be created at the destination
   until one or more intermediate collections have been created.

   412 (Precondition Failed) - The server was unable to maintain the
   liveness of the properties listed in the propertybehavior XML element
   or the Overwrite header is "F" and the state of the destination
   resource is non-null.

   423 (Locked) - The source or the destination resource was locked.

   502 (Bad Gateway) - This may occur when the destination is on another
   server and the destination server refuses to accept the resource.

=item 8.9.5 Example - MOVE of a Non-Collection

   This example shows resource
   http://www.ics.uci.edu/~fielding/index.html being moved to the
   location http://www.ics.uci.edu/users/f/fielding/index.html. The
   contents of the destination resource would have been overwritten if
   the destination resource had been non-null.  In this case, since
   there was nothing at the destination resource, the response code is
   201 (Created).

   >>Request

   MOVE /~fielding/index.html HTTP/1.1
   Host: www.ics.uci.edu
   Destination: http://www.ics.uci.edu/users/f/fielding/index.html

   >>Response

   HTTP/1.1 201 Created
   Location: http://www.ics.uci.edu/users/f/fielding/index.html


=item 8.9.6 Example - MOVE of a Collection

   >>Request

   MOVE /container/ HTTP/1.1
   Host: www.foo.bar
   Destination: http://www.foo.bar/othercontainer/
   Overwrite: F
   If: (<opaquelocktoken:fe184f2e-6eec-41d0-c765-01adc56e6bb4>)
       (<opaquelocktoken:e454f3f3-acdc-452a-56c7-00a5c91e4b77>)
   Content-Type: text/xml; charset="utf-8"
   Content-Length: xxxx

   <?xml version="1.0" encoding="utf-8" ?>
   <d:propertybehavior xmlns:d='DAV:'>
     <d:keepalive>*</d:keepalive>
   </d:propertybehavior>

   >>Response

   HTTP/1.1 207 Multi-Status
   Content-Type: text/xml; charset="utf-8"
   Content-Length: xxxx

   <?xml version="1.0" encoding="utf-8" ?>
   <d:multistatus xmlns:d='DAV:'>
     <d:response>
          <d:href>http://www.foo.bar/othercontainer/C2/</d:href>
          <d:status>HTTP/1.1 423 Locked</d:status>
     </d:response>
   </d:multistatus>

   In this example the client has submitted a number of lock tokens with
   the request.  A lock token will need to be submitted for every
   resource, both source and destination, anywhere in the scope of the
   method, that is locked.  In this case the proper lock token was not
   submitted for the destination http://www.foo.bar/othercontainer/C2/.
   This means that the resource /container/C2/ could not be moved.
   Because there was an error copying /container/C2/, none of
   /container/C2's members were copied.  However no errors were listed
   for those members due to the error minimization rules given in
   section 8.8.3.  User agent authentication has previously occurred via
   a mechanism outside the scope of the HTTP protocol, in an underlying
   transport layer.

=cut
sub MOVE_handler {
  my($R, $Opts) = @_;
  my $Path = $Opts->{Path};
  my $FS = $Opts->{FS};

  return 403 if $FS->IsReadOnly();

  my ($ParentNode, $name) = eval { $FS->ParentNodeAndName($Path) };
  return 404 if $@ || !$ParentNode;

  my @locks = _ParseIfHeader($R->header_in('If'));
  return 423 if $ParentNode->NameLocked($name, \@locks, 'w');
  return 403 if !$ParentNode->CanWrite();

  my $FileInfo = $ParentNode->ChildByName($name);
  return 404 if !$FileInfo;

  my $Depth     = $R->header_in('Depth');
  my $Overwrite = $R->header_in('Overwrite');
  if (!defined($Overwrite)) {
    $Overwrite = 'T';
  }

  my $Destination = $Opts->{Destination};
  my ($DestDir, $destname) = eval { $FS->ParentNodeAndName($Destination) };
  return 409 if $@ || !$DestDir;

  return 423 if $DestDir->NameLocked($destname, \@locks, 'w');

  return 403 if !$DestDir->CanWrite();

  my $DestFileInfo = $DestDir->ChildByName($destname);
  # notice if it's the same file
  return 403 if $DestFileInfo
             && $DestFileInfo->NodeId() == $FileInfo->NodeId()
             ;

  return 412 if ($DestFileInfo and $Overwrite ne 'T'); # conflict

  my $RC = $DestFileInfo ? 204 : 201; # 'No Content' or 'Created'

  eval {
    $FileInfo->MoveTo(
      ParentNode => $DestDir,
      NodeName => $destname,
      Overwrite => ($Overwrite eq 'T'),
      IfLocks => \@locks,
    );
  };

  if ($@) {
    warn "MOVE ERROR $@";
    return 409;
  }

  $R->status($RC);
  $R->set_content_length(0);
  $R->header_out('Pragma' => 'no-cache');
  $R->header_out('Cache-Control' => 'no-cache');
  $R->content_type('application/octet-stream');

  return OK;
}

=head2 8.10 LOCK Method

   The following sections describe the LOCK method, which is used to
   take out a lock of any access type.  These sections on the LOCK
   method describe only those semantics that are specific to the LOCK
   method and are independent of the access type of the lock being
   requested.

   Any resource which supports the LOCK method MUST, at minimum, support
   the XML request and response formats defined herein.

=item 8.10.1 Operation

   A LOCK method invocation creates the lock specified by the lockinfo
   XML element on the Request-URI.  Lock method requests SHOULD have a
   XML request body which contains an owner XML element for this lock
   request, unless this is a refresh request. The LOCK request may have
   a Timeout header.

   Clients MUST assume that locks may arbitrarily disappear at any time,
   regardless of the value given in the Timeout header.  The Timeout
   header only indicates the behavior of the server if "extraordinary"
   circumstances do not occur.  For example, an administrator may remove
   a lock at any time or the system may crash in such a way that it
   loses the record of the lock's existence. The response MUST contain
   the value of the lockdiscovery property in a prop XML element.

   In order to indicate the lock token associated with a newly created
   lock, a Lock-Token response header MUST be included in the response
   for every successful LOCK request for a new lock.  Note that the
   Lock-Token header would not be returned in the response for a
   successful refresh LOCK request because a new lock was not created.

=item 8.10.2 The Effect of Locks on Properties and Collections

   The scope of a lock is the entire state of the resource, including
   its body and associated properties.  As a result, a lock on a
   resource MUST also lock the resource's properties.

   For collections, a lock also affects the ability to add or remove
   members.  The nature of the effect depends upon the type of access
   control involved.

=item 8.10.3 Locking Replicated Resources

   A resource may be made available through more than one URI. However
   locks apply to resources, not URIs. Therefore a LOCK request on a
   resource MUST NOT succeed if can not be honored by all the URIs
   through which the resource is addressable.

=item 8.10.4 Depth and Locking

   The Depth header may be used with the LOCK method.  Values other than
   0 or infinity MUST NOT be used with the Depth header on a LOCK
   method.  All resources that support the LOCK method MUST support the
   Depth header.

   A Depth header of value 0 means to just lock the resource specified
   by the Request-URI.

   If the Depth header is set to infinity then the resource specified in
   the Request-URI along with all its internal members, all the way down
   the hierarchy, are to be locked.  A successful result MUST return a
   single lock token which represents all the resources that have been
   locked.  If an UNLOCK is successfully executed on this token, all
   associated resources are unlocked.  If the lock cannot be granted to
   all resources, a 409 (Conflict) status code MUST be returned with a
   response entity body containing a multistatus XML element describing
   which resource(s) prevented the lock from being granted.  Hence,
   partial success is not an option.  Either the entire hierarchy is
   locked or no resources are locked.

   If no Depth header is submitted on a LOCK request then the request
   MUST act as if a "Depth:infinity" had been submitted.

=item 8.10.5 Interaction with other Methods

   The interaction of a LOCK with various methods is dependent upon the
   lock type.  However, independent of lock type, a successful DELETE of
   a resource MUST cause all of its locks to be removed.

=item 8.10.6 Lock Compatibility Table

   The table below describes the behavior that occurs when a lock
   request is made on a resource.

   Current lock state/  |   Shared Lock   |   Exclusive
   Lock request         |                 |   Lock
   =====================+=================+==============
   None                 |   True          |   True
   ---------------------+-----------------+--------------
   Shared Lock          |   True          |   False
   ---------------------+-----------------+--------------
   Exclusive Lock       |   False         |   False*
   ------------------------------------------------------

   Legend: True = lock may be granted.  False = lock MUST NOT be
   granted. *=It is illegal for a principal to request the same lock
   twice.

   The current lock state of a resource is given in the leftmost column,
   and lock requests are listed in the first row.  The intersection of a
   row and column gives the result of a lock request.  For example, if a
   shared lock is held on a resource, and an exclusive lock is
   requested, the table entry is "false", indicating the lock must not
   be granted.

=item 8.10.7 Status Codes

   200 (OK) - The lock request succeeded and the value of the
   lockdiscovery property is included in the body.

   412 (Precondition Failed) - The included lock token was not
   enforceable on this resource or the server could not satisfy the
   request in the lockinfo XML element.

   423 (Locked) - The resource is locked, so the method has been
   rejected.

=item 8.10.8 Example - Simple Lock Request

   >>Request

   LOCK /workspace/webdav/proposal.doc HTTP/1.1
   Host: webdav.sb.aol.com
   Timeout: Infinite, Second-4100000000
   Content-Type: text/xml; charset="utf-8"
   Content-Length: xxxx
   Authorization: Digest username="ejw",
      realm="ejw@webdav.sb.aol.com", nonce="...",
      uri="/workspace/webdav/proposal.doc",
      response="...", opaque="..."

   <?xml version="1.0" encoding="utf-8" ?>
   <D:lockinfo xmlns:D='DAV:'>
     <D:lockscope><D:exclusive/></D:lockscope>
     <D:locktype><D:write/></D:locktype>
     <D:owner>
          <D:href>http://www.ics.uci.edu/~ejw/contact.html</D:href>
     </D:owner>
   </D:lockinfo>

   >>Response

   HTTP/1.1 200 OK
   Content-Type: text/xml; charset="utf-8"
   Content-Length: xxxx

   <?xml version="1.0" encoding="utf-8" ?>
   <D:prop xmlns:D="DAV:">
     <D:lockdiscovery>
          <D:activelock>
               <D:locktype><D:write/></D:locktype>
               <D:lockscope><D:exclusive/></D:lockscope>
               <D:depth>Infinity</D:depth>
               <D:owner>
                    <D:href>
                         http://www.ics.uci.edu/~ejw/contact.html
                    </D:href>
               </D:owner>
               <D:timeout>Second-604800</D:timeout>
               <D:locktoken>
                    <D:href>
               opaquelocktoken:e71d4fae-5dec-22d6-fea5-00a0c91e6be4
                    </D:href>
               </D:locktoken>
          </D:activelock>
     </D:lockdiscovery>
   </D:prop>

   This example shows the successful creation of an exclusive write lock
   on resource http://webdav.sb.aol.com/workspace/webdav/proposal.doc.
   The resource http://www.ics.uci.edu/~ejw/contact.html contains
   contact information for the owner of the lock.  The server has an
   activity-based timeout policy in place on this resource, which causes
   the lock to automatically be removed after 1 week (604800 seconds).
   Note that the nonce, response, and opaque fields have not been
   calculated in the Authorization request header.

=item 8.10.9 Example - Refreshing a Write Lock

   >>Request

   LOCK /workspace/webdav/proposal.doc HTTP/1.1
   Host: webdav.sb.aol.com
   Timeout: Infinite, Second-4100000000
   If: (<opaquelocktoken:e71d4fae-5dec-22d6-fea5-00a0c91e6be4>)
   Authorization: Digest username="ejw",
      realm="ejw@webdav.sb.aol.com", nonce="...",
      uri="/workspace/webdav/proposal.doc",
      response="...", opaque="..."

   >>Response

   HTTP/1.1 200 OK
   Content-Type: text/xml; charset="utf-8"
   Content-Length: xxxx

   <?xml version="1.0" encoding="utf-8" ?>
   <D:prop xmlns:D="DAV:">
     <D:lockdiscovery>
          <D:activelock>
               <D:locktype><D:write/></D:locktype>
               <D:lockscope><D:exclusive/></D:lockscope>
               <D:depth>Infinity</D:depth>
               <D:owner>
                    <D:href>
                    http://www.ics.uci.edu/~ejw/contact.html
                    </D:href>
               </D:owner>
               <D:timeout>Second-604800</D:timeout>
               <D:locktoken>
                    <D:href>
               opaquelocktoken:e71d4fae-5dec-22d6-fea5-00a0c91e6be4
                    </D:href>
               </D:locktoken>
          </D:activelock>
     </D:lockdiscovery>
   </D:prop>

   This request would refresh the lock, resetting any time outs.  Notice
   that the client asked for an infinite time out but the server choose
   to ignore the request. In this example, the nonce, response, and
   opaque fields have not been calculated in the Authorization request
   header.

=item 8.10.10 Example - Multi-Resource Lock Request

   >>Request

   LOCK /webdav/ HTTP/1.1
   Host: webdav.sb.aol.com
   Timeout: Infinite, Second-4100000000
   Depth: infinity
   Content-Type: text/xml; charset="utf-8"
   Content-Length: xxxx
   Authorization: Digest username="ejw",
      realm="ejw@webdav.sb.aol.com", nonce="...",
      uri="/workspace/webdav/proposal.doc",
      response="...", opaque="..."

   <?xml version="1.0" encoding="utf-8" ?>
   <D:lockinfo xmlns:D="DAV:">
     <D:locktype><D:write/></D:locktype>
     <D:lockscope><D:exclusive/></D:lockscope>
     <D:owner>
          <D:href>http://www.ics.uci.edu/~ejw/contact.html</D:href>
     </D:owner>
   </D:lockinfo>

   >>Response

   HTTP/1.1 207 Multi-Status
   Content-Type: text/xml; charset="utf-8"
   Content-Length: xxxx

   <?xml version="1.0" encoding="utf-8" ?>
   <D:multistatus xmlns:D="DAV:">
     <D:response>
          <D:href>http://webdav.sb.aol.com/webdav/secret</D:href>
          <D:status>HTTP/1.1 403 Forbidden</D:status>
     </D:response>
     <D:response>
          <D:href>http://webdav.sb.aol.com/webdav/</D:href>
          <D:propstat>
               <D:prop><D:lockdiscovery/></D:prop>
               <D:status>HTTP/1.1 424 Failed Dependency</D:status>
          </D:propstat>
     </D:response>
   </D:multistatus>

   This example shows a request for an exclusive write lock on a
   collection and all its children.  In this request, the client has
   specified that it desires an infinite length lock, if available,
   otherwise a timeout of 4.1 billion seconds, if available. The request
   entity body contains the contact information for the principal taking
   out the lock, in this case a web page URL.

   The error is a 403 (Forbidden) response on the resource
   http://webdav.sb.aol.com/webdav/secret.  Because this resource could
   not be locked, none of the resources were locked.  Note also that the
   lockdiscovery property for the Request-URI has been included as
   required.  In this example the lockdiscovery property is empty which
   means that there are no outstanding locks on the resource.

   In this example, the nonce, response, and opaque fields have not been
   calculated in the Authorization request header.

=cut
sub LOCK_handler {
  my($R, $Opts) = @_;
  my $Path = $Opts->{Path};
  my $FS = $Opts->{FS};

  return 403 if $FS->IsReadOnly();

  my ($ParentNode, $name) = eval { $FS->ParentNodeAndName($Path) };
  return 404 if $@ || !$ParentNode;

  my @locks = _ParseIfHeader($R->header_in('If'));
  return 403 if !$ParentNode->CanWrite();

  my $Depth = _ParseDepthHeader($R->header_in('Depth'));

  my $scope = 'exclusive';
  my $type = 'write';
  my $owner;
  if (my $len = $R->header_in('Content-length')) {
    my $content;
    $R->read($content, $len);
    my $doc;
    eval {
      $doc = $PARSER->parse_string($content);
    };
    if ($@ or not $doc or _HasNULLNamespace($doc)) {
      return 400;
    }

    my ($scopenode) = _GetNSNodes($doc, 'DAV:', 'lockinfo', 'lockscope');
    my ($typenode) = _GetNSNodes($doc, 'DAV:', 'lockinfo', 'locktype');
    my ($ownernode) = grep { $_->localname() eq 'owner' } _GetNSNodes($doc, 'DAV:', 'lockinfo');
    $scope = $scopenode->localname if $scopenode;
    $type = $typenode->localname if $typenode;
    $owner = $ownernode ? $ownernode->textContent() : 'unknown';
    # strip whitespace
    $owner =~ s/^\s+//s;
    $owner =~ s/\s+$//s;
  }

  # max timeout 2 hours
  my $timeout = _ParseTimeoutHeader($R->header_in('Timeout'), (5 * 60)); # 5 mins max

  # no fancy stuff for now!
  my $expire = time() + $timeout;
  my $lock;
  if (@locks) {
    return 423 if $ParentNode->NameLocked($name, \@locks, $scope eq 'exclusive' ? 'w' : 'r');
    foreach my $lock (@locks) {
      $ParentNode->RefreshLock($name, $lock, $expire);
    }
  } else {
    return 409 unless $type eq 'write';
    return 423 if $ParentNode->NameLocked($name, \@locks, $scope eq 'exclusive' ? 'w' : 'r');
    $lock = eval { $ParentNode->NewLock($name, $owner, $scope eq 'exclusive' ? 1 : 0, 1, $expire, (defined $Depth and $Depth eq '0') ? 0 : 1) };
    if ($@) {
      warn "LOCK ERROR: $@";
    }
    return 423 unless $lock;
  }

  my $doc = XML::LibXML::Document->new('1.0', 'utf-8');
  my $prop = $doc->createElement('prop');
  $prop->setNamespace('DAV:', 'D'); # apply it at once!
  $doc->setDocumentElement($prop);

  _MakeNSNodes($doc, $prop, 'DAV:', {
    lockdiscovery => _MakeLockDiscovery($ParentNode, $name),
  });

  my $Body = qq{<?xml version="1.0" encoding="utf-8"?>\n} .
             $doc->documentElement->toString(1);

  $R->status(200);
  $R->set_content_length(do { use bytes; length($Body) });
  $R->header_out('Content-Location' => _EscapeURI("$Opts->{Base}$Path"));
  $R->header_out('Pragma' => 'no-cache');
  $R->header_out('Cache-Control' => 'no-cache');
  if ($lock) {
    $R->header_out('Lock-Token' => '<opaquelocktoken:' . $lock . '>');
  }
  $R->content_type('text/xml; charset=utf-8');
  $R->print($Body);

  return OK;
}

=head2 8.11 UNLOCK Method

   The UNLOCK method removes the lock identified by the lock token in
   the Lock-Token request header from the Request-URI, and all other
   resources included in the lock.  If all resources which have been
   locked under the submitted lock token can not be unlocked then the
   UNLOCK request MUST fail.

   Any DAV compliant resource which supports the LOCK method MUST
   support the UNLOCK method.

   >>Request

   UNLOCK /workspace/webdav/info.doc HTTP/1.1
   Host: webdav.sb.aol.com
   Lock-Token: <opaquelocktoken:a515cfa4-5da4-22e1-f5b5-00a0451e6bf7>
   Authorization: Digest username="ejw",
      realm="ejw@webdav.sb.aol.com", nonce="...",
      uri="/workspace/webdav/proposal.doc",
      response="...", opaque="..."

   >>Response

   HTTP/1.1 204 No Content

   In this example, the lock identified by the lock token
   "opaquelocktoken:a515cfa4-5da4-22e1-f5b5-00a0451e6bf7" is
   successfully removed from the resource
   http://webdav.sb.aol.com/workspace/webdav/info.doc.  If this lock
   included more than just one resource, the lock is removed from all
   resources included in the lock.  The 204 (No Content) status code is
   used instead of 200 (OK) because there is no response entity body.

   In this example, the nonce, response, and opaque fields have not been
   calculated in the Authorization request header.

=cut
sub UNLOCK_handler {
  my($R, $Opts) = @_;
  my $Path = $Opts->{Path};
  my $FS = $Opts->{FS};
  
  return 403 if $FS->IsReadOnly();

  my ($ParentNode, $name) = eval { $FS->ParentNodeAndName($Path) };
  return 404 if $@ || !$ParentNode;

  my @locks = _ParseIfHeader($R->header_in('If'));
  return 403 if !$ParentNode->CanWrite();

  my $Remove = _ParseLockTokenHeader($R);
  push @locks, $Remove;

  if ($ParentNode->NameLocked($name, \@locks, 'r')) {
    return 423;
  }

  eval {
    my $rc = $ParentNode->RemoveLock($name, $Remove);
    return 409 unless $rc;
  };
  if ($@) {
    warn "UNLOCK FAILED: $@";
    return 409;
  }

  return 204;
}

# XML Support Functions

# search down a tree for a particular node (sort of a
# "poor man's xpath".
sub _GetNSNodes {
  my ($doc, $ns, @tree) = @_;

  my @nodes = ($doc->documentElement());
  while (my $req = shift @tree) {
    @nodes = grep { ref($_) eq 'XML::LibXML::Element' }
             map { $_->childNodes() }
             grep { _NSMatch($_->localname, $req) and $_->namespaceURI eq $ns }
             @nodes;
  }

  return @nodes;
}

sub _NSMatch {
  my $name = shift;
  my $match = shift;

  unless (ref $match) {
    return $name eq $match;
  }

  if (ref $match eq 'ARRAY') {
    foreach my $item (@$match) {
      return 1 if _NSMatch($name, $item);
    }
    return 0;
  }

  if (ref $match eq 'Regexp') {
    return $name =~ $match;
  }

  if (ref $match eq 'CODE') {
    return $match->($name);
  }

  return 0;
}

# generate a single element while reusing existing NamespaceURI
# declerations.
sub _MakeElement {
  my $doc = shift;
  my $ns = shift;
  my $name = shift;
  if (my $prefix = $doc->documentElement->lookupNamespacePrefix($ns)) {
    return $doc->createElement("$prefix:$name");
  } else {
    return $doc->createElementNS($ns, $name);
  }
}

# convert between a nice Perl hash structure and XML nodes, making it
# easier to build complex XML output.
sub _MakeNSNodes {
  my $doc = shift;
  my $basenode = shift;
  my $ns = shift;
  my $value = shift;

  return unless defined $value;
  if (ref($value) eq 'ARRAY') {
    foreach my $item (@$value) {
      _MakeNSNodes($doc, $basenode, $ns, $item);
    }
  } elsif (ref($value) eq 'HASH') {
    foreach my $key (sort keys %$value) {
      if (ref($value->{$key})eq 'ARRAY') {
        foreach my $item (@{$value->{$key}}) {
          my $node = _MakeElement($doc, $ns, $key);
          $basenode->appendChild($node);
          _MakeNSNodes($doc, $node, $ns, $item);
        }
      } else {
       my $node = _MakeElement($doc, $ns, $key);
       $basenode->appendChild($node);
       _MakeNSNodes($doc, $node, $ns, $value->{$key});
      }
    }
  } elsif (ref($value) eq 'CODE') {
    $value->($doc, $basenode, $ns);
  } elsif (ref($value) =~ m/^XML::LibXML/) {
    $basenode->appendChild($value);
  } else {
    $basenode->appendText($value);
  }
}

# Check for the 'NULL Namespace' bug as reported by litmus
sub _HasNULLNamespace {
  my $doc = shift;
  my @nodes = $doc->findnodes('//*');
  return grep { my $u = $_->namespaceURI; defined($u) and $u eq '' } @nodes;
}

##############################################################
# Header support functions

# RFC2518: 9.1 DAV Header
# - only outbound!

# RFC2518: 9.2 Depth Header
sub _ParseDepthHeader {
  my $Depth = shift;
  my $default = shift || 'infinity';
  return $default unless defined $Depth;
  return 'zero' if $Depth eq '0';
  return 'one' if $Depth eq '1';
  return 'infinity' if $Depth =~ m{ infinity }ixms;
  die "Invalid Depth Header $Depth";
}

# RFC2518: 9.3 Destination Header
# - fixed up above

# RFC2518: 9.4 If Header
sub _ParseIfHeader {
  my $If = shift;
  return () unless $If;
  my ($items) = $If =~ m/\(\s*([^\)]+)\s*\)/;
  my @list = split /\s*,\s*/, $items;
  s{ < opaquelocktoken: ([^>]+) > }{$1}xms for @list;
  return @list;
}

# RFC2518: 9.5 Lock-Token Header
sub _ParseLockTokenHeader {
  my $R = shift;
  my $header = $R->header_in('Lock-Token');
  return unless defined $header;
  return 
    unless $header =~ m{ < \s* opaquelocktoken : \s* ([^> ]+) \s* > }xms;
  my $token = $1;
  return $token;
}

# RFC2518: 9.6 Overwrite Header
sub _ParseOverwriteHeader {
  my $R = shift;
  my $default = shift || 'T';

  # grab the header
  my $header = $R->header_in('Overwrite');

  return $default if !defined $header;
  return 'T'      if $header =~ m{ T }ixms;
  return 'F'      if $header =~ m{ F }ixms;
  # bogus!
  die "Invalid Overwrite header $header\n";
}

# RFC2518: 9.7 Status-URI Response Header
# - we don't do that.

# RFC2518: 9.8 Timeout Request Header
sub _ParseTimeoutHeader {
  my $Timeout = shift;
  my $Max = shift;

  return $Max unless $Timeout;

  my @list = split /\s*,\s*/, $Timeout;
  for my $item (@list) {
    next unless $item =~ m{ Second- (\d+) }ixms;
    my $seconds = $1;
    next if $seconds > $Max;
    return $seconds;
  }
  return $Max;
}

# Icky client support!

sub _DetectClientQuirks {
  my $R = shift;
  my $ua = $R->user_agent();

  my %client;

  # check that a User-Agent header was even set!
  return \%client unless $ua;

  # Macs use a custom outdated quota method of 512 byte blocks!
  # http://mailman.lyra.org/pipermail/dav-dev/2004-February/005307.html
  if ($ua =~ m{ WebDAVFS }xms && $ua =~ m{ Darwin }xms) {
    $client{HasQuota512} = 1;
  }

  # Microsoft specific extentions, see
  # http://www.greenbytes.de/tech/webdav/webfolder-client-list.html
  # under "undefined-properties".
  if ($ua =~ m{ Microsoft \s Data \s Access }xms) {
    $client{HasWindowsProperties} = 1;
  }

  if ($ua =~ m{ Microsoft-WebDAV-MiniRedir }xms) {
    $client{WindowsVistaAllprop} = 1;
    $client{HasWindowsProperties} = 1;
  }

  if ($ua =~ m{ Konqueror\/3 }xms) {
    $client{HasUTF8Path} = 1;
  }

  if ($ua =~ m{ WebDrive }xms) {
    $client{WebDriveNS} = 1;
  }

  # XXX: others

  return \%client;
}

# Locking support functions

sub _MakeLockDiscovery {
  my $Info = shift;
  my $name = shift;
  my $Locks = $Info->GetLocks($name);

  my @Res;

  my $time = time();

  LOCK:
  foreach my $Lock (values %{$Locks}) {
    next LOCK if (not $Lock->{Expires} or $Lock->{Expires} < $time);
    push @Res, {
      locktype => { ($Lock->{IsWrite} ? 'write' : 'read') => undef },
      lockscope => { ($Lock->{IsExclusive} ? 'exclusive' : 'shared') => undef },
      owner => $Lock->{Owner},
      depth => ($Lock->{IsDeep} ? 'infinity' : '0'),
      timeout => "Second-" . ($Lock->{Expires} - $time),
      locktoken => { href => 'opaquelocktoken:' . $Lock->{LockName} },
    }
  }

  return undef unless @Res;
  return { activelock => \@Res };
}

sub _MakeSupportedLock {
  return {
    lockentry => [
      {
        lockscope => { shared => undef },
        locktype => { write => undef },
      },
      {
        lockscope => { exclusive => undef },
        locktype => { write => undef },
      },
    ],
  };
}

# Filesystem support

# Deep search for 'Depth: Infinity' propfind requests
sub _GetAllChildren {
  my $DirInfo = shift;
  my $list = shift;
  # Self is already on the list
  foreach my $node ($DirInfo->SubDirs()) {
    push @$list, $node;
    _GetAllChildren($node, $list);
  }
  push @$list, $DirInfo->Files();
}

sub _EscapeURI {
  # Escape all unicode chars + space and quote
  return escape_uri(shift, 0, q{"'` <>;});
}

# dodgy logging support 
sub _debuglog {
  my $User = shift;
  my $String = shift;

}

1;
