#!/usr/bin/perl -cw 

=pod

=head1 NAME

ME::VFS::Interface::FTP - The MessagingEngine personality

=head1 SYNOPSIS

  hm/scripts/ftpserver.pl [-d] [-v] [-p port] [-s] [-S] [-V] [-C conf_file]

=head1 DESCRIPTION

C<ME::VFS::Interface::FTP> is the MessagingEngine server

=head1 METHODS

=over 4

=cut

package ME::VFS::Interface::FTP;

use strict;
use warnings;

use vars qw($VERSION);
($VERSION) = ( '$Revision: 22691 $ ' =~ /\$Revision:\s+([^\s]+)/ ) ? $1 : 'unknown';

use ME::VFS;
use ME::User;
use ME::FMVars;
use IO::Socket::UNIX;

use ME::VFS::Interface::FTP::FileHandle;
use ME::VFS::Interface::FTP::DirHandle;
use Net::FTPServer;

use vars qw(@ISA);
@ISA = qw(Net::FTPServer);

use constant MAX_FILE_SIZE => 1073741824; # 1GB

# This is called before configuration.
sub pre_configuration_hook {
  my $self = shift;

  $self->{version_string} .= " ME::VFS::Interface::FTP/$VERSION";

  # Custom SITE commands.
  $self->{site_command_table}{USAGE} = \&_SITE_USAGE_command;

  # Add HASH command (http://tools.ietf.org/html/draft-ietf-ftpext2-hash-00)
  $self->{features}->{HASH} = "SHA-1*";
  $self->{options}->{HASH} = \&_OPTS_HASH_command;
  $self->{command_table}->{HASH} = \&_HASH_command;

  # Add XCRC command
  $self->{command_table}->{XCRC} = \&_XCRC_command;
  $self->{features}->{XCRC} = undef;

  # enable transfer logging
  # NOTE: expected to be a filehandle, but never checked as such,
  # and we override where it's written to;
  $self->{_xferlog} = 1;
}

# This is run after the log file is opened but before we've done
# any dangerous chat to the connecting user!
sub post_accept_hook {
  my $self = shift;

  # always run as fmuser
  my (undef, undef, $uid, $gid) = getpwnam('fmuser');
  $self->_drop_privs($uid, $gid, 'fmuser');

  # don't share the same seed for all forked copies
  srand();

  $self->{site_command_table}{USAGE} = \&_SITE_USAGE_command;
}

# Perform login against the database.
sub authentication_hook {
  my $self = shift;
  my $username = shift;
  my $password = shift;
  my $user_is_anon = shift;

  # Disallow anonymous access.
  return -1 if $user_is_anon;

  my ($u, undef, $d) = ME::User::SplitLoginName($username);
  return (-1, "Username must be complete user\@domain") if !$u || !$d;

  my ($User, $Reason, %Details) = ME::User->new_login(
    $username, $password, 'ftp', $self->{peeraddrstring},
  );

  # user has to exist, and be able to log in
  return (-1, $Reason) unless $User;

  # set up the environment for ease of checking the other conditions
  $self->{fs_user} = $User;
  my $VFS = $User->VFS(Init => 1);
  $self->{fs_me} = $VFS ;

  # check if this is a "MyFiles" session
  my ($Service) = grep { $_->Arg eq $self->{sockaddrstring} } @{ME::FMVars::FailoverServices()};
  if ($Service and $Service->Name =~ m/^ip_myfiles/) {
    $VFS->Chroot($VFS->HomeDir());
  }

  # tell the user stuff
  if ($Details{IsRestrictedLogin}) {
    $self->{fs_restricted} = 1;
    my $Message = $Details{LoginMessage} || "Restricted Session";
    $self->_set_config('welcome type', 'text', splat => 1);
    $self->_set_config('welcome text', "$Message. This session is read-only - you can rename or upload but not delete or overwrite existing files.", splat => 1);
  }
  elsif ($Details{LoginMessage}) {
    my $Message = $Details{LoginMessage};
    $self->_set_config('welcome type', 'text', splat => 1);
    $self->_set_config('welcome text', "$Message", splat => 1);
  }

  if ($Details{ExpireTime}) {
    $self->{fs_expire} = $Details{ExpireTime};
  }

  # only three connections at once
  my $OtherPids = $VFS->UserFlagGet('ftp session') || [];
  @$OtherPids = grep { kill(0, $_) } @$OtherPids;
  if (@$OtherPids >= 20) {
    return (-1, "Only twenty FTP logins permitted at a time");
  }
  $VFS->UserFlagSet('ftp session', [ @$OtherPids, $$ ]);

  # Put UnqName in the PS listing for easy seeing who's FTPing
  $0 = "$0 " . $User->UnqName();
  $self->{ps_basename} = $0;

  # User wants extra logging?
  my $logdir = "/var/log/ftpd.users/" . $User->UnqName();
  if (-d $logdir) {
    $self->_set_config('client logging', "$logdir/clientlog", splat => 1);
    $self->_open_client_log();
    $self->_set_config('error log', "$logdir/errorlog", splat => 1);
    $self->_open_error_log();
    #warn "LOGGING " . Data::Dumper::Dumper($self);
    $ME::VFS::Cache::DEBUG_CACHE = 1;
  }

  # Successful login.
  return 0;
}

#  Return an instance of DirHandle
# corresponding to the root directory.
sub root_directory_hook {
  my $self = shift;

  return ME::VFS::Interface::FTP::DirHandle->new($self, '/');
}

# The SITE USAGE command.
sub _SITE_USAGE_command {
  my $self = shift;
  my $cmd = shift;
  my $rest = shift;

  $self->reply (200, "Nothing to see here yet");
}

# The XCRC command
sub _XCRC_command {
  my $self = shift;
  my $cmd = shift;
  my $rest = shift;

  # Stupid non-standard command
  # http://help.globalscape.com/help/eft6/FileIntegrityChecking.htm
  my ($start, $end) = (0);
  if ($rest =~ /^"([^"]*)"(?: (\d+))?(?: (\d+))?/) {
    $rest = $1;
    $end = $2 if defined $2;
    ($start, $end) = ($2, $3) if defined $2 && defined $3;
  }

  # Find file by name.
  my ($dirh, $fileh, $filename) = $self->_get ($rest);

  unless ($fileh) {
    $self->reply (550, "File or directory not found.");
    return;
  }
  my $crc = $fileh->crc32($start, $end);
  unless (length($crc) == 8 ) {
    $self->reply (550, "Internal error calculating crc.");
    return;
  }

  $self->reply (250, "XCRC " . $crc );
}

# The HASH command
sub _HASH_command {
  my $self = shift;
  my $cmd = shift;
  my $rest = shift;

  # Find file by name.
  my ($dirh, $fileh, $filename) = $self->_get ($rest);

  unless ($fileh) {
    $self->reply (550, "File or directory not found.");
    return;
  }

  $self->reply (213, "SHA-1 " . $fileh->sha1() . " " . $filename);
}
sub _OPTS_HASH_command {
  my $self = shift;
  my $cmd = shift;
  my $rest = shift;

  if ($rest && uc($rest) ne 'SHA-1') {
    $self->reply (501, "Unknown algorithm, current selection not changed");
    return;
  }

  $self->reply (200, 'SHA-1');
}

# Count bytes and disconnect when file too large
sub transfer_hook {
  my ($self, $mode, $file, $sock, $buffer) = @_;

  # Only interested in uploads (writes)
  return if $mode ne 'w';

  $self->{fs_bytecount} += length $$buffer;
  if ($self->{fs_bytecount} > MAX_FILE_SIZE) {
    # We don't want to commit a partial file to VFS
    $file->DeleteOnClose();

    $self->log("notice", "aborting upload for ".$self->{fs_user}->UnqName.", exceeded size limit");
    return "File size limit exceeded";
  }

  return;
}

# RateTrack check
sub rate_check_hook {
  my $self = shift;
  my $file_name = shift;
  my $file_size = shift;
  my $direction = shift;

  my $VFS = $self->{fs_me};

  my $result = $VFS->RateCheck($file_size, $file_name);

  return $result;
}

# RateTrack update
sub rate_update_hook {
  my $self = shift;
  my $file_name = shift;
  my $file_size = shift;
  my $direction = shift;

  # rate_update_hook is only called at end of transfer, so its a convenient
  # place to do post-transfer stuff

  # clear the count of bytes transferred
  delete $self->{fs_bytecount};

  # log uploads
  if ($direction eq 'u' && $self->{_xferlog} && $self->{xfer}) {
    $self->log("notice", "upload complete, ip=$self->{peeraddrstring}, user=$self->{user}, path=$self->{xfer}->{path}, size=$file_size");
  }

  my $VFS = $self->{fs_me};

  my $result = $VFS->RateUpdate($file_size, $file_name, ForceFlag => ($direction eq 'u'));

  return $result;
}

# Log transfer amount
sub xfer_complete {
  my $self = shift;
  my $xfer = $self->{xfer} or return;

  # If no xfer log handle, return
  return unless $self->{_xferlog};

  # Never log if zero bytes transferred
  return unless $xfer->{bytes};

  # LogBwQuotaUse now implemented in RateUpdate

  # xfer_complete implementation from Net::FTPServer
  $xfer->{status} = 'c';
  # if it's just our hack, don't log it!
  $self->xfer_flush unless $self->{_xferlog} eq '1';

  # Wipe xfer ref to signify that it's flushed
  delete $self->{xfer};
}

# catch flushes too
sub xfer_flush {
  my $self = shift;
  return $self->xfer_complete(@_);
}

# pre-command items
sub pre_command_hook {
  my $self = shift;
  my ($cmdline) = @_;

  # Handle login timeouts
  if ($self->{fs_expire} and $self->{fs_expire} < time()) {
    $self->log ("info", "exiting on password expiry");
    print "421 Server closed the connection after password expiry.\r\n";
    $self->_log_line ("[PASSWORD EXPIRED!]");
    exit;
  }

  # Re-init cache so we don't get stale directory listings
  ME::VFS::Cache::Init();

  return 0;
}

# Syslog and wide_chars are not our friend.
sub log
{
  my $self = shift;

  Sys::Syslog::syslog (map { Encode::encode('utf8', $_) } @_) if $self->{_enable_syslog};
}

# Bug in Net::FTPServer dammit
sub _log_line
{
  my $self = shift;
  return unless exists $self->{_client_log};
  my $message = join ("",@_);
  my $io = $self->{_client_log};
  my $time = scalar localtime;
  my $authenticated = $self->{authenticated} ? $self->{user} : "-";
  $message =~ s/\n*$/\n/;
  $io->print ("[$time][$$:$authenticated]$message");
}

# update $0
sub _chdir_message {
  my $self = shift;
  my $cwd = $self->{cwd};
  my $path = $self->{cwd}->pathname;
  $0 = "$self->{ps_basename} $path";
  return $self->SUPER::_chdir_message(@_);
}

# Tidy up when the user leaves
sub quit_hook {
  my $self = shift;

  if (my $VFS = $self->{fs_me}) {
    my $Pids = $VFS->UserFlagGet('ftp session') || [];
    $VFS->UserFlagSet('ftp session', [ grep { $_ != $$ } @$Pids ]);
  }
}

1; # So that the require or use succeeds.

__END__

=back 4

=head1 FILES

  /etc/ftpd.conf
  /usr/lib/perl5/site_perl/5.005/Net/FTPServer.pm
  /usr/lib/perl5/site_perl/5.005/Net/FTPServer/DirHandle.pm
  /usr/lib/perl5/site_perl/5.005/Net/FTPServer/FileHandle.pm
  /usr/lib/perl5/site_perl/5.005/Net/FTPServer/Handle.pm
  /usr/lib/perl5/site_perl/5.005/Net/FTPServer/DBeg1/Server.pm
  /usr/lib/perl5/site_perl/5.005/Net/FTPServer/DBeg1/DirHandle.pm
  /usr/lib/perl5/site_perl/5.005/Net/FTPServer/DBeg1/FileHandle.pm

=head1 AUTHORS

Richard Jones (rich@annexia.org).

=head1 COPYRIGHT

Copyright (C) 2000 Biblio@Tech Ltd., Unit 2-3, 50 Carnwath Road,
London, SW6 3EG, UK

=head1 SEE ALSO

L<Net::FTPServer(3)>,
L<Net::FTP(3)>,
L<perl(1)>,
RFC 765,
RFC 959,
RFC 1579,
RFC 2389,
RFC 2428,
RFC 2577,
RFC 2640,
Extensions to FTP Internet Draft draft-ietf-ftpext-mlst-NN.txt.

=cut
