#!/usr/bin/perl -cs

=pod

=head1 NAME

ME::VFS::Interface::FTP::DirHandle - The Messaging Engine personality

=head1 SYNOPSIS

  use ME::VFS::Interface::FTP::DirHandle;

=head1 DESCRIPTION

=head1 METHODS

=over 4

=cut

package ME::VFS::Interface::FTP::DirHandle;

use strict;
use warnings;

use vars qw($VERSION);
( $VERSION ) = '$Revision: 18646 $ ' =~ /\$Revision:\s+([^\s]+)/;

use IO::Dir;
use Carp qw(confess);
use Encode;

use Net::FTPServer::DirHandle;

use vars qw(@ISA);

@ISA = qw(Net::FTPServer::DirHandle);

=pod

=item $handle = $dirh->get ($filename);

Return the file or directory C<$handle> corresponding to
the file C<$filename> in directory C<$dirh>. If there is
no file or subdirectory of that name, then this returns
undef.

=cut

sub dir { return $_[0] }

sub get {
  my $self = shift;
  my $filename = shift;

  # None of these cases should ever happen.
  confess "no filename" unless defined($filename) && length($filename);
  confess "slash filename" if $filename =~ /\//;
  confess ".. filename"    if $filename eq "..";
  confess ". filename"     if $filename eq ".";

  #note - this can be called either with an already-encoded name from
  #an internal callback, or direct from the wire.  FixString will
  #complain (fair enough) if called recursively, but it's too messy to
  #fix FTPServer to use different paths
  $filename = $self->{ftps}->{fs_me}->FixString($filename)
    unless Encode::is_utf8($filename);

  my $pathname = $self->{_pathname} . $filename;
  my $info = $self->{ftps}->{fs_me}->PathToNode($pathname);
  return undef unless $info;

  if ($info->IsDir()) {
    return ME::VFS::Interface::FTP::DirHandle->new($self->{ftps}, $pathname."/");
  }

  if ($info->IsFile()) {
    return ME::VFS::Interface::FTP::FileHandle->new($self->{ftps}, $pathname);
  }

  return undef;
}

=item $dirh = $dirh->parent;

Return the parent directory of the directory C<$dirh>. If
the directory is already "/", this returns the same directory handle.

=cut

sub parent {
  my $self = shift;

  my $parent = $self->SUPER::parent;
  bless $parent, ref $self;
  return $parent;
}

=pod

=item $ref = $dirh->list ([$wildcard]);

Return a list of the contents of directory C<$dirh>. The list
returned is a reference to an array of pairs:

  [ $filename, $handle ]

The list returned does I<not> include "." or "..".

The list is sorted into alphabetical order automatically.

=cut

sub list {
  my $self = shift;
  my $wildcard = shift;

  # Convert wildcard to a regular expression.
  if ($wildcard) {
    # convert to Perl Utf8
    $wildcard = $self->{ftps}->{fs_me}->FixString($wildcard);
    $wildcard = $self->{ftps}->wildcard_to_regex ($wildcard);
  }

  my $info = $self->{ftps}->{fs_me}->PathToNode($self->{_pathname});
  return undef unless $info;

  my @filenames;
  foreach my $file ($info->SubDirs(), $info->Files()) {
    my $name = $file->NodeName();
    next if $wildcard && $name !~ /$wildcard/;
    push @filenames, $name;
  }

  my @array = ();

  foreach my $name (sort @filenames) {
    if (my $handle = $self->get($name)) {
      push @array, [ Encode::encode_utf8($name), $handle ];
    }
  }

  return \@array;
}

=pod

=item $ref = $dirh->list_status ([$wildcard]);

Return a list of the contents of directory C<$dirh> and
status information. The list returned is a reference to
an array of triplets:

  [ $filename, $handle, $statusref ]

where $statusref is the tuple returned from the C<status>
method (see L<Net::FTPServer::Handle>).

The list returned does I<not> include "." or "..".

The list is sorted into alphabetical order automatically.

=cut

sub list_status {
  my $self = shift;

  my $arrayref = $self->list(@_);
  my $elem;

  foreach $elem (@$arrayref) {
    my @status = $elem->[1]->status;
    push @$elem, \@status;
  }

  return $arrayref;
}

=pod

=item ($mode, $perms, $nlink, $user, $group, $size, $time) = $handle->status;

Return the file or directory status. The fields returned are:

  $mode     Mode        'd' = directory,
                        'f' = file,
                        and others as with
                        the find(1) -type option.
  $perms    Permissions Permissions in normal octal numeric format.
  $nlink    Link count
  $user     Username    In printable format.
  $group    Group name  In printable format.
  $size     Size        File size in bytes.
  $time     Time        Time (usually mtime) in Unix time_t format.

In derived classes, some of this status information may well be
synthesized, since virtual filesystems will often not contain
information in a Unix-like format.

=cut

sub status {
  my $self = shift;

  my $info = eval { $self->{ftps}->{fs_me}->PathToNode($self->{_pathname}) };

  # Generate printable user/group.
  my $user  = $self->{ftps}->{fs_user}->{UserName};
  my $group = $self->{ftps}->{fs_user}->{Domain};
  $group =~ s{\..*}{};

  # If the directory has been removed since we created this
  # handle, then $dev will be undefined. Return dummy status
  # information.
  return ("d", 0000, 1, $user, $group, 0, 0) unless $info;

  my $type = 'd';

  # Permissions from mode.
  my $perms = 0700;

  my $size = 0;
  my $mtime = $info->Modified();

  # Always 1 link
  my $nlink = 1;

  return ($type, $perms, $nlink, $user, $group, $size, $mtime);
}

=pod

=item $rv = $handle->move ($dirh, $filename);

Move the current file (or directory) into directory C<$dirh> and
call it C<$filename>. If the operation is successful, return 0,
else return -1.

Underlying filesystems may impose limitations on moves: for example,
it may not be possible to move a directory; it may not be possible
to move a file to another directory; it may not be possible to
move a file across filesystems.

=cut

sub move {
  my $self = shift;
  my $dirh = shift;
  my $filename = shift;

  die if $filename =~ /\//;  # Should never happen.

  # convert to Perl Utf8
  $filename = $self->{ftps}->{fs_me}->FixString($filename);

  $self->{ftps}->{fs_me}->Move($self->{_pathname},
                               $dirh->{_pathname} . '/' . $filename,
                               KeepTimes => 1);

  $self->{_pathname} = $dirh->{_pathname} . $filename;

  return 0;
}

=pod

=item $rv = $dirh->delete;

Delete the current directory. If the delete command was
successful, then return 0, else if there was an error return -1.

It is normally only possible to delete a directory if it
is empty.

=cut

sub delete {
  my $self = shift;

  return -1 if $self->{ftps}->{fs_restricted};

  $self->{ftps}->{fs_me}->Delete($self->{_pathname});

  return 0;
}

=item $rv = $dirh->mkdir ($name);

Create a subdirectory called C<$name> within the current directory
C<$dirh>.

=cut

sub mkdir {
  my $self = shift;
  my $name = shift;

  die if $name =~ /\//;  # Should never happen.
  # no creating the special directories thanks
  return undef if $name eq '..';
  return undef if $name eq '.';

  $self->{ftps}->{fs_me}->Mkdir( $self->{_pathname} . $name );

  return 0;
}

=item $file = $dirh->open ($filename, "r"|"w"|"a") ;

Open or create a file called C<$filename> in the current directory,
opening it for either read, write or append. This function
returns a C<IO::File> handle object.

=cut

sub open {
  my $self = shift;
  my $filename = shift;
  my $mode = shift;

  die if $filename =~ /\//;  # Should never happen.

  # convert to Perl Utf8
  $filename = $self->{ftps}->{fs_me}->FixString($filename);

  my $Overwrite = not $self->{ftps}->{fs_restricted};

  return eval {
    $self->{ftps}->{fs_me}->OpenFile($self->{_pathname} . '/' . $filename, $mode,
                                     Overwrite => $Overwrite,
                                     SkipSizeCheck => 1);
  };
}

sub utime {
  my $self = shift;
  my $mtime = shift;

  my $info = $self->{ftps}->{fs_me}->PathToNode($self->{_pathname});
  return -1 unless $info;

  eval { $info->SetModified($mtime) };
  return -1 if $@;

  return 0;
}

1 # So that the require or use succeeds.

__END__

=back 4

=head1 AUTHORS

Richard Jones (rich@annexia.org).

=head1 COPYRIGHT

Copyright (C) 2000 Biblio@Tech Ltd., Unit 2-3, 50 Carnwath Road,
London, SW6 3EG, UK

=head1 SEE ALSO

L<Net::FTPServer(3)>, L<perl(1)>

=cut
