#!/usr/bin/perl -cw

=pod

=head1 NAME

ME::VFS::Interface::FTP::FileHandle - The Messaging Engine personality

=head1 SYNOPSIS

  use ME::VFS::Interface::FTP::FileHandle;

=head1 DESCRIPTION

=head1 METHODS

=over 4

=cut

package ME::VFS::Interface::FTP::FileHandle;

use strict;
use warnings;

use vars qw($VERSION);
( $VERSION ) = '$Revision: 18646 $ ' =~ /\$Revision:\s+([^\s]+)/;

use Net::FTPServer::FileHandle;
use ME::VFS::Interface::FTP::DirHandle;
use String::CRC32 qw();

use vars qw(@ISA);

@ISA = qw(Net::FTPServer::FileHandle);

=pod

=item $dirh = $fileh->dir;

Return the directory which contains this file.

=cut

sub dir {
  my $self = shift;

  my $dirname = $self->{_pathname};
  $dirname =~ s,[^/]+$,,;

  return ME::VFS::Interface::FTP::DirHandle->new ($self->{ftps}, $dirname);
}

=pod

=item $fh = $fileh->open (["r"|"w"|"a"]);

Open a file handle (derived from C<IO::Handle>, see
L<IO::Handle(3)>) in either read or write mode.

=cut

sub open {
  my $self = shift;
  my $mode = shift;

  my $Overwrite = not $self->{ftps}{fs_restricted};

  return eval { $self->{ftps}{fs_me}->OpenFile($self->{_pathname}, $mode, Overwrite => $Overwrite, SkipSizeCheck => 1) };
}

=pod

=item ($mode, $perms, $nlink, $user, $group, $size, $time) = $handle->status;

Return the file or directory status. The fields returned are:

  $mode     Mode        'd' = directory,
                        'f' = file,
                        and others as with
                        the find(1) -type option.
  $perms    Permissions Permissions in normal octal numeric format.
  $nlink    Link count
  $user     Username    In printable format.
  $group    Group name  In printable format.
  $size     Size        File size in bytes.
  $time     Time        Time (usually mtime) in Unix time_t format.

In derived classes, some of this status information may well be
synthesized, since virtual filesystems will often not contain
information in a Unix-like format.

=cut

sub status {
  my $self = shift;

  my $info = $self->{ftps}->{fs_me}->PathToNode($self->{_pathname});

  # Generate printable user/group.
  my $user  = $self->{ftps}->{fs_user}->{UserName};
  my $group = $self->{ftps}->{fs_user}->{Domain};
  $group =~ s{\..*}{};

  # If the file has been removed since we created this
  # handle, then $info will be undefined. Return dummy status
  # information.
  return ("f", 0000, 1, $user, $group, 0, 0) unless $info;

  # Permissions from mode.
  my $perms = 0600;

  my $type = 'f';

  my $size = $info->FileSize();
  my $mtime = $info->Modified();

  # Always 1 link
  my $nlink = 1;

  return ($type, $perms, $nlink, $user, $group, $size, $mtime);
}

=pod

=item $rv = $handle->move ($dirh, $filename);

Move the current file (or directory) into directory C<$dirh> and
call it C<$filename>. If the operation is successful, return 0,
else return -1.

Underlying filesystems may impose limitations on moves: for example,
it may not be possible to move a directory; it may not be possible
to move a file to another directory; it may not be possible to
move a file across filesystems.

=cut

sub move {
  my $self = shift;
  my $dirh = shift;
  my $filename = shift;

  # convert to Perl Utf8
  $filename = $self->{ftps}->{fs_me}->FixString($filename);

  die if $filename =~ /\//;  # Should never happen.

  my $new_name = $dirh->{_pathname} . $filename;

  $self->{ftps}{fs_me}->Move($self->{_pathname},
                             $dirh->{_pathname} . '/' . $filename,
                             KeepTimes => 1);

  $self->{_pathname} = $new_name;
  return 0;
}

=pod

=item $rv = $fileh->delete;

Delete the current file. If the delete command was
successful, then return 0, else if there was an error return -1.

=cut

sub delete {
  my $self = shift;

  return -1 if $self->{ftps}{fs_restricted};

  $self->{ftps}{fs_me}->Delete($self->{_pathname});

  return 0;
}

=item $link = $fileh->readlink;

If the current file is really a symbolic link, read the contents
of the link and return it.

=cut

sub readlink {
  my $self = shift;

  return $self->{_pathname};
}

sub utime {
  my $self = shift;
  my $mtime = shift;

  my $info = $self->{ftps}->{fs_me}->PathToNode($self->{_pathname});
  return -1 unless $info;

  eval { $info->SetModified($mtime) };
  return -1 if $@;

  return 0;
}

sub sha1 {
  my $self = shift;

  my $info = $self->{ftps}->{fs_me}->PathToNode($self->{_pathname});
  return -1 unless $info;

  return $info->Sha1Hex();
}

sub crc32 {
  my $self = shift;
  my $start = shift;
  my $end = shift;

  my $Fh =
    eval { $self->{ftps}{fs_me}->OpenFile( $self->{_pathname}, "r" ); };
  return -1 unless $Fh;

  if (!$start && !$end) {
    return sprintf( "%08x", String::CRC32::crc32($Fh) );

  } else {
    # bytes to read or -1 for eof
    my $rem = defined($end) && $end > $start ? $end - $start : -1;

    $Fh->seek($start // 0, 0);

    my $crc = 0;
    while ($rem) {
      $Fh->read(my $buf, my $amount = $rem > 4096 || $rem < 0 ? 4096 : $rem)
        || last;
      $crc = String::CRC32::crc32($buf, $crc);
      $rem -= $amount if $rem > 0;
    }

    return sprintf("%08x", $crc);
  }
}

1 # So that the require or use succeeds.

__END__

=back 4

=head1 AUTHORS

Richard Jones (rich@annexia.org).

=head1 COPYRIGHT

Copyright (C) 2000 Biblio@Tech Ltd., Unit 2-3, 50 Carnwath Road,
London, SW6 3EG, UK

=head1 SEE ALSO

L<Net::FTPServer(3)>, L<perl(1)>

=cut
