package ME::VFS::Node;

use strict;
use warnings;
use Carp;

use Scalar::Util qw( weaken );
use ME::VFS::Cache;
use ME::VFS::RootNode;
use ME::VFS::DirNode;
use ME::VFS::FileNode;
use Encode;
use Digest::MD5 qw(md5_hex);
use Socket;

sub new {
  my $class = shift;
  my $VFS = shift;
  my $NodeId = shift;

  croak "Need a VFS" unless $VFS;
  croak "ID must be numeric!" unless $NodeId =~ m/^\d+$/;

  return ME::VFS::RootNode->new($VFS) if $NodeId == 1;

  my $Self = bless {
    VFS => $VFS,
    NodeId => $NodeId,
    Data => {},
  }, ref($class) || $class;

  if ($class eq 'ME::VFS::Node') {
    if ($Self->IsDir()) {
      bless $Self, 'ME::VFS::DirNode';
    } elsif ($Self->IsFile()) {
      bless $Self, 'ME::VFS::FileNode';
    } else {
      die "Unknown node type " . $Self->NodeType();
    }
  }

  # we don't really exist, but have to be loadable to stop race conditions
  # where the node has just been deleted and is needed in cleanup checks.
  # if being newly created though, we just don't exist!
  #return undef if $Self->IsDeleted();

  return $Self;
}

sub SubNode {
  my $Self = shift;
  my $NodeId = shift;
  my $Type = shift;

  my $NewNode = bless {
    VFS => $Self->{VFS},
    NodeId => $NodeId,
    ParentNode => $Self,
  }, $Type || 'ME::VFS::Node';

  weaken $NewNode->{ParentNode};

  if (ref($NewNode) eq 'ME::VFS::Node') {
    if ($NewNode->IsDir()) {
      bless $NewNode, 'ME::VFS::DirNode';
    } elsif ($NewNode->IsFile()) {
      bless $NewNode, 'ME::VFS::FileNode';
    } else {
      die "Unknown node type " . $NewNode->NodeType();
    }
  }

  return $NewNode;
}

sub Equals {
  my $Self = shift;
  my $Node = shift;
  return ($Self->NodeId() == $Node->NodeId());
}

=item I<CreateAlternativeByName( $AlternativeName )>

Returns an object which is an alternative representation of
this node.  It can be treated much like a Node, but is not actually one!

=cut
sub CreateAlternativeByName {
  my $Self = shift;
  my $AlternativeName = shift;
  
  return $Self unless $AlternativeName;
  return $Self if ( $AlternativeName eq 'raw' );
  return ME::VFS::NodeAlternatives->CreateByName( 
    Node => $Self, AlternativeName => $AlternativeName );
}

sub NodeData {
  my $Self = shift;
  my $NodeData = ME::VFS::Cache::Get('N', $Self->{NodeId})
      || croak "Can't load node " . $Self->{NodeId};
  return $NodeData;
}

sub VFS {
  my $Self = shift;
  return $Self->{VFS};
}

sub NodeId {
  my $Self = shift;
  return $Self->{NodeId};
}

sub NodeType {
  my $Self = shift;
  return $Self->NodeData->{NodeType};
}

sub NodeName {
  my $Self = shift;
  return $Self->NodeData->{NodeName};
}

sub UserId {
  my $Self = shift;
  return $Self->NodeData->{UserId};
}

sub Modified {
  my $Self = shift;
  return $Self->NodeData->{Modified};
}

sub Created {
  my $Self = shift;
  return $Self->NodeData->{Created};
}

sub ParentNodeId {
  my $Self = shift;
  return $Self->NodeData->{ParentNodeId};
}

sub IsDeleted {
  my $Self = shift;
  return $Self->NodeData->{IsDeleted};
}

sub TempExpires {
  my $Self = shift;
  return $Self->NodeData->{TempExpires};
}

sub ParentNode {
  my $Self = shift;
  unless ($Self->{Data}{ParentNode}) {
    $Self->{Data}{ParentNode} = $Self->VFS()->Node($Self->ParentNodeId(), IgnoreACLs => 1);
  }
  return $Self->{Data}{ParentNode};
}

# Create code

sub Create {
  my $class = shift;
  my %Opts = @_;

  die "Filename must not contain / ($Opts{NodeName}))\n"
    if $Opts{NodeName} =~ m{/};

  die "Filename too long ($Opts{NodeName})\n"
    if length(Encode::encode_utf8($Opts{NodeName})) >= 254;

  my $ParentNode;
  unless ($ParentNode = delete $Opts{ParentNode}) {
    $ParentNode = $Opts{VFS}->Node($Opts{ParentNodeId});
  }

  die "ReadOnly Filesystem\n"
    if $ParentNode->VFS->IsReadOnly();

  unless ($ParentNode->CanWrite()) {
    Carp::cluck($ParentNode, \%Opts);
    die "Can't write to parent dir\n";
  }

   die "Can't create because of locks\n" if $ParentNode->NameLocked($Opts{NodeName}, $Opts{IfLocks});

  my $ThisNode = $ParentNode->ChildByName($Opts{NodeName});

  if ($ThisNode) {
    if ($Opts{Overwrite}) {
      $ThisNode->Delete(%Opts, SkipCommit => 1, SkipParentModified => 1);
    } else {
      die "File already exists " . $ThisNode->Path() . "\n" unless $Opts{IgnoreErrors};
      return;
    }
  }

  if (!Encode::is_utf8($Opts{NodeName}) && $Opts{NodeName} =~ /[^\x00-\x7f]/) {
    Carp::cluck("NOT A UTF8 VALUE WITH 8-BIT DATA $Opts{NodeName}\n");
  }

  my $now = time();

  my $NodeId = ME::VFS::Cache::Create('N', {
    NodeName => $Opts{NodeName},
    NodeType => $Opts{NodeType},
    UserId => $ParentNode->UserId(),
    ParentNodeId => $ParentNode->NodeId(),
    Modified => $Opts{Modified} || $now,
    Created => $Opts{Created} || $now,
    TempExpires => $Opts{TempExpires} || undef,
    Sha1Hex => $Opts{Sha1Hex} || undef,
    FileSize => $Opts{FileSize} || 0, # not null
    IsDeleted => 0,
  });

  die "Failed to create Node " . $ParentNode->Path() . "/$Opts{NodeName}\n"
    unless ($NodeId or $Opts{IgnoreErrors});
  return unless $NodeId;

  $ParentNode->Touch(%Opts, SkipCommit => 1) unless $Opts{SkipParentModified};

  my $Self = $ParentNode->VFS()->Node($NodeId);

  ME::VFS::Cache::Commit() unless $Opts{SkipCommit};

  return $Self;
}

sub Delete {
  my $Self = shift;
  my %Opts = @_;

  # check filesystem changes allowed
  die "ReadOnly Filesystem\n"
    if $Self->VFS->IsReadOnly();

  # check that this user may do them
  my $ParentNode = $Self->ParentNode();
  die "No permission to write directory for delete\n"
    unless ($Opts{IgnoreACLs} or $ParentNode->CanWrite());

  die "Can't delete because of locks\n" if $ParentNode->NameLocked($Self->NodeName(), $Opts{IfLocks}, 'w');

  # mark the node for future cleanup
  my $NodeId = $Self->NodeId();
  my $now = time();
  ME::VFS::Cache::Update('N', $NodeId, {
    IsDeleted => $Opts{IsDeleted} || $NodeId,
    TempExpires => $now + (7 * 24 * 60 * 60), # 1 week
  });

  $ParentNode->Touch(%Opts, SkipCommit => 1) unless $Opts{SkipParentModified};

  # XXX - delete locks, ACLs, websites, etc
  # at this point!

  # and finally commit.
  ME::VFS::Cache::Commit() unless $Opts{SkipCommit};

  return;
}

sub UnDelete {
  my $Self = shift;
  my %Opts = @_;

  # check filesystem changes allowed
  die "ReadOnly Filesystem\n"
    if $Self->VFS->IsReadOnly();

  # check that this user may do them
  my $ParentNode = $Self->ParentNode();
  die "No permission to write directory for delete\n"
    unless ($Opts{IgnoreACLs} or $ParentNode->CanWrite());

  die "Can't undelete because of locks\n" if $ParentNode->NameLocked($Self->NodeName(), $Opts{IfLocks}, 'w');

  # mark the node undelete time
  my $NodeId = $Self->NodeId();
  ME::VFS::Cache::Update('N', $NodeId, {
    IsDeleted => 0,
    TempExpires => undef,
  });

  $ParentNode->Touch(%Opts, SkipCommit => 1) unless $Opts{SkipParentModified};

  # XXX - delete locks, ACLs, websites, etc
  # at this point!

  # and finally commit.
  ME::VFS::Cache::Commit() unless $Opts{SkipCommit};

  return;
}

sub Chown {
  my $Self = shift;
  my %Opts = @_;

  my $NewUserId = $Opts{NewUserId};

  die "Need a ToUserId"
    unless $NewUserId;

  # check filesystem changes allowed
  die "ReadOnly Filesystem\n"
    if $Self->VFS->IsReadOnly();

  die "No write permission\n"
    unless $Self->CanWrite();

  die "Can't modify because of locks\n"
    if $Self->ParentNode->NameLocked($Self->NodeName(), $Opts{IfLocks}, 'w');

  # mark the node changed
  ME::VFS::Cache::Update('N', $Self->NodeId(), {
    UserId => $NewUserId,
  });

  # and finally commit.
  ME::VFS::Cache::Commit() unless $Opts{SkipCommit};

  return;
}

sub SetModified {
  my $Self = shift;
  my $NewModified = shift;
  my %Opts = @_;

  # check filesystem changes allowed
  die "ReadOnly Filesystem\n"
    if $Self->VFS->IsReadOnly();

  die "No write permission\n"
    unless $Self->CanWrite();

  die "Can't modify because of locks\n"
    if $Self->ParentNode->NameLocked($Self->NodeName(), $Opts{IfLocks}, 'w');

  # mark the node changed
  ME::VFS::Cache::Update('N', $Self->NodeId(), {
    Modified => $NewModified,
  });

  # and finally commit.
  ME::VFS::Cache::Commit() unless $Opts{SkipCommit};

  return;
}

sub SetExpires {
  my $Self = shift;
  my $NewExpires = shift;
  my %Opts = @_;

  # check filesystem changes allowed
  die "ReadOnly Filesystem\n"
    if $Self->VFS->IsReadOnly();

  die "No write permission\n"
    unless $Self->CanWrite();

  die "Can't modify because of locks\n"
    if $Self->ParentNode->NameLocked($Self->NodeName(), $Opts{IfLocks}, 'w');

  # mark the node changed
  ME::VFS::Cache::Update('N', $Self->NodeId(), {
    TempExpires => $NewExpires,
  });

  # and finally commit.
  ME::VFS::Cache::Commit() unless $Opts{SkipCommit};

  return;
}

sub IsRootNode {
  my $Self = shift;
  return 0; # overridden in RootNode
}

sub MoveTo {
  my $Self = shift;
  my %Opts = @_;

  # normalise options
  my $OldParent = $Self->ParentNode();
  my $OldName = $Self->NodeName();
  my $NewParent = $Opts{ParentNode} || $OldParent;
  my $NewName = $Opts{NodeName} || $OldName;
  my $Dest = $Opts{Node};
  if ($Dest) {
    $NewParent = $Dest->ParentNode();
    $NewName = $Dest->NodeName();
  } else {
    $Dest = $NewParent->ChildByName($NewName);
  }

  if ($Self->VFS->IsReadOnly()) {
    die "ReadOnly Filesystem\n";
  }

  # A move needs to write to the source dir or it's really just a copy!
  unless ($Opts{IgnoreACLs} or $OldParent->CanWrite()) {
    if (delete $Opts{FallbackToCopy}) {
      return $Self->CopyTo(%Opts);
    } else {
      die "No permission to write source dir for move, try copy instead\n";
    }
  }

  unless ($Opts{IgnoreACLs}) {
    die "No permission to read source for move\n" unless $Self->CanRead();
    die "No permission to write destination dir for move\n" unless $NewParent->CanWrite();
  }
  die "Can't move into a sub directory of self\n" if $NewParent->HasAncestor($Self);
  die "Can't move due to locks\n" if $OldParent->NameLocked($OldName, $Opts{IfLocks}, 'w');
  die "Can't move due to locks\n" if $NewParent->NameLocked($NewName, $Opts{IfLocks}, 'w');

  # check that it doesn't already exist
  if ($Dest) {
    return $Self if $Dest->NodeId() == $Self->NodeId();
    if ($Opts{Overwrite}) {
      $Dest->Delete(%Opts, SkipCommit => 1, SkipParentModified => 1);
    }
    elsif ($Opts{TryRename}) {
      my $BaseName = $NewName;
      my $n = 2;
      while ($n < 1000) {
        $NewName = ME::VFS::IOHandle::_GenerateName($BaseName, $n++);
        $Dest = $NewParent->ChildByName($NewName);
        last unless $Dest;
        return $Self if $Dest->NodeId() == $Self->NodeId();
      }
      die "Unable to create a temporary file based on $BaseName\n" if $Dest;
    }
    else {
      die "Destination already exists\n";
    }
  }

  # make the changes
  my %ExtraItems;
  unless ($Opts{KeepTemporary}) {
   $ExtraItems{TempExpires} = undef;
  }
  unless ($Opts{KeepTimes}) {
    $ExtraItems{Modified} = time();
  }
  unless ($NewParent->UserId() == $OldParent->UserId()) {
    $ExtraItems{UserId} = $NewParent->UserId();
  }
  ME::VFS::Cache::Update('N', $Self->NodeId(), {
    %ExtraItems,
    ParentNodeId => $NewParent->NodeId(),
    NodeName => $NewName,
  });

  $OldParent->Touch(%Opts, SkipCommit => 1) unless $Opts{SkipParentModified};
  unless ($OldParent->Equals($NewParent)) {
    $NewParent->Touch(%Opts, SkipCommit => 1) unless $Opts{SkipParentModified};
  }

  # some really local cached values..
  delete $Self->{Data};

  ME::VFS::Cache::Commit() unless $Opts{SkipCommit};

  return $Self;
}

sub CopyTo {
  my $Self = shift;
  my %Opts = @_;

  # normalise options
  my $OldParent = $Self->ParentNode();
  my $OldName = $Self->NodeName();
  my $NewParent = $Opts{ParentNode} || $OldParent;
  my $NewName = $Opts{NodeName} || $OldName;
  my $Dest = $Opts{Node};
  if ($Dest) {
    $NewParent = $Dest->ParentNode();
    $NewName = $Dest->NodeName();
  } else {
    $Dest = $NewParent->ChildByName($NewName);
  }

  if ($Self->VFS->IsReadOnly()) {
    die "ReadOnly Filesystem\n";
  }

  unless ($Opts{IgnoreACLs}) {
    die "No permission to read source for copy\n" unless $Self->CanRead();
    die "No permission to write destination dir for copy\n" unless $NewParent->CanWrite();
  }
  die "Can't copy into a sub directory of self\n" if $NewParent->HasAncestor($Self);
  die "Can't copy due to locks\n" if $OldParent->NameLocked($OldName, $Opts{IfLocks}, 'r');
  die "Can't copy due to locks\n" if $NewParent->NameLocked($NewName, $Opts{IfLocks}, 'w');

  # check that it doesn't already exist
  if ($Dest) {
    if ($Dest->NodeId() == $Self->NodeId()) {
      die "Can't copy over self\n";
    }
    if ($Opts{Overwrite}) {
      $Dest->Delete(%Opts, SkipCommit => 1, SkipParentModified => 1);
    }
    elsif ($Opts{TryRename}) {
      my $BaseName = $NewName;
      my $n = 2;
      while ($n < 1000) {
        $NewName = ME::VFS::IOHandle::_GenerateName($BaseName, $n++);
        $Dest = $NewParent->ChildByName($NewName);
        last unless $Dest;
      }
      die "Unable to create a temporary file based on $BaseName\n" if $Dest;
    }
    else {
      die "Destination already exists for copy\n";
    }
  }

  # copy details
  my %NodeData = %{$Self->NodeData};
  delete $NodeData{ParentNodeId};
  delete $NodeData{NodeId};

  my %ExtraItems;
  unless ($Opts{KeepTimes}) {
    my $time = time();
    $ExtraItems{Created} = $time;
    $ExtraItems{Modified} = $time;
  }
  unless ($NewParent->VFS->UserId() eq $OldParent->VFS->UserId()) {
    $ExtraItems{UserId} = $NewParent->VFS->UserId();
  }

  my $NewNode = $Self->Create(%NodeData, %ExtraItems,
    ParentNode => $NewParent,
    NodeName => $NewName,
    SkipCommit => 1,
    TempExpires => $Opts{TempExpires} || undef, # don't copy expiry
    SkipQuotaCheck => $Opts{SkipQuotaCheck},
  );

  # ACLs aren't copied, they get created from parent automatically.

  # Copy all the properties as is
  $NewNode->SetProperties(
    $Self->GetProperties(),
    SkipCommit => 1,
    ( $Opts{KeepTimes} ? ( SkipTouch => 1 ) : () )
  );

  $NewParent->Touch(%Opts, SkipCommit => 1) unless $Opts{SkipParentModified};

  # some really local cached values..
  delete $Self->{Data};

  ME::VFS::Cache::Commit() unless $Opts{SkipCommit};

  if ($Opts{NodeMap}) {
    $Opts{NodeMap}{$Self->NodeId()} = $NewNode->NodeId();
  }

  return $NewNode;
}

sub Touch {
  my $Self = shift;
  return unless $Self->CanWrite(); # no error
  return $Self->SetModified(time(), @_);
}

sub GetETag {
  my $Self = shift;
  return md5_hex($Self->NodeId() . $Self->Modified());
}

sub GetProperties {
  my $Self = shift;

  my $Ids = ME::VFS::Cache::Get('NP', $Self->NodeId());
  my %d;
  foreach my $NS (keys %$Ids) {
    foreach my $PN (keys %{$Ids->{$NS}}) {
      $d{$NS}{$PN} = ME::VFS::Cache::Get('P', $Ids->{$NS}{$PN});
    }
  }
  return \%d;
}

sub SetProperties {
  my $Self = shift;
  my $Props = shift;
  my %Opts = @_;

  if ($Self->VFS->IsReadOnly()) {
    die "ReadOnly Filesystem\n";
  }

  die "Can't set properties because of locks\n"
    if $Self->ParentNode->NameLocked($Self->NodeName(), $Opts{IfLocks}, 'w');

  unless ($Opts{IgnoreACLs}) {
    die "No permission to write node for properties\n" unless $Self->CanWrite();
  }

  my $OldProps = $Self->GetProperties();

  # remove stale nodes!
  foreach my $NS (keys %$Props) {
    foreach my $PN (keys %{$Props->{$NS}}) {
      # delete it if it exists!
      my $Val = $Props->{$NS}{$PN};
      my $OldVal = $OldProps->{$NS}{$PN};
      if ($OldVal->{PropertyId}) {
        if ($Val) {
          my $NewVal = { %$OldVal, %$Val };
          ME::VFS::Cache::Update('P', $OldVal->{PropertyId}, $NewVal);
        } else {
          ME::VFS::Cache::Delete('P', $OldVal->{PropertyId});
        }
      } elsif ($Val) {
        ME::VFS::Cache::Create('P', {
          NodeId => $Self->NodeId(),
          NamespaceURI => $NS,
          PropName => $PN,
          PropValue => $Val->{PropValue},
          IsXML => $Val->{IsXML},
        });
      }
    }
  }

  ME::VFS::Cache::Update( 'N', $Self->NodeId(), { Modified => time() } )
    unless $Opts{SkipTouch};

  ME::VFS::Cache::Commit() unless $Opts{SkipCommit};
}

# ACL stuff
sub ACLs {
  my $Self = shift;

  my $ACLIds = ME::VFS::Cache::Get('NA', $Self->NodeId());

  return $ACLIds;
}

sub HasACLs {
  my $Self = shift;
  my %Opts = @_;

  my $ACLs = $Self->ACLs();

  foreach my $item (keys %$ACLs) {
    next if ($Opts{IgnoreSelf} and $item == $Self->VFS->UserId());
    next unless $ACLs->{$item};
    next unless keys %{$ACLs->{$item}};
    return 1;
  }
  return 0;
}

sub AddACL {
  my $Self = shift;
  my $UserId = shift;
  my $ACLName = shift;
  my %Opts = @_;

  die "ReadOnly Filesystem\n" if $Self->VFS->IsReadOnly();
  die "No permission to alter ACLs\n" unless $Self->CanAdmin();

  my $ACLs = $Self->ACLs();

  my $Id = $ACLs->{$UserId}{$ACLName};

  unless ($Id) {
    $Id = ME::VFS::Cache::Create('A', {
      NodeId => $Self->NodeId(),
      UserId => $UserId,
      ACLName => $ACLName,
    });
  }

  ME::VFS::Cache::Commit() unless $Opts{SkipCommit};

  return $Id;
}

sub DelACL {
  my $Self = shift;
  my $UserId = shift;
  my $ACLName = shift;
  my %Opts = @_;

  die "ReadOnly Filesystem\n" if $Self->VFS->IsReadOnly();
  die "No permission to alter ACLs\n" unless $Self->CanAdmin();

  my $ACLs = $Self->ACLs();

  my $OldId = $ACLs->{$UserId}{$ACLName};

  if ($OldId) {
    ME::VFS::Cache::Delete('A', $OldId);
  }

  ME::VFS::Cache::Commit() unless $Opts{SkipCommit};

  return $OldId;
}


sub CanAdmin {
  my $Self = shift;
  return $Self->HasACL('a', $Self->VFS->UserId());
}

sub CanWrite {
  my $Self = shift;
  return $Self->HasACL('w', $Self->VFS->UserId());
}

sub CanRead {
  my $Self = shift;
  return $Self->HasACL('r', $Self->VFS->UserId());
}

sub IsDir {
  my $Self = shift;
  return $Self->NodeType eq 'd';
}

sub IsFile {
  my $Self = shift;
  return $Self->NodeType eq 'f';
}

sub Depth {
  my $Self = shift;
  unless ($Self->{Data}{Depth}) {
    $Self->{Data}{Depth} = $Self->ParentNode->Depth() + 1;
  }
  return $Self->{Data}{Depth};
}

sub Path {
  my $Self = shift;

  if ($Self->VFS->{Chroot}) {
    return '/' if $Self->Equals($Self->VFS->{Chroot});
  }

  unless (exists $Self->{Data}{Path}) {
    my $DirPath = $Self->ParentNode->Path();
    $DirPath = '' if $DirPath eq '/';
    $Self->{Data}{Path} = $DirPath . '/' . $Self->NodeName();
  }

  return $Self->{Data}{Path};
}

sub HasAncestor {
  my $Self = shift;
  my $Node = shift;
  return 1 if $Self->NodeId() == $Node->NodeId();
  return $Self->ParentNode->HasAncestor($Node);
}

sub ParentAtDepth {
  my $Self = shift;
  my $Depth = shift;

  die "Not deeper than $Depth" unless $Self->Depth() >= $Depth;
  my $Node = $Self;
  while ($Node->Depth() > $Depth) {
    $Node = $Node->ParentNode();
  }
  return $Node;
}

# This is subtly different than the creator UserId
sub PathOwnerUserId {
  my $Self = shift;
  return $Self->UserId() unless $Self->Depth() > 2;
  my $RootDir = $Self->ParentAtDepth(2);
  return $RootDir->UserId();
}

# NodeQuota support

sub QuotaNode {
  my $Self = shift;
  if ($Self->IsQuotaNode()) {
    return $Self;
  }
  else {
    return $Self->ParentNode->QuotaNode();
  }
}

# convenient access to user specific properties

sub GetPropertyValue {
  my $Self = shift;
  my $NS = shift;
  my $PN = shift;

  my $Props = $Self->GetProperties();

  return undef unless exists $Props->{$NS};
  return undef unless exists $Props->{$NS}{$PN};

  return $Props->{$NS}{$PN}{PropValue};
}

sub SetPropertyValue {
  my $Self = shift;
  my $NS = shift;
  my $PN = shift;
  my $Value = shift;
  my %Opts = @_;

  my $OldVal = $Self->GetPropertyValue($NS, $PN);

  return if (not defined $OldVal and not defined $Value);
  return if (defined $OldVal and defined $Value and $OldVal eq $Value);

  return $Self->SetProperties({
    $NS => {
      $PN => {
        PropValue => $Value,
        IsXML => 0,
      },
    },
  }, %Opts);
}

sub DelProperty {
  my $Self = shift;
  my $NS = shift;
  my $PN = shift;
  my %Opts = @_;

  my $Props = $Self->GetProperties();

  return undef unless exists $Props->{$NS};
  return undef unless exists $Props->{$NS}{$PN};

  return $Self->SetProperties({
    $NS => {
      $PN => undef,
    },
  }, %Opts);
}

sub HasProperty {
  my $Self = shift;
  my $NS = shift;
  my $PN = shift;

  my $Ids = ME::VFS::Cache::Get('NP', $Self->NodeId());
  return unless $Ids;
  return unless $Ids->{$NS};
  return unless $Ids->{$NS}{$PN};
  return 1;
}

sub GetUserProperty {
  my $Self = shift;
  my $NS = $Self->VFS->UserPropertyNamespace();
  return $Self->GetPropertyValue($NS, @_);
}

sub SetUserProperty {
  my $Self = shift;
  my $NS = $Self->VFS->UserPropertyNamespace();
  # User properties are namespaced to that user, so allow setting them
  #  even on read-only nodes. We need this so you can set
  #  IsExpanded, DisplayRows, IsSelected, etc
  return $Self->SetPropertyValue($NS, @_, IgnoreACLs => 1);
}

sub DelUserProperty {
  my $Self = shift;
  my $NS = $Self->VFS->UserPropertyNamespace();
  return $Self->DelProperty($NS, @_);
}

sub HasUserProperty {
  my $Self = shift;
  my $NS = $Self->VFS->UserPropertyNamespace();
  return $Self->HasProperty($NS, @_);
}

sub GetSysProperty {
  my $Self = shift;
  my $NS = $Self->VFS->SysPropertyNamespace();
  return $Self->GetPropertyValue($NS, @_);
}

sub SetSysProperty {
  my $Self = shift;
  my $NS = $Self->VFS->SysPropertyNamespace();
  return $Self->SetPropertyValue($NS, @_);
}

sub DelSysProperty {
  my $Self = shift;
  my $NS = $Self->VFS->SysPropertyNamespace();
  return $Self->DelProperty($NS, @_);
}

sub HasSysProperty {
  my $Self = shift;
  my $NS = $Self->VFS->SysPropertyNamespace();
  return $Self->HasProperty($NS, @_);
}

# pretty things for web interface

sub ShortPath {
  my $Self = shift;

  my $HomePath = $Self->VFS->HomeDir()->Path();
  my $TempPath = $Self->VFS->TempDir()->Path();
  my $Path = $Self->Path();

  if ($Path =~ m{^$HomePath(.*)}) {
    return "{My Files}$1";
  }
  elsif ($Path =~ m{^$TempPath(.*)}) {
    return "{My Temporary Space}$1";
  } 
  elsif ($Path =~ m{^/([^\.]+)[^/]+/files(.*)}) {
    return "{$1's Files}$2";
  }
  else {
    return $Path;
  }
}

sub ShortName {
  my $Self = shift;

  if ($Self->Depth() == 2) {
    return $Self->ShortPath();
  }

  return $Self->NodeName();
}

sub PathRelativeTo {
  my $Self = shift;
  my $RelNode = shift;

  return '' if $Self->NodeId == $RelNode->NodeId();

  my $ParentPath = $Self->ParentNode->PathRelativeTo($RelNode);
  if ($ParentPath eq '') {
    return $Self->NodeName();
  } else {
    return $ParentPath . '/' . $Self->NodeName();
  }
}

sub IsHomeDir { 0; }

# Lock stuff - XXX - needs access control!

sub NameLocked {
  my $Self = shift;
  my $NodeName = shift;
  my $HaveLocks = shift;
  my $CheckMode = shift;

  # check that we have any locks at all
  my $Locks = $Self->Locks($NodeName);
  return () unless keys %{$Locks};

  # are we holding any of those locks validly with this session?
  my %Ignore;
  foreach my $Lock (::AsList($HaveLocks)) {
    my $Id = $Locks->{$Lock};
    next unless $Id;
    $Ignore{$Lock} = 1;
  }

  # are any of the remaining locks not timed out?
  my $time = time();
  my @NameLocked;
  foreach my $Lock (keys %{$Locks}) {
    my $Id = $Locks->{$Lock};
    next unless $Id;
    my $Item = ME::VFS::Cache::Get('L', $Id);
    next unless $Item;
    if ($Item->{Expires} < $time) {
      ME::VFS::Cache::Delete('L', $Id);
      next;
    }
    next if $Ignore{$Lock};
    next if ($CheckMode eq 'r' and $Item->{IsExclusive} == 0);
    push @NameLocked, $Lock;
  }

  return @NameLocked;
}

sub Locks {
  my $Self = shift;
  my $NodeName = shift;

  my $Locks = ME::VFS::Cache::Get('NL', $Self->NodeId());
  return unless $Locks;

  $Locks = $Locks->{$NodeName} if $NodeName;
  return unless $Locks;

  return wantarray ? %{$Locks} : $Locks;
}

sub NewLock {
  my $Self = shift;
  my $NodeName = shift;
  my $Owner = shift;
  my $IsExclusive = shift;
  my $IsWrite = shift;
  my $Expires = shift;
  my $IsDeep = shift;

  my $UUID = $Self->VFS->MakeUUID();

  $Self->AddLock($NodeName, $Self->UserId(), $Owner, $UUID, $IsExclusive, $IsWrite, $Expires, $IsDeep);
  ME::VFS::Cache::Commit();

  return $UUID;
}

sub GetLock {
  my $Self = shift;
  my $NodeName = shift;
  my $LockName = shift;

  my $Locks = $Self->Locks($NodeName);

  my $Id = $Locks->{$LockName};

  if ($Id) {
    my $Lock = ME::VFS::Cache::Get('L', $Id);
    return $Lock;
  }

  return undef;
}

sub RemoveLock {
  my $Self = shift;
  my $NodeName = shift;
  my $LockName = shift;

  my $rc = $Self->DelLock($Self->UserId(), $NodeName, $LockName);
  ME::VFS::Cache::Commit();

  return $rc;
}

sub GetLocks {
  my $Self = shift;
  # disable this too!
  my $NodeName = shift;
  my $Locks = $Self->Locks($NodeName);
  my %Resp = map { $_ => ME::VFS::Cache::Get('L', $Locks->{$_}) } keys %{$Locks};
  return wantarray ? %Resp : \%Resp;
}

sub RefreshLock {
  my $Self = shift;
  my $NodeName = shift;
  my $LockName = shift;
  my $Expires = shift;

  my $Locks = $Self->Locks($NodeName);

  my $Id = $Locks->{$LockName};

  if ($Id) {
    ME::VFS::Cache::Update('L', $Id, {
      Expires => $Expires,
    });
  }
  ME::VFS::Cache::Commit();

  return undef;
}

sub AddLock {
  my $Self = shift;
  my $NodeName = shift;
  my $UserId = shift;
  my $Owner = shift;
  my $LockName = shift;
  my $IsExclusive = shift;
  my $IsWrite = shift;
  my $Expires = shift;
  my $IsDeep = shift;

  if ($Self->VFS->IsReadOnly()) {
    die "ReadOnly Filesystem\n";
  }

  die "No permission to alter Locks\n" unless $Self->CanWrite();

  my $Locks = $Self->Locks($NodeName);

  my $OldId = $Locks->{$LockName};

  my @Others = map { ME::VFS::Cache::Get('L', $Locks->{$_}) } grep { $_ ne $LockName } keys %{$Locks};

  # The table below describes the behavior that occurs when a lock
  # request is made on a resource.
  #
  # Current lock state/  |   Shared Lock   |   Exclusive
  # Lock request         |                 |   Lock
  # =====================+=================+==============
  # None                 |   True          |   True
  # ---------------------+-----------------+--------------
  # Shared Lock          |   True          |   False
  # ---------------------+-----------------+--------------
  # Exclusive Lock       |   False         |   False*
  # ------------------------------------------------------
  foreach my $Other (@Others) {
    die "Clashes with existing lock $Other->{LockName}"
      if ($IsExclusive || $Other->{IsExclusive});
    # writeness clash not specified in the spec!
  }

  if ($OldId) {
    ME::VFS::Cache::Update('L', $OldId, {
      IsDeep => $IsDeep || 0,
      Expires => $Expires,
      IsExclusive => $IsExclusive || 0,
      IsWrite => $IsWrite || 0,
      Owner => $Owner,
    });
  } else {
    my $Id = ME::VFS::Cache::Create('L', {
      NodeId => $Self->NodeId(),
      UserId => $UserId,
      NodeName => $NodeName,
      LockName => $LockName,
      Expires => $Expires,
      IsDeep => $IsDeep || 0,
      IsExclusive => $IsExclusive || 0,
      IsWrite => $IsWrite || 0,
      Owner => $Owner,
    });
  }

  # Note, you need to call Commit yourself

  return;
}

sub DelLock {
  my $Self = shift;
  my $UserId = shift;
  my $NodeName = shift;
  my $LockName = shift;

  die "No permission to alter Locks\n" unless $Self->CanWrite();

  my $Locks = $Self->Locks($NodeName);

  my $OldId = $Locks->{$LockName};

  if ($OldId) {
    ME::VFS::Cache::Delete('L', $OldId);
    return 1;
  }
  else {
    return 0;
  }

  # Note, you need to call commit yourself
}

sub _PostMoveDeep {
  my $Self = shift;
  # Here we sort out locks eventually?
}

1;
